#|
  module	: abstraction
  owner		: Franz Weber
  author	: Franz Weber/Olaf Mueller
  explanation	:



        Abstractions
        ---------------

        (defflavor abstraction (expression-node) (expression-node declaration)
            (:gettable-instance-variables expression-node)
            (:settable-instance-variables expression-node)
            (:init-keywords :expression-node)
        )       

        abstractions only have the position 't'

                a. generic implementations
                --------------------------

                        compare
                        copy
                        deallocate
                        index-of, set-elem, get-elem,
                            transfer-elem-out, transfer-elem-in,
                            iterate-non-recursive-method
                        apply-sub
                        expression-type
                        is-constant
                        is-variable

                        beta-reduce
                        beta-eta-reduce
                        delta-reduce

                        attribute-string



   history     :

        20.04.89
        90.06.12 Franz Weber
                 flavor base version
        91.03.04 Franz Weber
                 switch to typed lambda calculus
        91.04.13 Franz Weber
                 beta reduction inserted
	92.06.04 Olaf Mueller
		 unification extended

   errors      : equal between expressions does not work
   bugs        : 
   todo        : the mapping identity may be implemented by
                 a key stack mapping with structure shareing
                 on copy

                 copying of expression nodes is not done

                 :expression-type returns a type node
                 :type returns a type


|# 

(mod-require "data/glo/v2/simple-lin")
(mod-require "data/glo/v2/cop")
(mod-require "data/glo/v2/pointer")
(mod-require "data/tre/v2/simple-tre")
(mod-require "data/stack/v2/key-stack-list")
(mod-require "logic/v2/expression/declaration/impls/declaration")
(mod-require "logic/v2/expression/type/impls/type")
(mod-require "logic/v2/expression/basic-type/impls/basic-type")
(mod-require "logic/v2/expression/function-type/impls/function-type")
(mod-require "logic/v2/expression/parse-gen/impls/parse-gen")


; -------------------- NEW

(defflavor abstraction (expression-node) (expression-node declaration)
    (:gettable-instance-variables expression-node)
    (:settable-instance-variables expression-node)
    (:init-keywords :expression-node)
)
; settable von expression-node wird in parse-Regeln gebraucht!!


; -------- DEFINIT
; this method replaces a definitmethod, which is here unpossible,
; because the instance variable exp-node is at the time of make-instance
; not yet bounded. Remember: expression-node instead of en as parameter
; would generate a name-conflict here!
; for setting of expression-node out of binder, h,a do:
; (setq expression-node (assemble-exp [binder :cdr] head arguments))
(defmethod (abstraction :set-expression-node) (en)
 		(setq expression-node en)
		(setq binder [expression-node :binder |:snoc self])
		; remember: abstractions in binder are stored in reverse order!
		(setq head [expression-node :head])
		(setq arguments [expression-node :arguments])
                self
)


(defmethod (abstraction :compare) (right)
    (if (eql self right) 'equal
    (if (typep right 'abstraction)
        (progn
            [*exp-identity* :push-key self right]
            (prog1
                [expression-node :compare [right :expression-node]]
                [*exp-identity* :pop]
            )
        )
        (flavor-compare 'abstraction (type-of right))
    )
    )
)

(defmethod (abstraction :equal) (right)
    (or (eql self right)
       (and (typep right 'abstraction)
           (progn
               [*exp-identity* :push-key self right]
               (prog1
                   [:expression-node |:equal [right :expression-node]]
                   [*exp-identity* :pop]
               )
           )
       )
    )
)

; old
#|
(defmethod (abstraction :equal) (right)
    ;(let ((s (or [*exp-identity* :get self] self))
          ;(r (or [*exp-identity* :get right] right))
    ; ))
    (let ((s (or [*exp-identity* :get self :test 'eql] self))
          (r (or [*exp-identity* :get right :test 'eql] right))
         )
        (eql s r)
    )
)


; ----- copies

(defmethod (abstraction :copy-from) (from)
    (setq picture [from :picture])
    (setq type [from :type])
    [*exp-identity* :push-key from self]
    (setq expression-node [[from :expression-node] :copy])
    [*exp-identity* :pop]
    self 
)

(defmethod (abstraction :deallocate) () 
    (setq expression-node nil)
    (setq picture nil)
    (setq type nil)
)
|#

; ----- test

; inherited from declaration

; ----- simple aggregate

; positions is 'expression-node

(defmethod (abstraction :index-of) (pos) t)
; (defmethod (abstraction :set-elem) (i e) (setq expression-node e))
(defmethod (abstraction :get-elem) (i) expression-node)
; (defmethod (abstraction :transfer-elem-out) (i) 
;    (prog1 expression-node (setq expression-node nil)))
(defmethod (abstraction :transfer-elem-in) (i e)
    (setq expression-node e)
)

(defiterationmethod (abstraction :iterate-non-recursive-method) (f) 
    (funcall f expression-node t))

; ----- expression-nodes
(defmethod (abstraction :apply-sub) (sub)
    (let ((new (make-instance (type-of self) :picture picture :type type)))
        [*exp-identity* :push-key self new]
        [new :set-expression-node [expression-node :apply-sub sub]]
        [*exp-identity* :pop]
        new
    )
)

(defmethod (abstraction :expression-type) ()

    (let ((et [:expression-node |:expression-type]))
        (when et
          (make-instance 'function-type
              :domain [:type]
              :range  et
          )
       )
   )
)

; ----- reductions ; a simple variation of copy, does this give a hint ?
                   ; say some kind of mapping ???????

(defmethod (abstraction :beta-reduce) ()
    (let ((res (make-instance (type-of self) :picture picture :type type)))
        [*exp-identity* :push-key self res]
        [res :set-expression-node [[:expression-node] :beta-reduce]]
        [*exp-identity* :pop]
        res
    )
)

(defmethod (abstraction :beta-eta-reduce) ()
    (if (let ((en [:expression-node]))
            (and
                (typep en 'application)
                (let ((fn [en :functor])
                      (ar [en :argument])
                     )
                     (and
                         (typep ar 'reference)
                         (eql [ar :ref] self)
                         (not
                            [fn :exists 'recursive 'reflexive (x)
                                (and
                                    (typep x 'reference)
                                    ; could it be 'eql ??? @@@@
                                    (eql [x :ref] [ar :ref])
                                )
                            ]
                         )
                     )
                )
            )
        )
       ; eta redex x : (F x)
       [[[:expression-node] :functor] :beta-eta-reduce]
       ; no eta redex
       (let ((res (make-instance (type-of self) :picture picture :type type)))
           [*exp-identity* :push-key self res]
           [res :set-expression-node [[:expression-node] :beta-eta-reduce]]
           [*exp-identity* :pop]
           res
       )
    )
)
; not complete yet
(defmethod (abstraction :delta-reduce) ()
    (let ((res (make-instance (type-of self) :picture picture :type type)))
        [*exp-identity* :push-key self res]
        [res :set-expression-node [[:expression-node] :delta-reduce]]
        [*exp-identity* :pop]
        res
    )
)

; ----- declarations
(defmethod (abstraction :is-contant) () t)
(defmethod (abstraction :is-variable) () nil)

; ----- test

(defmethod (abstraction :attribute-string) () [:picture])
(defmethod (abstraction :attribute-lis) ()
    (make-instance 'lis-list :init-value (list [:type]))
)



