#|
   module      : basic-type
   owner       : Franz Weber
   author      : Franz Weber/Olaf Mueller
   explanation :


        Basic Types
        --------------

        (defflavor basic-type 
            (picture) 
            (type-node)
            (:gettable-instance-variables picture)
            (:settable-instance-variables picture)
            (:init-keywords picture)
        )

                a. generic implementations
                --------------------------

                        compare
                        deallocate
                        copy-from

   history     :

        91.03.04 Franz Weber
                 flavor base version
	92.06.04 Olaf Mueller
		 extended for unification

   errors      :
   bugs        : 
   todo        : 
|# 

(mod-require "logic/v2/expression/parse-gen/impls/parse-gen")
(mod-require "data/tre/v2/simple-tre")

; ------------------- Basic Type

(defflavor basic-type 
    (picture) 
    (type-node)
    (:gettable-instance-variables picture)
    (:settable-instance-variables picture)
    (:init-keywords picture)
)

; ----- comparison
(defmethod (basic-type :compare) (right)
    (if (typep right 'basic-type)
        (flavor-compare [:picture] [right :picture])
        (flavor-compare 'basic-type (type-of right))
    )
)

; copy
(defmethod (basic-type :copy-from) (from)

    [:set-picture [from :picture]]
    self
)

(defmethod (basic-type :deallocate) () (setf picture nil))


; ----- simple aggregate

(defmethod (basic-type :elem-error) () (error "basic-type have no elems"))
(defmethod (basic-type :index-of) (pos) [:elem-error])
(defmethod (basic-type :set-elem) (i e) [:elem-error])
(defmethod (basic-type :get-elem) (i) [:elem-error])
(defmethod (basic-type :transfer-elem-in) (i e) [:elem-error])
(defmethod (basic-type :transfer-elem-out) (i) [:elem-error])

(defiterationmethod (basic-type :iterate-method) (f)
            (when reflexive (funcall f self index-of-self)))

; ----- test
(defmethod (basic-type :attribute-string) () [:picture])

