#|
   module      : function-type
   owner       : Franz Weber
   author      : Franz Weber/Olaf Mueller
   explanation :

        4. Function Types
        -----------------

        (defflavor function-type 
            (domain range) 
            (type-node)
            (:gettable-instance-variables domain range)
            (:settable-instance-variables domain range)
            (:init-keywords domain range)
        )

                a. generic implementations
                --------------------------

                        compare
                        copy
                        deallocate
                        index-of, set-elem, get-elem
                            transfer-elem-out, transfer-elem-in
                            iterate-non-recursice-method

                        positions are 'domain 'range

   history     :

        91.03.04 Franz Weber
                 flavor base version
	92.06.04 Olaf Mueller
		 extended for unification

   errors      :
   bugs        : 
   todo        : 
|# 

(mod-require "logic/v2/expression/parse-gen/impls/parse-gen")
(mod-require "data/tre/v2/simple-tre")


; ------------------ Function Type
(defflavor function-type 
    (domain range) 
    (type-node)
    (:gettable-instance-variables domain range)
    (:settable-instance-variables domain range)
    (:init-keywords domain range)
)

; ----- comparison

(defmethod (function-type :compare) (right)
    (if (typep right 'function-type)
        (let ((res [range :compare [right :range]]))
            (if (equal res 'equal)
                [domain :compare [right :domain]]
                res
            )
        )
        (flavor-compare (type-of right) (type-of self))
    )
)

#|
; ----- copies
(defmethod (function-type :copy-from) (from)
    (setq range [[from :range] :copy])
    (setq domain [[from :domain] :copy])
    self
)

(defmethod (function-type :deallocate) ()
    (setq domain nil)
    (setq range nil)
)
|#

; ----- simple aggregate

(defmethod (function-type :index-of) (pos) pos)
(defmethod (function-type :get-elem) (i)
    (case i
        (domain domain)
        (range range)
    )
)
#|
(defmethod (function-type :set-elem) (i e)
    (case i
        (domain (setq domain e))
        (range (setq range e))
    )
)
(defmethod (function-type :transfer-elem-in) (i e)
    (case i
        (domain (setq domain e))
        (range (setq range e))
    )
)
|#
(defmethod (function-type :transfer-elem-out) (i)
    (case i
        (domain (prog1 domain (setq domain nil)))
        (range (prog1 range (setq range nil)))
    )
)

(defiterationmethod (function-type :iterate-non-recursive-method) (f)
    (funcall f domain 'domain)
    (funcall f range 'range)
)


; ------------------------  N  E  W  --------------

; -------------------- concerning unification
; -------- type/type-node?

; compute the types that are the arguments of a type-expression
; example: ((U V) W), i.e. W->(V->U) ==> (W,V)
; output is a lis-list-readonly of type-nodes 
(defmethod (basic-type   :argument-types)() (make-instance 'lis-list-readonly))
(defmethod (function-type :argument-types)() 
  [range :argument-types |:cons domain])
  

; very clumsy because :insert does'nt return self; my idea:
; [[range :argument-types] :insert domain])

; compute basic-type that results if an expression of type self is fed with
; as many arguments as possible 
(defmethod (basic-type    :output-type)() self)
(defmethod (function-type :output-type)() [range :output-type])

#|
; see new assemble-type downwards!
; make a copy of t replacing its output-type with o
(defmethod (basic-type    :new-output-type) (o) [o :copy])
(defmethod (function-type :new-output-type) (o)
  (make-instance 'function-type :domain [[:domain] :copy] 
		               :range  [[:range] :new-output-type o]))
|#


; ---------------- constructing function
(defun assemble-type (output-type input-types) 
	[input-types :reduce :iteration-order 'from-last (x) (p output-type)
                (make-instance 'function-type 
			       :domain x 
			       :range p)]

)


