#|
   module      : sub
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 
   
	substitutions

	substitutions are mappings from variables to expressions.
	variables are coded as a declaration !!!!

	Note, that the meaning of this mapping is not necessarily a
	simultaneous substitution. This implies, that certain
	insertions of pairs do not mean, what they mean in normal
	substitutions. However, because in the formula representation
	all variables are distinct, this is no problem.

	1. substitution methods:
	------------------------

		these methods depend on a method 'apply-sub' for
		expressions.

		(defmethod (substitution :function-apply) (for)
		support for the analysis of beta redices (expressions of
		the form 'for = (x : G) H'. the method inserts the pair
		x -> H into the substitution.

		(defmethod (substitution :function-abstract) (for)
		undoes the effects of function-apply

		(defmethod (substitution :expand) ()
		transforms the substitution into a form, that no
		variable of the left side occurs on the right side of
		the substitutions.


     
   history     :

	90.08.07 base version 1
   errors      :
   bugs        :
|# 

(mod-require "data/map/v2/map")
(mod-require "logic/v2/expression/parse-gen/impls/parse-gen")

(defflavor substitution () (map parse-gen))

;beta-reduction
(defmethod (substitution :function-apply) (for)
    [:move-in-key [for :functor |:variable] [for :argument]]
)

(defmethod (substitution :function-abstract) (for)
    (let* ((var [for :functor |:variable])
	   (help 
	        (make-instance 
		    'substitution 
		    :init-value `((,var [:get var])))))
	[:iterate ((key info) in)
	    [:set-info in [info :apply-sub help]]
	]
	[:remove-key var]
    )
)

; elimination of substitutions of the form x -> y y -> z
(defmethod (substitution :expand) ()
    [:iterate ((key info) in)
	 [:set-info in [info :apply-sub self :copy t]]
    ]
    self
)

(defmethod (substitution :unparse)
    (&key (stream t) 
	  (env *exp-environment*))
    (when env [env :iterate (x) [*exp-names* :push-key x [x :picture]]])
    (format stream "(")
    (let ((need-blank nil))
       [:iterate ((x f))
	   (if need-blank
	       (format stream " ")
	       (setq need-blank t)
	   )
	   [x :unparse :stream stream]
	   (format stream " ")
	   [f :unparse :stream stream]
       ]
    )
    (format stream ")")
    (when env [env :iterate (x) [*exp-names* :pop]])
)
