#|
   module      : uni-main 
   author      : Olaf Mueller
   explanation : unification method for formulas
                 with K-Imitation for C and AC as rigid Determinations
                 
   history     : 29.3.93 create
   owner       : Olaf Mueller
                 5.4.93 all examples work
    
   errors      :
   bugs        : - Simplify is only invoked, if at least one rig-rig pair exist
                   in the dis-set. Nevertheless the flex-rig pairs are considered
                   in the simplification process, too. So: Are flex-rig pairs
                   necessary in Simplify? Or the other way around: Are even sets
                   with onlu flex-rig pairs necessary? 
   todo        : - minimality of solutions 
                   for this purpose simplify must immediatley follow imitate
                 - improvements concerning the number of search-nodes
                   a first step could be to produce new nodes after simplify only,
                   if more than one solution is returned.


|# 

; Modules to be imported
; ADT's
(mod-require "data/lis/v2/lis-list")
(mod-require "data/lis/v2/lis-list-readonly")
(mod-require "data/lis/v2/lis-lazy")
(mod-require "data/queue/v2/queue-list")
(mod-require "data/glo/v2/simple-aggregate-empty")

; logic modules
(mod-require "logic/v2/substitution/substitution-hash")
(mod-require "logic/v2/expression")

; Unification method for expression
(defmethod (expression-node :E-unify) 
  (e2 environment &key (substitutions (make-instance 'substitution-hash)) 
	               (patience 5000)
                  &allow-other-keys)
    (setq *uni-pic-count* 0)
    (setq *Neu-count* 0)     ; neue AC-Variablen
    (setq *NV-count* 0)      ; neue Lifting-Variablen
    (setq *count-I* 0)
    (setq *count-P* 0)
    (setq *count-KI* 0)
    (setq *count-S* 0)
    (setq *count-SS* 0)
    (setq *count-triv-S* 0)
    (setq *count-ho-S* 0)
    (setq *count-deadends* 0)
    (setq *count-nodes* 0)
    (setq *count-KI-apply* 0)
    (setq *count-S-apply* 0)
    (setq *count-comp* 0)
    (make-instance 'lis-lazy :init-value 
		 (uni-unify (uni-init-searchtree self e2 environment substitutions patience))
    )
)

(defmethod (expression-node :time-unify) (e2 env)
    (time  (progn (setq res [ :E-unify e2 env]) nil))
   (format t "~%~%--------------------------")
   (format t "~% Auswertung mit KI (Qian)~%")
   (format t "--------------------------~%~%")
   (format t "Anzahl Lsg.:          ~A~%" [res :count])
   (format t "Anzahl toter Aeste:   ~A~%" *count-deadends*)
   (format t "Knoten:          ~A~%" *count-nodes*)
   (format t "Imitationen:     ~A~%" *count-I*)
   (format t "K-Imitationen:   ~A~%" *count-KI*)
   (format t "Projektionen:    ~A~%" *count-P*)
   (format t "Vereinfachungen: ~A~%" *count-S*)
   (format t "K-Imitationen-real:   ~A~%" *count-KI-apply*)
   (format t "Vereinfachungen-real: ~A~%" *count-S-apply*)
   (format t "Triv. Vereinfachungen (Gr. 2): ~A~%In Prozent aller (erf.r.)  Vereinfachungen: ~A~%" *count-triv-S* (float (/ *count-triv-S* *count-S*)))
   (format t "Durchschnittliche Vereinf. Groesse: ~A~%" (float (/ *count-SS* *count-S*)))
   (format t "ho-Vereinfachungen: ~A~%" (float (/ *count-ho-S* *count-S*)))
   (format t "Komplexitaet/Knoten: ~A~%" (float (/ *count-comp* *count-nodes*)))
)


;(format t "~%Anzahl der durchlaufenen Knoten: ~A~%" count-nodes)
; Function traversing the searchtree. Output is a lisp-lazy-list of preunifiers
(defun uni-unify (tree)
  (do (node dis sub all-sub (pat [tree :patience] (- pat 1)))
      ([tree :is-empty] nil); While there are nodes not yet visited
      (if (= pat 0)(throw 'uni-patience nil))
      (setq node      [tree :pick-node]           ; Pick one of these nodes
	    dis       [node :disagreements]       ; And look at its
	    sub       [node :new-sub]             ; components
            all-subs  [node :all-subs])
      (setq *count-nodes* (1+ *count-nodes*))
      [all-subs :insert-aggregate sub] ; Substitution application that has been postponed
      ; Normalize all disagreemnent pairs
      ;(format t "dis: ~A~%" dis)
      [dis :iterate (x) [[[x :apply-sub sub] :beta-reduce] :toplevel-eta-expand]]
      [dis :count-complexity]
     ; [node :propagate-solved-pairs]
     ; [[node :copy] :print-node]
      (let* ((p [dis :get-rigid-rigid])
            (q-idx [dis :get-flexible-rigid-idx])
            (q (if q-idx [dis :get-elem q-idx] nil))
            (new-diss (make-instance 'lis-list))
            (subs (make-instance 'lis-list))
           )
        (if (or p q)
          (if p
              ; ---------- Vereinfachung -----------------------
              (progn
                 (setq new-diss [dis :simplify [tree :env]])
                 (if new-diss
                    (progn
                      (setq *count-S* (+ [new-diss :count] *count-S*))
                      (setq *count-S-apply* (1+ *count-S-apply*))
                      [tree :insert-new-nodes node new-diss subs]
                    )
                    (setq *count-deadends* (1+ *count-deadends*))
                 )
              )
              (progn
                 ; ----------- Projektion -------------------------
                 (setq new-diss (make-instance 'lis-list))
                 (setq subs [q :project [tree :env]])
                 (setq *count-P* (+ [subs :count] *count-P*))
                 [tree :insert-new-nodes node new-diss subs]
                 ; ----------- Imitation --------------------------
                 (if (typep [q :right |:head |:ref] 'c-declaration)
                    (progn
                        [dis :move-out `(at-index ,q-idx)]
                        (setq *count-KI* (1+ *count-KI*))
                        [tree :K-imitate-for-C node q]
                    )
                    (if (typep [q :right |:head |:ref] 'ac-declaration)
                        (progn
                            [dis :move-out `(at-index ,q-idx)]
                            (setq *count-KI* (+ 4 *count-KI*))
                            (setq *count-KI-apply* (1+ *count-KI-apply*))
                            [tree :K-imitate-for-AC node q]
                        )
                        (progn
                            (setq new-diss (make-instance 'lis-list))
                            (setq subs [q :imitate [tree :env]])
                            (setq *count-I* (+ [subs :count] *count-I*))
                            [tree :insert-new-nodes node new-diss subs]
                        )
                    )
                 )
              )   ; *progn*
          )
          (return (lazy-cons                                ; Create a lisp-lazy-list 
		     (let ((us (make-instance 'uni-sub)))   ; whose car is the first solution
		       [all-subs :iterate ((key info))      ; consisting of the accumulated subs
			    [us :move-in-key key info]]
		       [us :set-flex-flex dis]              ; and the remaining flex-flex pairs.
		       us)
		     (uni-unify tree) ; the cdr ist the promise to compute all solutions
                   )
          )
       ) ; *if*
      )  ;*let*    
   )     ;*do*
)


;--------------------------------------------------------------------------------
;         Substitutions with preunifier-components
;--------------------------------------------------------------------------------

;flex-flex is a disagreement-set containing only flexible-flexible-pairs
(defflavor uni-sub (flex-flex) (substitution-hash)
  (:gettable-instance-variables flex-flex)
  (:settable-instance-variables flex-flex)
  (:initable-instance-variables flex-flex))
  
;--------------------------------------------------------------------------------
;                   searchtree
;--------------------------------------------------------------------------------

; This ADT hides the representation and search strategy of Huet's unification
; search tree
; Additionally it hides the environment and patience

(defflavor searchtree (env patience) (queue-list) ; breadth first search strategy
  (:gettable-instance-variables env patience)
  (:settable-instance-variables env patience)
  (:initable-instance-variables env patience))

; Build an initial searchtree from two expressions, an environment and an initial substitution.
(defun uni-init-searchtree (e1 e2 en substitution patience)
  (let* ((left  e1)  
	 (right e2)
	 (pair  (make-instance 'disagreement-pair :left  left 
			                          :right right))
	 (set   (make-instance 'disagreement-set :init-value `(,pair)))
	 (mtsub (make-instance 'substitution-hash))
	 (node  (make-instance 'searchtree-node :disagreements set
		                                :new-sub       substitution
		                                :all-subs      mtsub
                                                :depth         0))
	 (tree  (make-instance 'searchtree :env en :patience patience)))
    [tree :push node]
    tree))

 
; Give the next searchtree-node to be visited
(defmethod (searchtree :pick-node)() [:dequeue])


; Build new searchtreenodes into the tree after both match and simplify
(defmethod (searchtree :insert-new-nodes)(node new-dissets subs)
  (if [new-dissets :is-empty]
    ; case match
    [subs :iterate (x) 
        (let ((new-node [node :copy]))
	  [new-node :set-new-sub x]
          [new-node :set-depth (1+ [node :depth])]
	  [:push new-node]
        )
    ]
    ; case simplify
    [new-dissets :iterate (x)
        (let ((new-node [node :copy]))
          [new-node :set-new-sub (make-instance 'substitution-hash)]
          [new-node :set-disagreements x]
          [new-node :set-depth (1+ [node :depth])]
          [:push new-node]
        )
    ]
  ) 
)


(defmethod (searchtree :K-imitate-for-C) (node q)
  (let* ((new-node [node :copy])
         (dis [new-node :disagreements])
         (sub-in-list [q :imitate [:env]])
         (sub [sub-in-list :move-out 'first])
         (pair [q :apply-sub sub | :beta-reduce])
         (l [pair :left])
         (r [pair :right])
        )
     (multiple-iterate ((a [l :arguments])   
	                (b [r :arguments]))
	      [dis :move-in (make-instance 'disagreement-pair 
                          :left  [(assemble-exp-binder [l :binder] a) :toplevel-eta-expand] 
			  :right [(assemble-exp-binder [r :binder] b) :toplevel-eta-expand])
              ]
     )
     ;(format t "dis: ~A~%" dis)
     [new-node :set-new-sub sub]
     [new-node :set-depth (1+ [node :depth])]
     [:push new-node]
  )
)

(defmethod (searchtree :K-imitate-for-AC) (node q)
  (let* ((sub-in-list [q :imitate [:env]])
         (sub [sub-in-list :move-out 'first])
         (new-node1 [node :copy])
         (new-node2 [node :copy])
         (new-node3 [node :copy])
         (new-node4 [node :copy])
         (dis1 [new-node1 :disagreements])
         (dis2 [new-node2 :disagreements])
         (dis3 [new-node3 :disagreements])
         (dis4 [new-node4 :disagreements])
         (pair [q :apply-sub sub | :beta-reduce])
         (l [pair :left])
         (r [pair :right])
         (arg1 [l :arguments])
         (arg2 [r :arguments])
         (b1   [l :binder])
         (b2   [r :binder])
         (x1 [arg1 :get-by-pos 1])
         (x2 [arg1 :get-by-pos 2])
         (y1 [arg2 :get-by-pos 1])
         (y2 [arg2 :get-by-pos 2])
         (ac-decl [q :left |:head |:ref])
         (var-type [ac-decl :type |:range |:domain])
         (NV1 [env :sub2-variable var-type "Neu"]) 
         (NV2 [env :sub2-variable var-type "Neu"])
         (NV3 [env :sub2-variable var-type "Neu"])
         (NV4 [env :sub2-variable var-type "Neu"])
        )
     (multiple-iterate ((a arg1)   
	                (b arg2))
	      [dis1 :move-in (make-instance 'disagreement-pair 
                          :left  [(assemble-exp-binder [l :binder] a) :toplevel-eta-expand] 
			  :right [(assemble-exp-binder [r :binder] b) :toplevel-eta-expand])
              ]
     )  
     [dis2 :move-in (make-instance 'disagreement-pair
               :left  (assemble-exp-binder b1 x1)
               :right (assemble-exp-binder b2 [ac-decl :appl-build NV2 y2]))]
     [dis2 :move-in (make-instance 'disagreement-pair
               :left (assemble-exp-binder b1 [ac-decl :appl-build x2 NV2])
               :right (assemble-exp-binder b2 y1))]
     [dis3 :move-in (make-instance 'disagreement-pair
               :left (assemble-exp-binder b1 x2)
               :right  (assemble-exp-binder b2 [ac-decl :appl-build NV4 y2]))]
     [dis3 :move-in (make-instance 'disagreement-pair
               :left  (assemble-exp-binder b1 [ac-decl :appl-build NV4 x1])
               :right  (assemble-exp-binder b2 y1))]
     [dis4 :move-in (make-instance 'disagreement-pair
               :left   (assemble-exp-binder b1 x1)
               :right  (assemble-exp-binder b2 [ac-decl :appl-build NV2 NV1]))]
     [dis4 :move-in (make-instance 'disagreement-pair
               :left   (assemble-exp-binder b1 x2)
               :right  (assemble-exp-binder b2 [ac-decl :appl-build NV4 NV3]))]
     [dis4 :move-in (make-instance 'disagreement-pair
               :left   (assemble-exp-binder b1 [ac-decl :appl-build NV4 NV2])
               :right  (assemble-exp-binder b2 y1))]
     [dis4 :move-in (make-instance 'disagreement-pair
               :left   (assemble-exp-binder b1 [ac-decl :appl-build NV3 NV1])
               :right  (assemble-exp-binder b2 y2))]
     [new-node1 :set-new-sub sub]
     [new-node2 :set-new-sub sub]
     [new-node3 :set-new-sub sub]
     [new-node4 :set-new-sub sub]
     [new-node1 :set-depth (1+ [node :depth])]
     [new-node2 :set-depth (1+ [node :depth])]
     [new-node3 :set-depth (1+ [node :depth])]
     [new-node4 :set-depth (1+ [node :depth])]
     [:push new-node1]
     [:push new-node2]
     [:push new-node3]
     [:push new-node4]
  )
)

  
;--------------------------------------------------------------------------------
;              searchtree-node
;--------------------------------------------------------------------------------

; simple structure  storing a disagreement-set, 
; a substitution to be applied to the disagreements
; and a substitution caching the substitutions found so far

(defflavor searchtree-node (disagreements new-sub all-subs depth)(simple-aggregate-empty simple-aggregate)
  :settable-instance-variables
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (searchtree-node :print-node) ()
   (format t "~%~%Knoten: ~A~%" *count-nodes*)
   [disagreements :iterate (pair)
       [pair :left |:unparse]
       (format t " =?= ")
       [pair :right |:unparse]
       (format t" , ")
   ]
   (format t "Tiefe: ~A" depth)
   (print-real-sub new-sub)
)

; auxiliary function for generic implementation of :print-self
(defmethod (searchtree-node :attribute-lis)()
  (make-instance 'lis-list :init-value `(,disagreements ,new-sub ,all-subs)))

; A :copy of a searchtree-node is really only a copy of new-sub and all-subs
(defmethod (searchtree-node :copy)()
  (make-instance 'searchtree-node :disagreements [disagreements :copy]
		                  :new-sub       [new-sub :copy]
				  :all-subs      [all-subs :copy]
                                  :depth         depth))


; entdeckt und propagiert sowohl flex-rig als auch flex-flex Paare in gel"oster Form.
; flex-rig enstehen oft nach Simplify, flex-flex nach Imitation
(defmethod (searchtree-node :propagate-solved-pairs) ()
   (let* ((dis      [:disagreements])
          (all-subs [:all-subs])
         )
     [dis :iterate (pair idx)
       (if (and (not (eq [pair :get-type] :rigid-rigid))
                [pair :is-solved-formed]
           )
         (progn
           [dis :move-out `(at-index ,idx)]
           (let* ((sub (make-instance 'substitution-hash :init-value `((
                         ,[pair :left |:head |:ref]
                         ,[pair :right]))))
                 )
               [dis :iterate (p) [p :apply-sub sub |:beta-reduce]]
               [all-subs :insert-aggregate sub]
           )
         )
       )
     ]
   )
)

;--------------------------------------------------------------------------------
;                    disagreement-pairs
;--------------------------------------------------------------------------------

; Pair of uni-exp's that has to be unified
(defflavor disagreement-pair (left right) (simple-aggregate-empty simple-aggregate)
  (:gettable-instance-variables left right)  
  (:settable-instance-variables left right)
  (:initable-instance-variables left right))

; auxiliary function for generic implementation of :print-self
(defmethod (disagreement-pair :attribute-lis)()
  (make-instance 'lis-list :init-value `(,left ,right)))

; extend some expression-methods to both components
(defmethod (disagreement-pair :apply-sub) (sub)
  (setq left  [left  :apply-sub sub])
  (setq right [right :apply-sub sub])
  self)

(defmethod (disagreement-pair :copy)()
  (make-instance 'disagreement-pair :left  left
		                    :right right))

(defmethod (disagreement-pair :beta-reduce) ()
  (setq left  [left  :beta-reduce])
  (setq right [right :beta-reduce])
  self)

(defmethod (disagreement-pair :toplevel-eta-expand) ()
  (setq left  [left  :toplevel-eta-expand])
  (setq right [right :toplevel-eta-expand])
  self)


; nur fuer rigid-flex und flex-flex
; flex-flex wird - wenn in solved-form - so umgestellt, dass Variable links
(defmethod (disagreement-pair :is-solved-formed) ()
   (or (typep left 'reference)     ; eig. eta-Kurzform pruefen, z.B. durch 
                                   ; (eq? [left :head |:toplevel-eta-expand] left)
       (and (eq [:get-type] :flexible-flexible)
            (progn
               (psetq left right
                      right left)
               (typep left 'reference)    ; s.o. + occur-check!!!
            )
       )
   )
)

; A component of a disagreement-pair is flexible iff it's head is a free variable
; else it is called rigid.
; Therefore there are four types of disagreement-pairs:
; rigid-rigid
; (rigid-flexible)
; flexible-rigid
; flexible-flexible
; If :get-type finds a rigid-flexible pair it normalizes it to
; a flexible-rigid pair and returns flexible-rigid
; both expression has to be beta-normalized
(defmethod (disagreement-pair :get-type)()
  (if [left :is-rigid]
      (if [right :is-rigid]
          :rigid-rigid
	  (progn (psetq left  right
			right left)
		 :flexible-rigid))
      (if [right :is-rigid]
	  :flexible-rigid
	  :flexible-flexible)))

; project returns of a set of substitutions determined by projection
; The env argument specifies an environment where
; new variables are made public
(defmethod (disagreement-pair :project) (env)
  (let* ((subs (make-instance 'set-lis)) ; Substitution to be computed. 
	 (e1 [:left])                    ; bind some parts of the pair
	 (e2 [:right])
	 (a1 [e1 :arguments])
	 (a2 [e2 :arguments])
	 (h1 [e1 :head])
	 (h2 [e2 :head])
	 (beta [[e1 :expression-type] :output-type]) ; Result type of e1 and e2
	 ; some easy parts of the output substitutions
         ; binder for all produced substitutions
	 (binder  [a1 :map :iteration-order 'from-last (x) (uni-new-abstraction x)])
         ; remember: binders are stored in reverse order!
         (rev-binder [binder :reduce (x) 
                             (p (make-instance 'lis-list-readonly))
                             [p :cons x]])
         ; Arguments for new Variables
	 (vars    [binder :map :iteration-order 'from-last (x) (make-instance 'reference :ref x)])
         ; binder is inverse, not so arguments.
        )
       ;----------project--------------------------------------------------
    ; inspect all arguments of e1 for fitting type
    (multiple-iterate ((e1i a1)
                       (z rev-binder))
      (if [beta :equal [[e1i :expression-type] :output-type]]
	  (let ((itypes [[e1i :expression-type] :argument-types]))
	    [subs :move-in 
	          (make-instance 'substitution-hash :init-value `(( 
		     ,[h1 :ref]
		     ,(assemble-exp
		        binder
			(make-instance 'reference :ref z)
  		        [itypes :map (x) 
		          (assemble-exp-head 
                             [env :new-variable 
                                  (assemble-type 
                                     x  
                                     [h1 :expression-type |:argument-types]
                                  )
                                  "I"
                             ]
			     vars
                          )
                        ]
                      )
                   )))
            ]
          )
       )
    )
    subs)) 


; imitate returns all substitutions found by imitation
; The env argument specifies an environment where
; new variables are made public
(defmethod (disagreement-pair :imitate) (env)
  (let* ((subs (make-instance 'set-lis)) ; Substitution to be computed. 
	 (e1 [:left])                    ; bind some parts of the pair
	 (e2 [:right])
	 (a1 [e1 :arguments])
	 (a2 [e2 :arguments])
	 (h1 [e1 :head])
	 (h2 [e2 :head])
	 (beta [[e1 :expression-type] :output-type]) ; Result type of e1 and e2
	 ; some easy parts of the output substitutions
         ; binder for all produced substitutions
	 (binder  [a1 :map :iteration-order 'from-last (x) (uni-new-abstraction x)])
         ; remember: binders are stored in reverse order!
         (rev-binder [binder :reduce (x) 
                             (p (make-instance 'lis-list-readonly))
                             [p :cons x]])
         ; Arguments for new Variables
	 (vars    [binder :map :iteration-order 'from-last (x) (make-instance 'reference :ref x)])
         ; binder is inverse, not so arguments.
         (decl2 [h2 :ref])
         (head-list (make-instance 'lis-list))
        )
    ;----------imitate--------------------------------------------------
     [[decl2 :imitation-decls env] :iterate (decl-i)
            [head-list :move-in (make-instance 'reference :ref decl-i)]
     ]
     [head-list :iterate (hi)
       (if (not [[e2 :binder] :find [hi :ref]]) ; is this check necessary for all hi??
	  [subs :move-in 
	        (make-instance 'substitution-hash :init-value `((
	            ,[h1 :ref]
		    ,(assemble-exp
		       binder
		       hi
		       [a2 :map (x) 
		           (assemble-exp-head 
                              [env :new-variable 
                                   (assemble-type 
                                       [x :expression-type] 
                                       [h1 :expression-type |:argument-types]
                                   )
	                           "G"
                              ]
			      vars
                            )
                       ]
                     )
                  )))
           ]
       )
     ]
     subs
  )
) 

;------------------------------------------------------------------------------
;              disagreement-sets
;------------------------------------------------------------------------------

; set of disagreement pairs
; the following line is efficient but wrong
(defflavor disagreement-set () (lis-list))
;(defflavor disagreement-set () (set-list))

; copy a disagreement set keeping the pairs untouched
(defmethod (disagreement-set :copy)()
  (let ((c (make-instance 'disagreement-set)))
    [:iterate (x) [c :move-in [x :copy]]]
    c))
 

; Apply Substitution sub to each element of self
(defmethod (disagreement-set :apply-sub)(sub)
  [:iterate (el) 
      [el :apply-sub sub] 
      ; Note: apply-sub for pairs is modifying
  ]
)

(defmethod (disagreement-set :count-complexity) ()
 (let ((set  [ :reduce (pair) (h 0)
               (+ h [pair :right |:count-leafs] 
                    [pair :left |:count-leafs])
             ])
      )
  (setq *count-comp* (+ set *count-comp*))
 )
)

; Remove one pair of the specified type
; Heuristics for choosing pairs intelligently are hidden here
(defmethod (disagreement-set :pick-rigid-rigid)()
  (let ((x [:find-property (el) (eq [el :get-type] :rigid-rigid)]))
    (if x [:move-out `(at-index ,x)] nil))) 
(defmethod (disagreement-set :get-rigid-rigid)()
  (let ((x [:find-property (el) (eq [el :get-type] :rigid-rigid)]))
    (if x [:get-elem x] nil))) 
(defmethod (disagreement-set :pick-flexible-rigid)()
  (let ((x [:find-property (el) (eq [el :get-type] :flexible-rigid)]))
    (if x [:move-out `(at-index ,x)] nil)))
; same as pick but don't remove the pair
(defmethod (disagreement-set :get-flexible-rigid)()
  (let ((x [:find-property (el) (eq [el :get-type] :flexible-rigid)]))
    (if x [:get-elem x] nil)))
(defmethod (disagreement-set :pick-flexible-flexible)()
  (let ((x [:find-property (el) (eq [el :get-type] :flexible-flexible)]))
    (if x [:move-out `(at-index ,x)] nil)))

(defmethod (disagreement-set :get-flexible-rigid-idx)()
  (let ((x [:find-property (el) (eq [el :get-type] :flexible-rigid)]))
    x))

; --------------------------------------------------------------------
;      Theory-dependent table of suitable declarations for imitation
; --------------------------------------------------------------------

; returns all declarations that are compatible to self in a list
; in order to use references to them as possible heads in the imitation rule.
; In normal declarations only self is able to produce a suitable head for
; imitation. For all root-preserving operators (C, AC, ...) thats the same.
; Therefore in this cases this method need not to be overwritten. 
(defmethod (declaration :imitation-decls ) (env)
  (let  ((result-list (make-instance 'lis-list))
        )
    [result-list :move-in self]
    result-list
  )
)
    


; --------------------------------------------------------------------
;                      Output - Utilities
; --------------------------------------------------------------------

(defun print-sub (s)

    [s :iterate ((var exp))
        (format t "~a  ->  " [var :picture])
        [exp :beta-eta-reduce |:unparse]
        (format t "~%")
    ]
    (format t "- flex~%")
    [s :flex-flex |:iterate (pair)
        [pair :left  |:unparse]
        (format t "  ->  ")
        [pair :right |:beta-eta-reduce |:unparse]
        (format t "~%")
 
    ]
)


(defun print-real-sub (s)
    (if (not [s :is-empty]) (format t "~%"))
    [s :iterate ((var exp))
        (format t "~a  ->  " [var :picture])
        [exp :beta-eta-reduce |:unparse]
        (format t " , ")
    ]
)

(defun print-subs (l)
    [l :iterate (x)
        (format t "----~%")
        (print-sub x)

    ]
)

(defun print-expand (s sub)

    [s :iterate ((var exp))
        (format t "~a  ->  " [var :picture])
        [exp :apply-sub sub |:beta-eta-reduce |:unparse]
        (format t "~%")
    ]
    (format t "- flex~%")
    [s :flex-flex |:iterate (pair)
        [pair :left  |:unparse]
        (format t "  ->  ")
        [pair :right |:apply-sub sub |:beta-eta-reduce |:unparse]
        (format t "~%")
 
    ]
)

(defun print-expand-s (l)
    [l :iterate (x)
        (format t "----~%")
        (print-expand x x)

    ]
)

