#|
   module      : f-o-dependence
   author      : Olaf Mueller
   explanation : detection and elimanation of dependent substitutions
                 in the first-order-case
                
   history     : 17.3.93 create
   owner       : Olaf Mueller
                 1.4.93 all examples work
    
   errors      :
   bugs        : - Muessen Substitutionen vorher nicht entflochten werden? (expand)
                 - Ist = von Subs wirklich modulo Variablenumbenenung zu sehen?
   todo        :


|#
; ---------------------------------------------------------------------------------------
;                                      Methode less-equal
; ---------------------------------------------------------------------------------------

(defmethod (substitution-hash :less-equal) (rho env)
  (let ((sigma [:expand self]) 
        (rho   [rho :expand rho])
        (dis-set (make-instance ' disagreement-set))
        (sub (make-instance 'substitution-hash))
       )
     [sigma :iterate ((var exp))
        (let ((varx (make-instance 'reference :ref var)))
           [dis-set :move-in (make-instance 'disagreement-pair 
                 :left exp
                 :right [varx :apply-sub rho | :free-vars-to-new-cons sub])
           ]
        )
     ]
     ;(format t "dis-set: ~A" dis-set)
     (if [dis-set :poormans-AC (make-instance 'substitution-hash) env] t nil)
  )
)

(virtualmethod (expression-node :collect-free-variables) (var-list) )

(defmethod (application :collect-free-variables) (var-list)
  (let ((list1 [functor  :collect-free-variables [var-list :copy]])
        (list2 [argument :collect-free-variables [var-list :copy]]) 
       )
     [list1 :iterate (var)
        (if (not [list2 :find var])
            [list2 :move-in var]
        ) 
     ]
     list2
  )
)

(defmethod (abstraction :collect-free-variables) (var-list)
  [expression-node :collect-free-variables [var-list :copy]]
)

(defmethod (reference :collect-free-variables) (var-list)
  (if (not [ :ref | :is-constant]) [var-list :move-in [self :ref]])
  var-list
)

(defmethod (expression-node :free-vars-to-new-cons) (sub) 
  (let ((frees [ :apply-sub sub |:collect-free-variables (make-instance 'lis-list)])
       )
     [frees :iterate (var) 
        (let ((cons (make-instance 'declaration 
                           :type [var :type]
                           :picture (format nil "~A~D" "cons-" [var :picture])
                           :is-constant t))
             )
           [sub :move-in-key var (make-instance 'reference :ref cons)]
        )
     ]
     [ :apply-sub sub]
  )
)

(defmethod (lis-list :eliminate-redundant-subs) (env)
    [ :iterate (sub1 idx1)
       [ :iterate (sub2 idx2)
           (if (AND (not (eq idx1 idx2))
                    [sub2 :less-equal sub1 env])
              [ :move-out `(at-index ,idx1)]
           )
       ]
       ; (format t "Rest-subs: ~%")
       ; (print-subsx self)  Ausgabe zu Testzwecken
    ]
    self
)


(defmethod (substitution-hash :expand) (sub)
   (let ((res (make-instance 'substitution-hash))
         )
      [ :iterate ((var exp))
         [res :move-in-key var [exp :apply-sub sub]]
      ]
      res
   )
)


#|
Ausgabefunktionen zu Testzwecken:

(defun print-subx (s)

    [s :iterate ((var exp))
        (format t "~a  ->  " [var :picture])
        [exp :beta-eta-reduce |:unparse]
        (format t "~%")
    ]
)

(defun print-subsx (l)
    [l :iterate (x)
        (format t "----~%")
        (print-subx x)

    ]
)

|#

