#|
   module      : h-o-dependence
   author      : Olaf Mueller
   explanation : detection and of dependent substitutions
                 in the higher-order-case
                
   history     : 17.3.93 create
   owner       : Olaf Mueller
                 1.4.93 all examples work
    
   errors      :
   bugs        : 
   todo        :


|#

; ---------------------------------------------------------------------------------------
;                                 (h-o) - Unifikation fuer Mengen
; ---------------------------------------------------------------------------------------

; Unification method for disagreement-sets
(defmethod (disagreement-set :set-unify) 
     (environment &key (substitutions (make-instance 'substitution-hash)) 
	               (patience 500)
                  &allow-other-keys)
  (setq *count-nodes* 0)
  (setq *uni-pic-count* 0)
  (let* ((mtsub (make-instance 'substitution-hash))
	 (node  (make-instance 'searchtree-node :disagreements self
		                                :new-sub       substitutions
		                                :all-subs      mtsub
                                                :depth         0))
	 (tree  (make-instance 'searchtree :env environment :patience patience))
        )
    [tree :push node]
    (make-instance 'lis-lazy :init-value (uni-unify tree))
  )
)
; ---------------------------------------------------------------------------------------
;                                      Methode h-o-less-equal
; ---------------------------------------------------------------------------------------

(defmethod (substitution-hash :h-o-less-equal) (rho env)
  (let ((sigma self)  ; expand??@@@
        (dis-set (make-instance ' disagreement-set))
        (sub (make-instance 'substitution-hash))
       )
     [sigma :iterate ((var exp))
        (let ((varx (make-instance 'reference :ref var)))
           [dis-set :move-in (make-instance 'disagreement-pair 
                 :left exp
                 :right [varx :apply-sub rho | :free-vars-to-new-cons sub])
           ]
        )
     ]
     (if [[dis-set :set-unify env] :get-by-pos '1] t nil)
  )
)

#|
Gibts es schon fuer f-o-dependence !!!

(virtualmethod (expression-node :collect-free-variables) (var-list) )

(defmethod (application :collect-free-variables) (var-list)
  (let ((list1 [functor  :collect-free-variables [var-list :copy]])
        (list2 [argument :collect-free-variables [var-list :copy]]) 
       )
     [list1 :iterate (var)
        (if (not [list2 :find var])
            [list2 :move-in var]
        ) 
     ]
     list2
  )
)

(defmethod (abstraction :collect-free-variables) (var-list)
  [expression-node :collect-free-variables [var-list :copy]]
)

(defmethod (reference :collect-free-variables) (var-list)
  (if (not [ :ref | :is-constant]) [var-list :move-in [self :ref]])
  var-list
)

(defmethod (expression-node :free-vars-to-new-cons) (sub) 
  (let ((frees [ :apply-sub sub |:collect-free-variables (make-instance 'lis-list)])
       )
     [frees :iterate (var) 
        (let ((cons (make-instance 'declaration 
                           :type [var :type]
                           :picture (format nil "~A~D" "cons-" [var :picture])
                           :is-constant t))
             )
           [sub :move-in-key var (make-instance 'reference :ref cons)]
        )
     ]
     [ :apply-sub sub]
  )
)
|#


(defmethod (uni-sub :expand) (sub)
   (let ((res (make-instance 'uni-sub))
         )
      [ :iterate ((var exp))
         [res :move-in-key var [exp :apply-sub sub]]
      ]
      [res :set-flex-flex flex-flex]
      res
   )
)

(defmethod (uni-sub :project-help-variables) (exp1 exp2)
 (let* ((res (make-instance 'uni-sub))
       (exp (make-instance 'application 
                   :functor exp1
                   :argument exp2))
       (var-list [exp :collect-free-variables (make-instance 'lis-list)])
      )
   [ :iterate ((var exp))
      (if [var-list :find var]
          [res :move-in-key var exp]
      )
   ]
   [res :set-flex-flex flex-flex]
   res
 )
)


(defmethod (lis-list :eliminate-redundant-ho-subs) (env)
    [ :iterate (sub1 idx1)
       [ :iterate (sub2 idx2)
           (if (AND (not (eq idx1 idx2))
                    [sub2 :h-o-less-equal sub1 env])
              [ :move-out `(at-index ,idx1)]
           )
       ]
       ; (format t "Rest-subs: ~%")
       ; (print-subsx self)  Ausgabe zu Testzwecken
    ]
    self
)


; Eliminiert das Ergebnis von :E-unify um abhaengige ho-subs, indem vorher alle
; Hilfsvariablen wegprojeziert wurden. Geht aber nur bei richtigen Substitutionen,
; meist nciht, wenn flex-flex Paare da sind.

(defmethod (lis-lazy :optimize) (env exp1 exp2)
   (let ((list (make-instance 'lis-list))
        )
     [ :iterate (uni-sub) 
        [list :move-in [uni-sub :expand uni-sub |:project-help-variables exp1 exp2]]
     ]
     [list :eliminate-redundant-ho-subs env]
     list
   )
)

#| 
Problem bei h-o-less-equal:

Trotz :expand und :project-help-variables lassen sich Abhaengigkeiten
haeufig nicht ermitteln, weil Hilfsvariablen in den flex-flex Paaren
stecken. Richtig funktioniert es nur fuer richtige Substitutionen!!!
Bsp: Franz Bsp V(x)^W(x)=?=c(1)^c(y), wo ueberall solche flex-flex-Paare 
     dastehen, die suggerieren, dass alle Lsg. abhaengig seien!!

|#
