#|
   module      : uni-ac-simplify  
   author      : Olaf Mueller
   explanation : simplification rule for unification
                 by Qian/Wang; for c and ac symbols also
                
   history     : 14.1.93 create
   owner       : Olaf Mueller
                 11.3.93 all examples work
    
   errors      :
   bugs        : 
   todo        : - same H for same alien-subterm should be searched; until now the occur-
                   check is senseless!
                 


|#

; ---------------------------------------------------------------------------------------
;                                      Methode Simplify
; ---------------------------------------------------------------------------------------

(defmethod (disagreement-set :print-simp) ()
 (let ((mfus  [*PMAS* :reduce (pair) (h 0)
                (+ h [pair :right |:count-leafs] 
                     [pair :left |:count-leafs])
              ])
       (set  [ :reduce (pair) (h 0)
               (+ h [pair :right |:count-leafs] 
                    [pair :left |:count-leafs])
             ])
      )
  (format t "~%Vereinfachung:~%")
  (format t "Anzahl der mfu's: ~A~%" [*PMAS* :count])
  (format t "Einzelne Groessen: ")
  [*PMAS* :iterate (pair)
    (format t "~A, " [pair :right |:count-leafs])
  ]
  (format t "~%Anzahl der uebergebenen Paare: ~A~%" [:count])
  (format t "Groesse dieses Sets: ~A" set)
  (if (< set 3)  (setq *count-triv-S* (1+ *count-triv-S*)))
  (setq *count-SS* (+ set *count-SS*))
  (if (> mfus set) (setq *count-ho-S* (1+ *count-ho-S*)))
 )
)


(virtualmethod (expression-node :count-leafs) () )

(defmethod (application :count-leafs) () 
   (+ [functor :count-leafs] [argument :count-leafs])
)

(defmethod (abstraction :count-leafs) () 
   (1+ [expression-node :count-leafs])
)

(defmethod (reference :count-leafs) () 1)


(defmethod (disagreement-set :simplify) (env)
(block simplify
  (let ((S1 self)
        (S2 (make-instance 'disagreement-set))    ; dis-set fuer erste Stufe
        (S-print (make-instance 'disagreement-set))    ; S2-copy fuer Ausgabe
        (sigma-list (make-instance 'lis-list))    ; Liste von Unifikatoren erster Stufe
        (res (make-instance 'lis-list))           ; Liste der neuen dis-sets
        (H-env (make-instance 'environment-lis))
        (l-env (make-instance 'environment-lis))
       )
     (setq *exp-identity* (make-instance 'key-stack-list))
     (setq *pmas* (make-instance 'lis-list))
     (setq S2 [S1 :abstract H-env]) 
     ;(format t "S2: ~A~%" S2)
     (setq *H-contexts* (make-instance 'key-stack-list))
     (setq *lambda-list* (make-instance 'lambda-list))
     (setq *converter* (make-instance 'converter))
     [S2 :translate l-env]
     (setq S-print [S2 :copy])
     ;(format t "S2: ~A~%" S2)
     (setq sigma-list [S2 :poormans-AC (make-instance 'substitution-hash) l-env])
     ;(format t "sigs: ~A~%" sigma-list)
     (setq *organizer* (make-instance 'new-variable-organizer))  ; @@or sigma-dependent?
     (if (not sigma-list) (return-from simplify nil))  ; no unifiers at all
     [S-print :print-simp]
     ;(format t "sigs-before: ~A~%" sigma-list)
     (if (> [sigma-list :count] 0) [sigma-list :eliminate-redundant-subs env])
     ;(format t "sigs-after: ~A~%" sigma-list)
     [sigma-list :iterate (sigma)
        (let ((S-hilf (make-instance 'disagreement-set))
             )
          [*pmas* :iterate (pair)
            (let* ((H         [pair :left])
                   (H-decl    [H :ref])
                   (binder    [*H-contexts* :get H-decl :test 'eql])
                   (abs-list  (make-instance 'lis-list))
                   (Y         (make-instance 'algebraic-reference 
                                  :ref [*converter* :convert-variable H-decl l-env]))
                   (Y-sub     [Y :apply-sub sigma])
                   (Y-sub-re  (make-instance 'expression-node))
                   (erg-term  (make-instance 'expression-node))
                  )
              [binder :iterate (x) 
                (let ((abs (make-instance 'abstraction 
                                 :type [x :type]
                                 :picture [x :picture]
                                 :is-constant t))
                     )
                  [*exp-identity* :push-key x abs]
                  [abs-list :move-in abs :to-pos 'last]
                )
              ]
              (setq Y-sub-re  [Y-sub :retranslate binder env])
              ;(format t "Y-sub-re: ~A~%" Y-sub-re)
              (setq erg-term 
                 [abs-list :reduce :iteration-order 'from-last (x) (p Y-sub-re)
                        [x :set-expression-node p]
                 ]
              )     
              [abs-list :iterate (x) [*exp-identity* :pop]]  
              [S-hilf :move-in (make-instance 'disagreement-pair
                                     :left  erg-term
                                     :right [pair :right])
              ]
            )  
          ] 
          [S1 :iterate (pair)              ; alle flex-flex Paare von S1 hinzunehmen
               [S-hilf :move-in pair]
          ]   
          [res :move-in S-hilf]
        )
     ]
    ; (format t "~%Zeit:   ")
     res   
  )
)
)


; ---------------------------------------------------------------------------------------
;                                      Methode Abstract
; ---------------------------------------------------------------------------------------

(defmethod (disagreement-set :abstract) (H-env)
  (let ((S1 self)
        (S2 (make-instance 'disagreement-set)) 
       )     
    (setq *Hi-count* 0)    ; Zaehler der Hi innerhalb von ganz Simplify  
    [S1 :iterate (pair idx)
       (if (eq [pair :get-type] :rigid-rigid)
              ; (eq [pair :get-type] :flexible-rigid)
          ; )    
         (progn
           [S1 :move-out `(at-index ,idx)]
           [S2 :move-in [pair :replace-alien-subterms H-env]]
         )
       )
     ]
     S2
  )
)

; --------------------------------------------------------------------------
;                          Methode replace-alien-subterms
; --------------------------------------------------------------------------
; method is only allowed for beta-reduced terms

(defmethod (disagreement-pair :replace-alien-subterms) (env)
  (setq left  [left  :replace-alien-subterms (make-instance 'lis-list) env])
  (setq right [right :replace-alien-subterms (make-instance 'lis-list) env])
  self
)

(virtualmethod (expression-node :replace-alien-subterms) (abs-list env))

(defmethod (application :replace-alien-subterms) (abs-list env)
  (let* ((fun    functor)
        )
     (if (not [head :ref |:is-constant])  ; only beta-reduced terms allowed!
       (let*  ((type-list [abs-list :map (x) [x :type]])
               (type [type-list :reduce (x) (p [ :expression-type])
                          (make-instance 'function-type
                               :domain x
                               :range p)])
               (H [env :sub-variable type "H"])
              )      
            [*pmas* :move-in (make-instance 'disagreement-pair 
                         :left  H 
                         :right (assemble-exp-binder abs-list self)) :to-pos 'last
            ]  ; therefore correct order.
            [abs-list :reduce :iteration-order 'from-last (x) (p H) 
                   (make-instance 'application
                           :functor p
                           :argument (make-instance 'reference 
                                            :ref [*exp-identity* :get x :test 'eql]))
            ]
         )
         (make-instance (type-of self)                       
                        :functor  [fun       :replace-alien-subterms abs-list env]
                        :argument [argument  :replace-alien-subterms abs-list env])
     )
  )
)

(defmethod (abstraction :replace-alien-subterms) (abs-list env)
  (let ((res (make-instance (type-of self) 
                        :picture picture
                        :type type))
       )
     [abs-list :move-in self]              ; reverse order: directly fit for binder
     [*exp-identity* :push-key self res]
     [res :set-expression-node [ :expression-node | :replace-alien-subterms abs-list env]]
     [*exp-identity* :pop]
     [abs-list :move-out 'first]
     res
  )
)

(defmethod (reference :replace-alien-subterms) (abs-list env)
  (if (not [ :ref | :is-constant])
    (let* ((type-list [abs-list :map (x) [x :type]])
           (type [type-list :reduce (x) (p [ :expression-type])
                          (make-instance 'function-type
                               :domain x
                               :range p)])
           (H [env :sub-variable type "H"])
          )
       [*pmas* :move-in (make-instance 'disagreement-pair 
             :left  H 
             :right (assemble-exp-binder abs-list self)) :to-pos 'last
       ]  ; therefore correct order.
       [abs-list :reduce :iteration-order 'from-last (x) (p H) 
              (make-instance 'application
                      :functor p
                      :argument (make-instance 'reference 
                                       :ref [*exp-identity* :get x :test 'eql]))
       ]
    )
    (let ((bound [*exp-identity* :get [ :ref] :test 'eql]))
       (if bound
          (make-instance 'reference :ref bound)
          [ :copy]
       )
    )
  )
)

; --------------------------------------------------------------------------------
;                         Hilfsprozeduren zur Abstraktion
; --------------------------------------------------------------------------------

; create a declaration with a unique picture beginning with base,
; insert it into the environment and return a reference to it
(defmethod (environment :sub-variable) (type-node base)
  (let ((dec (make-instance 'declaration              ;note the similarity to :declare 
			    :type type-node
			    :picture (new-pic base)
			    :is-constant nil)))
    [:move-in dec]
    (make-instance 'reference :ref dec)))

(setq *Hi-count* 0) 
(defun new-pic (base)
  (setq *Hi-count* (1+ *Hi-count*))
  (format nil "~A~D" base *Hi-count*))


;------------------------------------------------------------------------------
;              Datentyp lambda-list
; -----------------------------------------------------------------------------

(defflavor const-declaration (var) (declaration)
                (:settable-instance-variables var) )

(defflavor lambda-declaration (cons next) (declaration) 
                (:settable-instance-variables cons next) )

(defflavor lambda-list (list top depth) ()
        (:gettable-instance-variables list top depth)
        (:settable-instance-variables list top depth)
        (:initable-instance-variables list top depth) )

(defmethod (lambda-list :new-lambda) (Te S env)
  (incf depth)
  ;(format t "S: ~A~%" S)
  (let* ((T-new [Te :translate])
         (TtoS (make-instance 'function-type 
                     :domain Te
                     :range  S))
         (type (make-instance 'function-type 
                     :domain [S :translate]
                     :range  [TtoS :translate]))
        )
  (if (equal top [list :last])              ; if at the end of list
    (let ((lambda (make-instance 'lambda-declaration           
			    :type type
			    :picture (format nil "~A~D" "lambdax" depth)
			    :is-constant t))
          (varx (make-instance 'declaration           
			    :type Te
			    :picture (format nil "~A~D" "x" depth)
			    :is-constant nil))
          (cons   [env :introduce-constant T-new depth])
         )                   ; create new lambda-declaration
       [cons :set-var varx]
       [lambda :set-cons cons]
       [lambda :set-next nil]
       [ :set-top [list :move-in lambda :to-pos 'last]] ; move it in list
       lambda                               ; and return it
    )
    (let ((lambda nil)                       ; if not at the end of list
         )
       (if top                               ; increase top
          [ :set-top [list :next top]]
          [ :set-top [list :first]]
       )
       (setq lambda [list :get-elem top])    ; get (next) lambda-declaration
       (do () 
           ((or (not [lambda :next])        ; if type disagree + next lambda exists
                (equal [lambda :type] type)))
         (setq lambda [lambda :next])        ; get next lambda
       )
       (if [[lambda :type] :compare type]    ; if type correct. only compare, not equal!!
         lambda                              ; return lambda
                                             ; else no next lambda exist
                                             ; therefore create new lambda
         (let ((next-lambda  (make-instance 'lambda-declaration           
			    :type type
			    :picture (format nil "~A~D" "lambdax" depth)
			    :is-constant t))
               (varx (make-instance 'declaration           
			    :type Te
			    :picture (format nil "~A~D" "x" depth)
			    :is-constant nil))
               (next-cons   [env :introduce-constant T-new depth])
              )
           [next-cons :set-var varx]
           [next-lambda :set-cons next-cons]    
           [lambda :set-next next-lambda]     ; kitt it to the old
           next-lambda                        ; and return it
         )
       )   
    )   ; *let*
  )     ; *if*
  )
)

(defmethod (lambda-list :dispose-lambda) ()
  (decf depth)
  [ :set-top [list :previous top]]
)

(setq *lambda-list* (make-instance 'lambda-list))

(definitmethod (lambda-list) ()
   (setq list (make-instance 'lis-list))
   (setq top nil)
   (setq depth 0)
)

(defmethod (environment :introduce-constant) (type counter)
  (let ((dec (make-instance 'const-declaration           
			    :type type
			    :picture (format nil "~A~D" "consx" counter)
			    :is-constant t)))
    [:move-in dec]  
    dec
  )
)

;------------------------------------------------------------------------------
;              Datentyp converter
; -----------------------------------------------------------------------------

(defflavor converter (to-first to-higher) ()
        (:gettable-instance-variables to-first to-higher)
        (:settable-instance-variables to-first to-higher)
        (:initable-instance-variables to-first to-higher) )

(setq *converter* (make-instance 'converter))

(definitmethod (converter) ()
   (setq to-first  (make-instance 'key-stack-list))
   (setq to-higher (make-instance 'key-stack-list))
)
; here to-higher should be initialized with all constants in the theory, 
; they could be generated by the first-order-precess!

(defmethod (converter :convert-constant) (c l-env)
  (let ((c_exists [to-first :get c :test 'eql]))
    (if c_exists
       c_exists
       (let* ((arg-types [c :type | :argument-types])
              (in  [arg-types :map (type) [type :translate]])
              (out [c :type | :output-type])
              (type (assemble-type out in))
              (new_c  (make-instance (type-of c) 
                           :type type
                           :picture [c :picture]
                           :is-constant t))
             )
         [l-env :move-in new_c]
         [to-first  :push-key c     new_c]
         [to-higher :push-key new_c c    ]
         new_c
       )
    )
  )
)

(defmethod (converter :reconvert-constant) (c)
  [to-higher :get c :test 'eql]
)

(defmethod (converter :convert-variable) (H l-env)
  (let ((Y_exists [to-first :get H :test 'eql]))
    (if Y_exists
       Y_exists
       (let* ((out [H :type | :output-type])
              (Y (make-instance 'declaration 
                      :type out   ; exactly here the argument-types get lost!
                      :picture (format nil "~A~D" 
                                 "Y" (string-left-trim "H" [H :picture]))
                      :is-constant nil))
             )
         [l-env :move-in Y]
         [to-first :push-key H Y]
         Y
       )
    )
  )
)

(defmethod (basic-type :translate) ()
  self 
)

(defmethod (function-type :translate) ()
  (make-instance 'basic-type 
       :picture (format nil 
          "~A~D~A~D~A" "(" [domain :as-string] "->" [range :as-string] ")" )) 
)

(defmethod (function-type :as-string) ()
  (format nil "~A~D~A~D~A" "(" [domain :as-string] "->" [range :as-string] ")" )
)

(defmethod (basic-type :as-string) ()
  picture
)

; ---------------------------------------------------------------------------------------
;                                      Methode Translate
; ---------------------------------------------------------------------------------------

(defflavor algebraic-expression-node () (expression-node) )

(defflavor algebraic-application () (application) )

(defflavor algebraic-reference () (reference) )

(defmethod (disagreement-set :translate) (l-env)
  [ :iterate (pair)
      [pair :translate l-env]
  ]
)

(defmethod (disagreement-pair :translate) (l-env)
  (setq left  [left  :translate (make-instance 'lis-list) l-env])
  (setq right [right :translate (make-instance 'lis-list) l-env])
  self
)

(virtualmethod (expression-node :translate) (x-list l-env))

(defmethod (application :translate) (x-list l-env)
  (if [head :ref | :is-constant]
    [arguments :reduce (s) (h [head :translate x-list l-env])
        (make-instance 'algebraic-application                        
             :functor  h
             :argument [s :translate x-list l-env])
    ]
    [head :translate x-list l-env]  ; self is a simple variable subterm; here
                                    ; the types of the arguments get lost!
  )
)

(defmethod (abstraction :translate) (x-list l-env)
  (let* ((node-type [expression-node :expression-type])
         (l-decl    [*lambda-list* :new-lambda type node-type l-env])
         (fun       (make-instance 'algebraic-reference :ref l-decl))
         res
        )
     [x-list :move-in [l-decl :cons | :var] :to-pos 'last]
     ; (format t "x-list: ~A~%" x-list)
     [*exp-identity* :push-key self l-decl]
     (setq res (make-instance 'algebraic-application
                    :functor  fun
                    :argument [expression-node :translate x-list l-env]))
     [*exp-identity* :pop]
     [x-list :move-out 'last]
     [*lambda-list* :dispose-lambda]
     res
  )
)

(defmethod (reference :translate) (x-list l-env)
  (if [ref :is-constant]
    (let ((lambda-i [*exp-identity* :get [ :ref] :test 'eql]))
      (if lambda-i
        (make-instance 'algebraic-reference :ref [lambda-i :cons])
        (make-instance 'algebraic-reference 
                   :ref [*converter* :convert-constant ref l-env])
      )
    )
    (progn 
      [*H-contexts* :push-key ref [x-list :map (x) x]]
      (make-instance 'algebraic-reference 
                   :ref [*converter* :convert-variable ref l-env])
    )
  )
)


; --------------------------------------------------------------------------------
;                                Methode Retranslate
; --------------------------------------------------------------------------------

(virtualmethod (algebraic-expression-node :retranslate) (var-list env) )

(defmethod (algebraic-application :retranslate) (var-list env)
  (if (and (typep functor 'reference)
           (typep [functor :ref] 'lambda-declaration)
      )
    (let* ((var [functor :ref | :cons | :var])
           (res (make-instance 'abstraction 
                      :picture [var :picture]
                      :type [var :type]))          
         )
       [var-list :move-in var :to-pos 'last] 
       [*exp-identity* :push-key var res]
       [res :set-expression-node [argument :retranslate var-list env]]
       [*exp-identity* :pop]
       [var-list :move-out 'last]
       res
    )
    [arguments :reduce (s) (h [head :retranslate var-list env])
        (make-instance 'application                        
             :functor  h
             :argument [s :retranslate var-list env])
    ]
  )
)

(defmethod (algebraic-reference :retranslate) (var-list env)
  (if [ref :is-constant]
    (if (typep [ :ref] 'const-declaration)
                             ; the first includes two cases: ci of initial var-list
                             ; or of the created part of var-list!
      (make-instance 'reference  
              :ref [*exp-identity* :get [ :ref | :var] :test 'eql])
                             ; these are all constants without the ci.
      (make-instance 'reference 
              :ref [*converter* :reconvert-constant ref])  ; other constants
    )
    [ :lift-variable var-list env]   ; all variables
  )
)

(defmethod (algebraic-reference :lift-variable) (var-list env)
  (let* ((new-decl [*organizer* :new-variable ref var-list env]) 
         (NV (make-instance 'reference :ref new-decl))
       )
     [var-list :reduce (x) (p NV) 
           (make-instance 'application
                   :functor p
                   :argument (make-instance 'reference 
                                :ref [*exp-identity* :get x :test 'eql]))
     ]
  )
)

;------------------------------------------------------------------------------
;              Datentyp New-variable-organizer
; -----------------------------------------------------------------------------

(defflavor new-variable-organizer (old-new) ()
        (:gettable-instance-variables old-new)
        (:settable-instance-variables old-new)
        (:initable-instance-variables old-new) )

(setq *organizer* (make-instance 'new-variable-organizer))

(definitmethod (new-variable-organizer) ()
   (setq old-new (make-instance 'key-stack-list))
)

(defmethod (new-variable-organizer :new-variable) (old-variable var-list env)
 (let* ((pair (make-instance 'disagreement-pair
               :left old-variable
               :right var-list))
        (exist [ :give pair])
       )
    (if exist
       [old-new :get exist :test 'eql]  ; a new decl already exists: return it
       (let* ((type-list [var-list :map (x) [x :type]])
              (type (assemble-type [old-variable :type] type-list))
              (decl (make-instance 'declaration
                      :type type
                      :picture (new-pic-NV "NV")
                      :is-constant nil)) 
             )
          [env :move-in decl]
          [old-new :push-key pair decl]
          decl
       )
     )
  )
)

(defmethod (new-variable-organizer :give) (pair)
  ; Sei pair = (x,<y1, .., yn>)
  (let ((left  [pair :left])
        (right [pair :right])
        (pair-eq nil)
        (existing-pair (make-instance 'disagreement-pair))
       )
     [old-new :iterate ((key info)) 
       ; Sei key = (x',<y'1, .., y'k>)
       (if (equal [key :left] left)    ; x = x'
         (if (equal [key :right |:count] [right :count])  ; n = k
           (let ((list-eq t))
             (multiple-iterate ((dec1 [key :right])
                                (dec2 right))
               (if (not (equal dec1 dec2))
                 (setq list-eq nil)
               )
             )
             (if list-eq    ; y1 = y'1, ... , yn = y'k
               (progn
                 (setq pair-eq t)
                 (setq existing-pair key)
               )
             )
           )    ; let
         )      ; n = k
       )        ; x = x'
     ]
     (if pair-eq
        existing-pair
        nil
     )
  )
)

(setq *NV-count* 0) 
(defun new-pic-NV (base)
  (setq *NV-count* (1+ *NV-count*))
  (format nil "~A~D" base *NV-count*))


; --------------------------------------------------------------------------------
;                          Methode poormans-AC
; --------------------------------------------------------------------------------

; Gibt Liste mit Unifikatoren erster Stufe zurueck, die self AC-gerecht unifizieren.
; env ist das l-env, indem alle first-order Konstanten udn Variablen enthalten sind,
; bis auf die lambda-i!

(defmethod (disagreement-set :poormans-AC) (sub env)
  (if [ :is-empty]
    (make-instance 'lis-list :init-value `(,sub))  ; ready 
    (let* ((pair [ :move-out 'last])
           (left [pair :left])
           (right [pair :right])
          )
      (if (eq [pair :get-type] :rigid-rigid)
         (if (equal [left :head |:ref] [right :head |:ref])  ; case [f(si), f(ti)]
            (if (typep [left :head |:ref] 'c-declaration)
              [:c-variante left right sub env]
              (if (typep [left :head |:ref] 'ac-declaration)
                [:ac-variante [left :head |:ref] left right sub env]
                (progn
                  (multiple-iterate ((a [left :arguments])
                                     (b [right :arguments]))
                    [:move-in (make-instance 'disagreement-pair :left a :right b)] 
                  )
                  [ :poormans-AC sub env]
                )
              )
            )
            nil                                        ; case [f(si), g(ti)]
         )
         (if (eq [pair :get-type] :flexible-rigid)     ; case [x,f(si)] or [f(si),x]
         (progn             
            (if (not [[pair :right] :occur-check [pair :left]])
              (let* ((left  [pair :left])
                     (right [pair :right])
                     (new-sub  (make-instance 'substitution-hash 
                                :init-value `(( ,[left :ref] ,right))))
                   )
                [ :apply-sub new-sub] 
                [sub :move-in-key [left :ref] right] 
                ;[sub :expand]
                [ :poormans-AC sub env]
              )
              nil
            ))
            (if (equal [left :ref] [right :ref])       ; case [x,x]  
               [ :poormans-AC sub env]
               (let ((new-sub  (make-instance 'substitution-hash    ; case [x,y]
                                  :init-value `(( ,[left :ref] ,right))))
                    )
                  [ :apply-sub new-sub] 
                  [sub :move-in-key [left :ref] right] 
                  ;[sub :expand]
                  [ :poormans-AC sub env]
               )
            )
          )
       )   ; if
    )      ; let
  )        ; if
)

(defmethod (disagreement-set :c-variante) (left right sub env)
   (let ((help-dis [self :copy]))
      (multiple-iterate ((a [left :arguments])
                         (b [right :arguments]))
        [:move-in (make-instance 'disagreement-pair :left a :right b)] 
      )
      (multiple-iterate ((a [left :arguments])
                         (b [[right :arguments] 
                   :map :iteration-order 'from-last (x) x]))
        [help-dis :move-in (make-instance 'disagreement-pair :left a :right b)] 
      )
      (let ((first [ :poormans-AC [sub :copy] env])
            (sec   [help-dis :poormans-AC [sub :copy] env]))
        (if first
           (if sec
              (progn
                [sec :iterate (x) [first :move-in x]]
                first
              )
              first
           )
           (if sec
              sec
              nil
           )
         ) 
      )
    )
)

(defmethod (disagreement-set :ac-variante) (ac-decl left right sub env)
 (let* ((res-list (make-instance 'lis-list))
        (solution1 [self :copy])
        (solution2 [self :copy])
        (solution3 [self :copy])
        (solution4 [self :copy])
        (solution5 [self :copy])
        (solution6 [self :copy])
        (solution7 [self :copy])
        (arg1 [left :arguments])
        (arg2 [right :arguments])
        (x1 [arg1 :get-by-pos 1])
        (x2 [arg1 :get-by-pos 2])
        (y1 [arg2 :get-by-pos 1])
        (y2 [arg2 :get-by-pos 2])
        (var-type [ac-decl :type |:range |:domain])
        (NV1 [env :sub2-variable var-type "Neu"]) 
        (NV2 [env :sub2-variable var-type "Neu"])
        (NV3 [env :sub2-variable var-type "Neu"])
        (NV4 [env :sub2-variable var-type "Neu"])
       )
         (multiple-iterate ((a arg1)
                            (b arg2))
           [solution1 :move-in (make-instance 'disagreement-pair 
                               :left a 
                               :right b)] 
         )          
         [solution2 :move-in (make-instance 'disagreement-pair
                                :left  x1
                                :right y2)]
         [solution2 :move-in (make-instance 'disagreement-pair
                                :left  x2
                                :right y1)]
         [solution3 :move-in (make-instance 'disagreement-pair
                                :left  x1
                                :right [ac-decl :appl-build y1 NV1])]
         [solution3 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build x2 NV1] 
                                :right y2)]
         [solution4 :move-in (make-instance 'disagreement-pair
                                :left  x1
                                :right [ac-decl :appl-build NV2 y2])]
         [solution4 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build x2 NV2]
                                :right y1)]
         [solution5 :move-in (make-instance 'disagreement-pair
                                :left  x2
                                :right [ac-decl :appl-build y1 NV3])]
         [solution5 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build NV3 x1]
                                :right y2)]
         [solution6 :move-in (make-instance 'disagreement-pair
                                :left  x2
                                :right [ac-decl :appl-build NV4 y2])]
         [solution6 :move-in (make-instance 'disagreement-pair
                                 :left [ac-decl :appl-build NV4 x1]
                                 :right y1)]
         [solution7 :move-in (make-instance 'disagreement-pair
                                :left  x1
                                :right [ac-decl :appl-build NV2 NV1])]
         [solution7 :move-in (make-instance 'disagreement-pair
                                :left  x2
                                :right [ac-decl :appl-build NV4 NV3])]
         [solution7 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build NV4 NV2]
                                :right y1)]
         [solution7 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build NV3 NV1]
                                :right y2)]
         [res-list :move-in [solution1 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution2 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution3 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution4 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution5 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution6 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution7 :poormans-AC [sub :copy] env]]
         (let ((sol nil))
           (block suche-eine-lsg
           [res-list :iterate (sub idx) 
              (if sub 
                  (progn
                     (setq sol sub)
                     [res-list :move-out `(at-index ,idx)]
                     (return-from suche-eine-lsg nil)
                  ))
           ]
           )
           [res-list :iterate (sub) 
               (if sub
                   [sub :iterate (x) [sol :move-in x]]
               )
           ]
           sol
         )
  )
)

; @@@@@@@@@Sollte man mit sub-variable zusammentun!!
(defmethod (environment :sub2-variable) (type-node base)
  (let ((dec (make-instance 'declaration              ;note the similarity to :declare 
			    :type type-node
			    :picture (neu-pic base)
			    :is-constant nil)))
    [:move-in dec]
    (make-instance 'algebraic-reference :ref dec)))

(setq *Neu-count* 0) 
(defun neu-pic (base)
  (setq *Neu-count* (1+ *Neu-count*))
  (format nil "~A~D" base *Neu-count*))


(defflavor c-declaration () (declaration) )

(defflavor ac-declaration () (declaration) )

(defmethod (environment-lis :new-declare) (decl)
  [:move-in decl]
)

(defmethod (ac-declaration :appl-build) (x y)
  (make-instance 'algebraic-application 
          :functor (make-instance 'algebraic-application 
                             :functor (make-instance 'algebraic-reference :ref self)
                             :argument x)
          :argument y
  )
)

; Beachte: Solange :replace-alien-subterms jedes Vorkommen einer Variable 
; (oder eines Terms) durch eine andere Variable ersetzt, braucht man den
; occur-check nicht ...
(virtualmethod (expression-node :occur-check) (x))

(defmethod (application :occur-check) (x)
  (let ((occurs nil))
    [arguments :iterate (si)
      (if [si :occur-check x] 
        (setq occurs t)
      )
    ]
    occurs
  )
)

(defmethod (reference :occur-check) (x)
  (equal ref [x :ref])
)

