#|
   module      : bag-hash
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 

	ADT bag implemented as a hash table

   history     : 90.06.26 Version 1
   errors      :
   bugs        : iteration method does not obey all iteration variants
		 object is stored only once with counter
    
|#

(defflavor bag-hash (content) (bag))

; -------------------- constructors (page 57)
(definitmethod (bag-hash) (&key init-value) 
    (setq content (make-hash-table))
    (dolist (e init-value)
	[:move-in e]
    )
)

(defmethod (bag-hash :clear) () (clrhash content) self)

(defmethod (bag-hash :move) (from-pos &key to-pos from) 
    [self :move-in [from :move-out from-pos]]
)

(defmethod (bag-hash :move-in) (e  &key to-pos) 
    (let ((nr (gethash e content)))
	(if nr
	    (setq nr 0)
	    (setq nr (+ 1 nr))
	)
	(setf (gethash e content) nr)
	(cons e nr)
    )
)

(defmethod (bag-hash :move-out) (pos) 
    (setq pos (if (listp pos) (cadr pos) pos))
    (let ((nr (gethash (car pos) content)))
	(if (equal nr 1)
            (remhash (car pos) content) 
	    (setf (gethash (car pos) content) (- nr 1))
	)
    )
    (car pos)
)

(defmethod (bag-hash :index-of) (i) (cons i (gethash i content)))
(defmethod (bag-hash :get-elem) (e) (if (listp e) (cadr e) e))

; -------------------- Iterators (page 80)
(defmethod (bag-hash :iterate-method) (f)
    (maphash #'(lambda (x y) 
		   (dotimes (i y)
		       (funcall f x x)
		      )) content))
