#|

   module      : str.test
   owner       : Franz Weber
   author      : Franz Weber
   explanation : tests for str
   history     : 90.06.26 Version (first lof)
   errors      :
   bugs        :
|#

(defflavor aggregate-test () (aggregate))
; -------------------- constructors (page 57)
(defmethod (aggregate-test :generate-any) () (virtual))
    ; generates one insertable element
(defmethod (aggregate-test :generate-lof) () (virtual))
    ; generates a list of insertable elements
(defmethod (aggregate-test :generate-objects) () (virtual))
    ; generates the objects to test
(defmethod (aggregate-test :test) ()

    (dolist
	(tt
	    (list :test-clear
	      :test-insert
	      :test-insert-s-exp
	      :test-insert-struct
	      :test-find-or-insert
	      :test-remove
	      :test-remove-elem
	      :test-remove-all
	      :test-move-in
	      :test-move-out
	      :test-move
	      :test-move-struct
	      :test-swap
	      :test-copy
	      :test-copy-from
	      :test-transfer
	      :test-transfer-from
	      :test-index-of
	      :test-set-elem
	      :test-get-elem
	      :test-transfer-elem-out
	      :test-transfer-elem-in
	      :test-modify-elem
	      :test-count
	      :test-is-empty
	      :test-find
	      :test-count-elem
	      :test-is-elem
	      :test-equal
	      :test-iterate
	      :test-iterate-move
	      :test-reduce
	      :test-find-property
	      :test-count-property
	      :test-exists
	      :test-forall
	      :test-compare
	      :test-<=
	      :test-<
	      :test->=
	      :test->
	      :test-max
	      :test-min
	      :test-remove-min
	      :test-remove-max
	     )
	)
	(dolist (o [:generate-objects])
	   (format t "test ~A, ~A~%" tt o)
	   (unless [[o :copy] tt [:generate-any] [:generate-lof]] 
	       (format t "test ~A failed for object ~A~%" tt o)
	       (print o)
	   )
       )
    )
)

(defmethod (aggregate-test :test-clear) (any lof) [[:clear] :is-empty])

(defmethod (aggregate-test :test-insert) (any lof)
    [:insert any]
    (not [:is-empty])
)

(defmethod (aggregate-test :test-insert-s-exp) (any lof)
    [:insert-s-exp lof]
    (let ((b t))
    (dolist (e lof)
	(setq b (and b [:is-elem e]))
    )
    b
    )
)

(defmethod (aggregate-test :test-insert-struct) (any lof)

    (let ((s (make-instance (type-of self))))
	[s :insert-struct self]
	[:forall (x) [s :is-elem x]]
    )
)

(defmethod (aggregate-test :test-find-or-insert) (any lof)
  [:find-or-insert any]
  (let ((c [:copy]))
      [c :find-or-insert any]
      [c :find-or-insert any]
      [c :equal self]
  )
)

(defmethod (aggregate-test :test-remove) (any lof)

    (let ((c [:copy]))
	(let ((index [c :insert any]))
	   (let ((i [c :count-elem any]))
	       [c :remove `(at-index ,index)]
	       (equal (- i 1) [c :count-elem any])
	   )
	)
    )
)

(defmethod (aggregate-test :test-remove-elem) (any lof)

    (let ((c [:copy]))
	[c :insert any]
	(let ((i [c :count-elem any]))
	    [c :remove-elem any]
	    (equal (- i 1) [c :count-elem any])
	)
    )
)

(defmethod (aggregate-test :test-remove-all) (any lof)
    [:remove-all any]
    (equal [:count-elem any] 0)
)

(defmethod (aggregate-test :test-move) (any lof)

    (let ((c [:copy]))
	[c :insert any]
	[:remove-all any]
	[:move `(at-index ,[c :find any]) :from c]
	[:is-elem any]
    )
)

(defmethod (aggregate-test :test-move-in) (any lof)

    [:remove-all any]
    [:move-in any]
    [:is-elem any]
)

(defmethod (aggregate-test :test-move-out) (any lof)

    [:remove-all any]
    [:insert any]
    [:move-out `(at-index ,[:find any])]
    (not [:is-elem any])
)

(defmethod (aggregate-test :test-move-struct) (any lof)

    [(make-instance (type-of self)) :move-struct self]
    [:is-empty]
)

(defmethod (aggregate-test :test-swap) (any lof) 
    [:insert-s-exp lof]
    (let ((op
    (if (< [:index-pos [:find (second lof)]] [:index-pos [:find (first lof)]])
	'>
	'<
    )
    ))
    [:swap
	`(at-index ,[:find (first lof)])
	`(at-index ,[:find (second lof)])
    ]
    (funcall op [:index-pos [:find (second lof)]] [:index-pos [:find (first lof)]])
    )
)

(defmethod (aggregate-test :test-copy) (any lof)
    [:equal [:copy]]
)

(defmethod (aggregate-test :test-copy-from) (any lof)
    (let ((c [:copy]))
	[c :equal [:copy-from c]]
    )
)

(defmethod (aggregate-test :test-transfer) (any lof)
    (let ((c [:count])
	  (s (make-instance (type-of self)))
	 )
	(setq s [:transfer])
	(and
	    (equal c [s :count])
	    [:is-empty]
	)
    )
)

(defmethod (aggregate-test :test-transfer-from) (any lof)
    (let ((c [:count])
	  (s (make-instance (type-of self)))
	 )
	[s :transfer-from self]
	(and
	    (equal c [s :count])
	    [:is-empty]
	)
    )
)

(virtualmethod (aggregate-test :test-index-of) (any lof))

(defmethod (aggregate-test :test-get-elem) (any lof)
    [:insert any]
    (flavor-equal [:get-elem [:find any]] any)
)

(defmethod (aggregate-test :test-set-elem) (any lof)
    [:insert any]
    (let ((i [:find any]))
	[:set-elem i (second lof)]
	(flavor-equal [:get-elem i] (second lof))
    )
)

(defmethod (aggregate-test :test-transfer-elem-out) (any lof)
    (flavor-equal [:transfer-elem-out [:insert any]] any)
)

(defmethod (aggregate-test :test-transfer-elem-in) (any lof)
   [:insert any]
   [:remove-all (third lof)]
   (let ((i [:find any])
	)
       [:transfer-elem-in i (third lof)]
       [:is-elem (third lof)]
    )
)

(defmethod (aggregate-test :test-modify-elem) (any lof)
   [:insert any]
   [:remove-all (third lof)]
   (let ((i [:find any])
	)
	[:modify-elem i #'(lambda (x) (third lof))]
	[:is-elem (third lof)]
    )
)

(defmethod (aggregate-test :test-count) (any lof)

    (and
	(progn [:clear] (equal [:count] 0))
	(progn [:insert (first lof)] (equal [:count] 1))
	(progn [:insert (second lof)] (equal [:count] 2))
    )
)

(defmethod (aggregate-test :test-count-elem) (any lof)

    (and
	(progn [:clear] (equal [:count-elem (second lof)] 0))
	(progn [:insert (first lof)] (equal [:count-elem (second lof)] 0))
	(progn [:insert (second lof)] (equal [:count-elem (second lof)] 1))
    )
)

(defmethod (aggregate-test :test-is-empty) (any lof)
    (and
       [[:clear] :is-empty]
       (progn [:insert any] (not [:is-empty]))
    )
)

(defmethod (aggregate-test :test-find) (any lof)

    [:insert any]
    (flavor-equal [:get-elem [:find any]] any)
)

(defmethod (aggregate-test :test-is-elem) (any lof)

    (and
       (progn [:insert any] [:is-elem any])
       (progn [:remove-all any] (not [:is-elem any]))
    )
)


(defmethod (aggregate-test :test-equal) (any lof)
    [:equal [:copy]]
)

(defmacro plus (a b)
   `(if (instancep ,a)
       [,a :+ ,b]
       (if (instancep ,b)
	   [,b :+ ,a]
           (+ ,a ,b)
	)
    )

)
(defmethod (aggregate-test :test-iterate) (any lof)

    [:clear]
    [:insert-s-exp lof]
    (let ((s 0))
    [:iterate
	(el in)
	(setq s (plus s (if (listp el) (cadr el) el)))
    ]
    (let ((fun #'(lambda (b a) 
		     (plus (if (listp a) (cadr a) a)
		        (if (listp b) (cadr b) b)))))
	(equal s (reduce fun lof)))
    )
)

(defmethod (aggregate-test :test-iterate-move) (any lof)

    [:clear]
    [:insert-s-exp lof]
    [:iterate 'move
	(el in)
	(when (flavor-equal el (second lof)) [:remove `(at-index ,in)])
    ]
    (not [:is-elem (second lof)])
)

(defmethod (aggregate-test :test-reduce) (any lof)
    [:clear]
    [:insert-s-exp lof]
    (let ((fun #'(lambda (b a &optional in) 
		     (plus (if (listp a) (cadr a) a)
		        (if (listp b) (cadr b) b)))))
	(equal [:reduce-method fun :unit 0] (reduce fun lof)))
)

(defmethod (aggregate-test :test-find-property) (any lof)
    [:clear]
    [:insert-s-exp lof]
    (flavor-equal [:get-elem 
	       [:find-property (x) (flavor-equal x (second lof))]] (second lof))

)

(defmethod (aggregate-test :test-count-property) (any lof)
    [:clear]
    [:insert-s-exp lof]
    (equal 
	 [:count-property (x) (or (equal x (first lof))(equal x (second lof)))]
	2)

)

(defmethod (aggregate-test :test-exists) (any lof)
    [:clear]
    [:insert-s-exp lof]
    [:exists (x) (flavor-equal x (second lof))]
)

(defmethod (aggregate-test :test-forall) (any lof)
    [:clear]
    [:insert-s-exp lof]
    [:forall (x) (find x lof :test 'equal)]
)

(defun data-glo-aggregate-test-generate-differs (self a1 a2)
    (let ((a [self :copy])
	  (b [self :copy])
	  (c [self :copy])
	 )
	[b :insert a1]
	(when [a :equal b]
	    [b :remove-all a1]
	)
	[c :insert a2]
	(when [a :equal c]
	    [c :remove-all a2]
	)
	(if [a :< b]
	    (if [b :< c]
		(list a b c)
		(if [c :< a]
		    (list c a b)
		    (list a c b)
		)
	    )
	    (if [a :< c]
		(list b a c)
		(if [c :< b]
		    (list c b a)
		    (list b c a)
		)
	    )
	)
    )
)

(defmethod (aggregate-test :test-compare) (any lof)

    (let ((l (data-glo-aggregate-test-generate-differs self 
		 (first lof) (second lof))))
	(and
	    (equal [:compare self] 'equal)
	    (equal [(first l) :compare (second l)] 'less)
	    (equal [(second l) :compare (first l)] 'greater)
	)
    )
)

(defmethod (aggregate-test :test-<=) (any lof)

    (let ((l (data-glo-aggregate-test-generate-differs self 
		 (first lof) (second lof))))
	(and
	    [(first l) :<= (second l)]
	    [(first l) :<= (third l)]
	    (not [(second l) :<= (first l)])
	    [:<= self]
	)
    )
)

(defmethod (aggregate-test :test-<) (any lof)

    (let ((l (data-glo-aggregate-test-generate-differs self 
		 (first lof) (second lof))))
	(and
	    [(first l) :< (second l)]
	    [(first l) :< (third l)]
	    (not [(second l) :< (first l)])
	    (not [:< self])
	)
    )
)

(defmethod (aggregate-test :test->=) (any lof)

    (let ((l (data-glo-aggregate-test-generate-differs self 
		 (first lof) (second lof))))
	(and
	    [(second l) :>= (first l)]
	    [(third l) :>= (first l)]
	    (not [(first l) :>= (second l)])
	    [:>= self]
	)
    )
)

(defmethod (aggregate-test :test->) (any lof)

    (let ((l (data-glo-aggregate-test-generate-differs self 
		 (first lof) (second lof))))
	(and
	    [(second l) :> (first l)]
	    [(third l) :> (first l)]
	    (not [(first l) :> (second l)])
	    (not [:> self])
	)
    )
)

(defmethod (aggregate-test :test-max) (any lof) 

        [:clear]
        [:insert-s-exp lof]
	(flavor-equal [:max] (third lof))
)

(defmethod (aggregate-test :test-min) (any lof) 

        [:clear]
        [:insert-s-exp lof]
	(flavor-equal [:min] (first lof))
)

(defmethod (aggregate-test :test-remove-max) (any lof) 

        [:clear]
        [:insert-s-exp lof]
	[:remove-max]
	(not [:is-elem (third lof)])
)

(defmethod (aggregate-test :test-remove-min) (any lof) 

        [:clear]
        [:insert-s-exp lof]
	[:remove-min]
	(not [:is-elem (first lof)])
)
