#|
   module      : com
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 
	
	general functions for comparison
	class 'comparison' of objects, which have comparison methods

	provides macros which universally compares a lot of data structure independent of
	their type


	(defmacro if-comp (object then &optional (else object)) ... )
	Tests, wether the object is of class 'comparison'. If so, the then action is performed
	otherwise the 'else'-action is performed

	(defmacro flavor-compare (a b) ... )
	Universally compares lisp data structures
	Returns 'less, 'greater or 'equal

	(defun fla-compare (a b) ... )
	Same as a fucntion

	(defmacro flavor-equal (a b)
	tests universally, wether a and b are equal
		    

	(defflavor comparison () () )

        (virtualmethod (comparison :compare) (right)
            ; returns ternary result 'less 'greater 'equal
        )
        (defmethod (comparison :<=) (right) )
        (defmethod (comparison :>=) (right))
        (defmethod (comparison :<) (right))
        (defmethod (comparison :>) (right))

        (defmethod (comparison :equal) (right))
        

   history     : 90.07.24 Version 1
   errors      :
   bugs        :
|#

(mod-require "flavor/v4")

(defflavor comparison () () )


; -------------------- Compare Macros
(defmacro if-comp (object then &optional (else object))
    `(if (typep ,object 'comparison) ,then ,else)
)

(defmacro flavor-compare (a b)
    `(if-comp ,a
	[,a :compare ,b]
        (if-comp ,b
	    (case [,b :compare ,a]
		('greater 'less)
		('less 'greater)
		('equal 'equal)
	    )
            (if (equal ,b ,a) 'equal
		(if (equal (type-of ,a) (type-of ,b))
		(if (characterp ,a)
		    (if (char< ,a ,b) 'less 'greater)
		(if (typep ,a 'sequence)
		   (if (equal (length ,a) (length ,b))
		       (block the-block
		       (map nil
			   #'(lambda (x y &aux c)
				 (setq c (fla-compare x y))
				 (unless (equal c 'equal)
				     (return-from the-block c)
				 )
			     )
			   ,a
			   ,b
		       )
			   (return-from the-block 'equal)
		       )
		       (if (< (length ,a) (length ,b))
			   'less
			   'greater
		       )
		   )
		   (if (symbolp ,a)
		      (if (string<= (string ,a) (string ,b)) 
			  'less
		          'greater
		      )
		      (if (<= ,a ,b)
			  'less
			  'greater
		      )
		   )
		))
		(fla-compare (type-of ,a) (type-of ,b))
		)
	    )
	)
    )
)

(defun fla-compare (a b) (flavor-compare a b))

(defmacro flavor-equal (a b)
    `(if-comp ,a
	[,a :equal ,b]
        (if-comp ,b
	    [,b :equal ,a]
            (equal (flavor-compare ,b ,a) 'equal)
	)
    )
)

; -------------------- Order Dependent Operations

(virtualmethod (comparison :compare) (right)
    ; returns ternary result 'less 'greater 'equal
)
(defmethod (comparison :<=) (right) 
  (case [:compare right]
      ((less equal) t)
      (t nil)
  )
)
(defmethod (comparison :>=) (right)
  (case [:compare right]
      ((greater equal) t)
      (t nil)
  )
)

(defmethod (comparison :<) (right)
  (case [:compare right]
      (less t)
      (t nil)
  )
)

(defmethod (comparison :>) (right)
  (case [:compare right]
      ((greater) t)
      (t nil)
  )
)

(defmethod (comparison :equal) (right) (equal [:compare right] 'equal))

(compile-flavor-methods comparison)
