#|
   module      : key
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 

	Instanciation of key generic with general aggregates.
	No logic extension but change of the set of virtual kernel methods.

	This module serves to easily implement 'efficient' key
	aggregates. This efficiency is achieved by turning some
	non-virtual methods of 'key-generic' in virtual methods and
	turning  some virtual methods of 'aggregate' in non virtual
	methods. 

	The kernel basis of aggregate are move and transfer methods.
	The module 'key generic' implements move and transfer methods
	based on key or info parts on top of these virtual methods.
	However, concrete key structures often suppport especially the
	methods which identify obects by key, so that these operations
	are more suited as kernel methods for key aggregates.

	The module presented here offers no additional logical
	properties. But when you implement a concrete instanciation
	of key aggregates on the basis of this module you have to
	implement other virtual methods than on the basis of
	'aggregate' and 'key generic' alone. For details, see below.

   history     : 90.06.26 Version 1
   errors      :
   bugs        :
|#

(mod-require "data/glo/v2/key-gen")
(defflavor key-aggregate () (key-generic-aggregate))

; -------------------- constructors (page 57)

(defmethod (key-aggregate :move-in) (e &key to-pos)
   [self :move-in-key (key-of e) (info-of e) :to-pos to-pos])
(virtualmethod (key-aggregate :move-in-key) (k i &key to-pos))
(defmethod (key-aggregate :move-out) (pos)
    (multiple-value-setq (k i) [self :move-out-key pos])
    (pair k i)
) 
(virtualmethod (key-aggregate :move-out-key) (pos))

; -------------------- Operations based on indices (page 73) 

(defmethod (key-aggregate :get-elem) (i) 
    (pair [self :get-key i] [self :get-info i]))
(hidemethod (key-aggregate :set-elem))
(virtualmethod (key-aggregate :get-key) (i) (virtual))
(virtualmethod (key-aggregate :get-info) (i) (virtual))
(virtualmethod (key-aggregate :set-info) (i info) (virtual))
(hidemethod (key-aggregate :transfer-elem-in))
(hidemethod (key-aggregate :transfer-elem-out))
(virtualmethod (key-aggregate :transfer-info-in) (i info) (virtual))
(virtualmethod (key-aggregate :transfer-info-out) (i) (virtual))
(hidemethod (key-aggregate :modify-elem))
(defmethod (key-aggregate :modify-info) (i modification)
    [self :set-info i (funcall modification [self :get-info i])]
)

; -------------------- Access by Position Count (page 76)

; -------------------- Selector Operations (page 77)

; -------------------- Iterators (page 80)

; -------------------- Reduction (page 86)

; -------------------- Find, Skip and Count (page 89)

; -------------------- Existential and Universal Quantifiers

; -------------------- Order Dependent Operations

(compile-flavor-methods key-aggregate)

