#|
   module      : pointer
   owner       : Franz Weber
   author      : Franz Weber
   explanation : pointer objects

	pointer objects are objects, which point to another object. This is useful in
	hierarchical structure of aggregates to delimit hierarchical objects. Iterators over
	pointers don't find a component. However some other methods are propagated to the value
	object of the pointer.  The pointer is realized by a simple instance variable 'ref'.
	Some virtual methods of copies and comparsion are implemented. Due to the simple
	strucutre of pointers, this is an easy task.

   history     : 90.08.03
   errors      :
   bugs        : There should be no pointers but null aggregates !
		 see e.g. 'basic-type'
|#

(mod-require "data/glo/v2/simple-aggregate-empty")

(defflavor pointer (ref) (simple-aggregate-empty)
    (:gettable-instance-variables ref)
    (:settable-instance-variables ref)
    (:init-keywords :ref)
)

(defmethod (pointer :copy-from) (from)
    (setq ref [from :ref])
)

(defmethod (pointer :deallocate) () (setq ref nil))

(defmethod (pointer :compare) (right)
    (flavor-compare ref [right :ref])
)

