#|
   module      : unordered
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 
   
	abstract data type unordered
	it is an abstraction from both, bags and sets
	different unordered data types differ only in their semantic
	of 'move-in'. This implies the semantic of insert and similar
	operations.

	1. S-Expressions
	----------------

	simply lisp lists

	methods:

	(defmethod (unordered :insert-s-exp) (l) ... )

	2. Generic Implementations
	--------------------------

	(defmethod (unordered :copy-from) (from) ... )
	(defmethod (unordered :transfer-from) (from) ... )
	(defmethod (unordered :equal) (right) ... )
	(defiterationmethod (unordered :iterate-move-method) (f) ... )
	    here the implementation sub tree for
	    iterate-non-recursive-method is altered to

	    iterate-non-recursive-method
		iterate-move-method
		    -> iterate-normal-method
                iterate-normal-method

	(defmethod (unordered :compare) (right) ... )


	3. Hiding Methods based on Indices
	----------------------------------

	(hidemethod (unordered :set-elem))
	(hidemethod (unordered :transfer-elem-in))
	(hidemethod (unordered :transfer-elem-out))
	(hidemethod (unordered :modify-elem))
	(hidemethod (unordered :slice))


	4. Unorderd Specific Operations
	--------------------------

	(defmethod (unordered :subset) (right)
		self <= right
	(defmethod (unordered :proper-subset) (right)
		self < right
	(defmethod (unordered :superset) (right)
	(defmethod (unordered :proper-superset) (right)


   history     : 91.03.01 Version 1
   errors      :
   bugs        : 
|#

(mod-require "data/glo/v2/aggregate")
(mod-require "data/lis/v2/lis-list")

(defflavor unordered () (aggregate))

; -------------------- constructors (page 57)

(defmethod (unordered :insert-s-exp) (l)
    (dolist (e l) [:insert e])
)

(defmethod (unordered :move-in-s-exp) (l)
    (dolist (e l) [:move-in e])
)

(defmethod (unordered :insert-aggregate) (a)
    [a :iterate (x) [:move-in x]]
)

(defmethod (unordered :move-aggregate) (a)
    [a :iterate (x i) [:move :from-pos i :from a ]]
)

(defmethod (unordered :copy-from) (from)
    [:clear]
    [from :iterate (el) [:move-in el]]
    self
)

(defmethod (unordered :transfer-from) (from)
    [:copy-from from]
    [from :clear]
    self
)

; -------------------- Operations based on indices (page 73)

(hidemethod (unordered :set-elem))
(hidemethod (unordered :transfer-elem-in))
(hidemethod (unordered :transfer-elem-out))
(hidemethod (unordered :modify-elem))

; -------------------- Selector Operations (page 77)

(defmethod (unordered :equal) (right)

    (and
	[right :subset self]
	[:subset right]
    )
)

(hidemethod (unordered :slice))

; -------------------- Iterators (page 80)

(defiterationmethod (unordered :iterate-move-method) (f)

     (let ((c [:copy]))
	 [:calliterationmacro :move nil (el) :iterate
	     (funcall f el [:find el])
	 ]
     )
)

; -------------------- Compare

(defmethod (unordered :compare) (right)
    [:reduce (el) (res (make-instance 'lis-list)) 
          [res :move-in el] res 
    |:sort
    |:compare
     [right :reduce (el) (res (make-instance 'lis-list)) 
	 [res :move-in el] res 
     |:sort
     ]
    ]
)

; -------------------- Specific Operations

(defmethod (unordered :subset) (right)
  [:forall (elem) (<= [:count-elem elem] [right :count-elem elem])]
)

(defmethod (unordered :proper-subset) (right)
    (and
	[self :subset right]
	(not [:equal right])
    )
)

(defmethod (unordered :superset) (right)
    [right :subset self]
)

(defmethod (unordered :proper-superset) (right)
    (and
	[:superset right]
	(not [:equal right])
    )
)
