#|
   module      : com-lis-list
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 

	variant of lis-list, whereby  each  member of the list is
	a component, that is a father pointer is maintained (see class
	component in global definitions).

   history     : 90.06.26 Version 1
   errors      :
   bugs        :
|#

(mod-require "data/lis/v2/lis")

(defflavor com-lis-list (content) (lis-list))

; --------------------

(defmethod (com-lis-list :insert-s-exp) (e) 
    (dolist (el e) [el :set-father self])
    (setf content (append content (copy-list e)))
)

(defmethod (com-lis-list :move-in-s-exp) (e) 
    (dolist (el e) [el :set-father self])
    (setf content (append content e))
)
(defmethod (com-lis-list :clear) ()
    (dolist (el content)  [el :set-father nil])
    (setf content nil)
    self
)
(defmethod (com-lis-list :slice) (sub) 
    (error " not implemented for component lists com-lis-list")
)
(defmethod (com-lis-list :move-out) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos
	 (prog1 (cadr pos) 
		[(cadr pos) :set-father nil]
		(setf (cdr pos) (cddr pos)))
	 (prog1 (car content) 
		[(car content) :set-father nil]
		(setf content (cdr content)))
    )
)
(defmethod (com-lis-list :move-in) (e &key (to-pos nil to-pos-p))
    (unless to-pos-p (setf to-pos 'dont-care))
    (setq to-pos [:normalize-insert-pos to-pos])
    [e :set-father self]
    (if to-pos
	 (setf (cdr to-pos) (cons e (cdr to-pos)))
	 (setf content (cons e content))
    )
)

; -------------------- Operations based on indices (page 73)

(defmethod (com-lis-list :set-elem) (i e) 
    [(car i) :set-father nil]
    [e :set-father self]
    (setf (car i) e)
)
(defmethod (com-lis-list :transfer-elem-out) (i)
    [(car i) :set-father nil]
    (car i)
)
(defmethod (com-lis-list :transfer-elem-in) (i e) 
    [e :set-father self]
    (setf (car i) e)
)
  
(compile-flavor-methods com-lis-list)
