#|
   module      : lis-lazy
   owner       : Peter Sanders
   author      : Peter Sanders, Franz Weber
   explanation : 

	lazy implementation of lists.
	all kind of modifying access are forbidden, except

	initialization using the operation 
        (lazy-cons <car> <cdr>)
        The cdr may be a function call producing a very long or even
        infinitely long list.

	iteration order 'from-last and 'move is forbidden

        all operations depending on index comparison are
        not yet implemented

        these lists will usually be accessed by get-elem, first, next and
        nondestructive iterators.


   history     : 91.03.01 Version 1
   errors      :
   bugs        : index comparison is wrong
		 hence, the keywords 'to' and `part` for iterations does not work

|#

(mod-require "data/lis/v2/lis")
(mod-require "lazy/lazy-list")

(defflavor lis-lazy (content) (lis)
  :gettable-instance-variables)

; -------------------- Constructors (page 57)

; kernel method
(definitmethod (lis-lazy) (&key init-value) (setf content init-value)
)

(defmethod (lis-lazy :copy-from) (from)
    (setq content [from :content])
)

(defmethod (lis-lazy :deallocate) () (setq content nil))

; -------------------- Operations based on indices (page 73)

(defmethod (lis-lazy :normalize-remove-pos) (pos)
    (if (symbolp pos)
	(case pos
	    (first     content)
	)
	(case (car pos)
	    (at-index (cadr pos))
	    (after    (lazy-cdr (cadr pos)))
	    (first     content)
	)
    )
)


(defmethod (lis-lazy :index-of) (pos) [:normalize-remove-pos pos])

; kernel method
(defmethod (lis-lazy :get-elem) (i) (lazy-car i))

; -------------------- Selectors (page 77)

(defmethod (lis-lazy :first) () content)
(defmethod (lis-lazy :next) (i) (lazy-cdr i))

; -------------------- Iterators (page 80)

(defiterationmethod (lis-lazy :iterate-non-recursive-method) 
    (f)
    ; obey iteration-order from 

    (unless from (setq from content))
    (loop
	(unless from (return))
	(funcall f (lazy-car from) from)
	(setq from (lazy-cdr from))
    )
)

; -------------------- Parts (page 130)

; testing

; ----- print-self

(defmethod (lis-lazy :print-self-component-lis) 
    (stream prindepth indent count)

    (if *print-length*
	(data-glo-print-lis [:component-lis] stream prindepth indent count)
	(progn (write content :stream stream) count)
    )
)

(compile-flavor-methods lis-lazy)

