#|
   module      : lis-lis
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 

	concrete implementation of lis by lisp lists

	indices are the sublist beginning with the element pointed to

   history     : 90.06.26 Version 1
   errors      :
   bugs        : parts not implemented
		 normalize-remove/insert-pos
		 should be functions instead of methods !!
		 iterate method does not obey all variants
|#

(mod-require "data/lis/v2/lis")

(defflavor lis-list (content) (lis))

; --------------------

; kernel method
(definitmethod (lis-list) (&key init-value) (setf content nil)
    [:move-in-s-exp init-value]
)
; kernel method
(defmethod (lis-list :insert-s-exp) (e) 
    (setf content (append content (copy-list e))))
(defmethod (lis-list :move-in-s-exp) (e) (setf content (append content e)))
; kernel method
(defmethod (lis-list :clear) () (setf content nil) self)

; kernel method
(defmethod (lis-list :slice) (sub) 
    (setf content (car sub))
    (setf (cdadr sub) nil)
)

(defmethod (lis-list :move) 
    (from-pos &key (to-pos nil to-pos-p) (from nil from-p) &aux help)

    (unless to-pos-p (setf to-pos '(dont-care)))
    (unless from-p (setf from self))
    (setf help [from :move-out from-pos])
    [:move-in help :to-pos to-pos]
)

; kernel method
(defmethod (lis-list :move-out) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos
	 (prog1 (cadr pos) (setf (cdr pos) (cddr pos)))
	 (prog1 (car content) (setf content (cdr content)))
    )
)

; kernel method
(defmethod (lis-list :move-in) (e &key (to-pos nil to-pos-p))
    (unless to-pos-p (setf to-pos 'dont-care))
    (setq to-pos [:normalize-insert-pos to-pos])
    (if to-pos
	 (setf (cdr to-pos) (cons e (cdr to-pos)))
	 (setf content (cons e content))
    )
)

; -------------------- Operations based on indices (page 73)

(defmethod (lis-list :normalize-remove-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (first     nil)
	    (last      [:previous [:last]])
	)
	(case (car pos)
	    (at-index [:previous (cadr pos)])
	    (after    pos)
	    (first     nil)
	    (last      [:previous [:last]])
	)
    )
)

(defmethod (lis-list :normalize-insert-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (dont-care nil)
	    (first     nil)
	    (last      [:last])
	)
	(case (car pos)
	    (before [:previous (cadr pos)])
	    (after    pos)
	    (dont-care nil)
	    (first     nil)
	    (last      [:last])
	)
    )
)

(defmethod (lis-list :index-of) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos (cdr pos) content))

; kernel method
(defmethod (lis-list :set-elem) (i e) (setf (car i) e))
(defmethod (lis-list :get-elem) (i) (car i))
(defmethod (lis-list :transfer-elem-out) (i) (car i))
; kernel method
(defmethod (lis-list :transfer-elem-in) (i e) (setf (car i) e))

; -------------------- Iterators (page 80)

(defiterationmethod (lis-list :iterate-non-recursive-method) 
    (f)
    ; obey iteration-order move part from to 
    ; move is transparently obeyed

     (if (eql iteration-order 'from-last)
	(progn
	   ; to should contain the whole list to iterate upon
	   ; from should contain the cdr of the first element
	   (if (consp part)
	       (progn
		   (setq from (cdadr part))
		   (setq to (car part))
	       )
	       (progn
	           (if to 
		       (setq to [:index-of to])
		       (setq to content))
		   (when from 
		       (setq from (cdr [:index-of from]))
		   )
	       )
	   )
	   (labels
	     ((rec (index)
		  (unless (eql index from) 
		      (rec (cdr index))
		      (funcall f (car index) index)
		  )
	      )
	     )
	     (when to (rec to))
	   )
	)
	(progn
	   ; from should contain whole list to iterate over
	   ; to should contain cdr of last element to iterate
	   (if (consp part)
	       (progn
		   (setq from (car part))
		   (setq to (cdadr part))
	       )
	       (progn
	           (if from 
		       (setq from [:index-of from])
		       (setq from content))
		   (when to 
		       (setq to (cdr [:index-of to]))
		   )
	       )
	   )
	    (when from
		(let ((n (cdr from)))
		    (loop
			(funcall f (car from) from)
			(when (equal n to) (return))
			(setq from n)
			(setq n (cdr from))
		    )
		)
	    )
	)
     )
)

; -------------------- Shift and Invert (page 132)

; this is an optimization
(defmethod (lis-list :invert) ()
    (setf content (list-nreverse content))
)


; -------------------- Parts (page 130)

(compile-flavor-methods lis-list)
