#|
   module      : lis-tab
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 

	concrete implementation of lis by an adjustable lisp array

   history     : 90.06.26 Version 1
   errors      :
   bugs        : Operations based on indices
		 should possibly be seprated from 'str'
		 parts not implemented
		 normalize-remove/insert-pos
		 should be functions instead of methods !!
		 iterate method does not obey all variants
|#


(mod-require "data/lis/v2/lis")

(defflavor lis-tab (content) (lis))

; --------------------

; kernel method
(definitmethod (lis-tab) (&key init-value)
    (setf content (make-array 100 :adjustable t :fill-pointer 0))
    [:move-in-s-exp init-value]
)

(defmethod (lis-tab :clear) () (setf (fill-pointer content) 0) self)

(defmethod (lis-tab :slice) (sub) 
    (error "not implemented")
)

(defmethod (lis-tab :move) 
    (from-pos &key (to-pos nil to-pos-p) (from nil from-p) &aux help)

    (unless to-pos-p (setf to-pos '(dont-care)))
    (unless from-p (setf from self))
    (if (eql self from)
       (progn
	  (setq from-pos [from :normalize-remove-pos from-pos])
	  (setq to-pos [from :normalize-insert-pos to-pos])
	  (setf help (aref content from-pos))
	  (if (< from-pos to-pos)
	      (let ((i from-pos))
		  (loop
		      (when (equal i to-pos) (return))
		      (setf (aref content i) (aref content (+ 1 i)))
		      (incf i)
		  )
	      )
	      (let ((i to-pos))
		  (loop
		      (when (equal i from-pos) (return))
		      (setf (aref content (- i 1)) (aref content i))
		      (decf i)
		  )
	      )
	  )
	  (setf (aref content to-pos) help)
       )
       (progn
	  (setf help [from :move-out from-pos])
	  [self :move-in help :to-pos to-pos]
       )
    )
)

(defmethod (lis-tab :move-out) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (prog1
	(aref content pos)
	(let ((i pos))
	    (loop
		(when (equal i (- (fill-pointer content) 1)) (return))
		(setf (aref content i) (aref content (+ 1 i)))
		(incf i)
	    )
	)
        (decf (fill-pointer content))
    )
)

(defmethod (lis-tab :move-in) (e &key (to-pos nil to-pos-p))
    (unless to-pos-p (setf to-pos 'dont-care))
    (setq to-pos [:normalize-insert-pos to-pos])
    (incf (fill-pointer content))
    (let ((i (- (fill-pointer content) 1)))
      (loop
	  (when (equal i to-pos) (return))
	  (setf (aref content (- i 1)) (aref content i))
	  (decf i)
      )
     ) 
     (setf (aref content to-pos) e)
     to-pos
)

; -------------------- Operations based on indices (page 73)

(defmethod (lis-tab :normalize-remove-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (first     (min (fill-pointer content) 0))
	    (last      (- (fill-pointer content) 1))
	)
	(case (car pos)
	    (at-index  (cadr pos))
	    (after     (+ (cadr pos) 1))
	    (first     (min (fill-pointer content) 0))
	    (last      (- (fill-pointer content) 1))
	)
    )
)

 (defmethod (lis-tab :normalize-insert-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (dont-care (fill-pointer content))
	    (first     0)
	    (last      (fill-pointer content))
	)
	(case (car pos)
	    (before  pos)
	    (after    (+ pos 1))
	    (dont-care (fill-pointer content))
	    (first     0)
	    (last      (fill-pointer content))
	)
    )
)

(defmethod (lis-tab :index-of) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (unless (> 0 pos) pos)
)

; kernel method
(defmethod (lis-tab :set-elem) (i e) (setf (aref content i) e))
(defmethod (lis-tab :get-elem) (i) (aref content i))
(defmethod (lis-tab :transfer-elem-out) (i) (aref content i))
; kernel method
(defmethod (lis-tab :transfer-elem-in) (i e) (setf (aref content i) e))

(defmethod (lis-tab :index-val) (i) (- i 1))
(defmethod (lis-tab :index-pos) (p) (+ p 1))

; -------------------- Iterators (page 80)

(defiterationmethod (lis-tab :iterate-non-recursive-method) 
    (f)
    ; obey iteration-order move part from to

    (if (and move  (equal iteration-order 'from-last))
	; you may modify all elements and delete current
	(progn
	    (if part
		(progn
		    (setq from (car part))
		    (setq to (cadr part))
		)
		(progn
		   (unless from (setq from 0))
		   (unless to (setq to (- (fill-pointer content) 1)))
		)
	    )
	    (let ((ff (fill-pointer content)))
	       (loop
		  (when (> from to) (return))
		  (funcall f (aref content from) from)
		  (if (equal ff (fill-pointer content))
		     (incf from)
		     (progn
			 (decf ff)
			 (decf to)
		     )
		  )
	       )
	    )
	 )
	; not move
	(if (equal iteration-order 'from-last)
	    ; reverse iteration
	    (progn
	       (if part
		   (progn
		       (setq from (cadr part))
		       (setq to (car part))
		   )
		   (progn
		      (unless from (setq from (- (fill-pointer content) 1)))
		      (unless to (setq to 0))
		   )
	       )
	       (loop
		  (when (< from to) (return))
		  (funcall f (aref content from) from)
		  (decf from)
	       )
	    )
	    ; normal iteration
	   (progn
	       (if part
		   (progn
		       (setq from (car part))
		       (setq to (cadr part))
		   )
		   (progn
		      (unless from (setq from 0))
		      (unless to (setq to (- (fill-pointer content) 1)))
		   )
	       )
	       (loop
		  (when (> from to) (return))
		  (funcall f (aref content from) from)
		  (incf from)
	       )
	    )
	)
    )
)

#|
(defmethod (lis-tab :iterate-non-recursive-method) 
    (f &key (iteration-order 'dont-care) &allow-other-keys) 
 (if (eql iteration-order 'from-last)
    (let ((i (fill-pointer content)))
	(loop
	    (when (<= i 0) (return))
	    (decf i)
	    (funcall f (aref content i) i)
	)
    )
    (if move
       (let ((i 0) (ff (fill-pointer content)))
	   (loop 
	       (when (>= i (fill-pointer content)) (return)) 
	       (funcall f (aref content i) i) 
	       (if (equal ff (fill-pointer content))
		   (incf i) 
		   (decf ff)
	       )
	   ) 
       )
       (let ((i 0))
	   (loop
	       (when (>= i (fill-pointer content)) (return))
	       (funcall f (aref content i) i)
	       (incf i)
	   )
       )
    )
  )
)
|#

(compile-flavor-methods lis-tab)
