#|
   module      : map-hash
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 
   
	abstract data type mapping implemented as hash table

   history     : 90.06.26 Version 1
   errors      :
   bugs        : iteration methods
		 don't obey all variants
|#

(mod-require "data/map/v2/map")
(defflavor map-hash (content) (map))

; -------------------- constructors (page 57)
(definitmethod (map-hash) (&key init-value) 
    (setq content (make-hash-table))
    (dolist (pair init-value)
	[:move-in pair]
    )
)

(defmethod (map-hash :clear) () 
    #|
    (iterate
	(el in self)
	[el :clear]
    )
    |#
    (clrhash content)
    self
)

(defmethod (map-hash :move) 
    (from-pos &key to-pos from)
    [self :move-in [from :move-out from-pos]]
)

(defmethod (map-hash :move-in-key) (k i &key to-pos) 
    (setf (gethash k content) i) k)

(defmethod (map-hash :move-out-key) (pos)

    (setq pos (cadr pos))
    (let ((res (gethash pos content)))
       (remhash pos content)
       (values pos res)
    )
)

(defmethod (map-hash :transfer-info-in) (i info)

    (setf (gethash i content) info)
)

(defmethod (map-hash :transfer-info-out) (i) (gethash i content))

; -------------------- Operations based on indices (page 73)
(defmethod (map-hash :index-of) (pos) (cadr pos))
(defmethod (map-hash :get-key) (i) i)
(defmethod (map-hash :get-info) (i) (gethash i content))
(defmethod (map-hash :set-info) (i info) [self :insert-key i info])

; -------------------- Access by Position Count (page 76)

; -------------------- Selector Operations (page 77)

; -------------------- Iterators (page 80)
(defiterationmethod (map-hash :iterate-normal-method) (f)

    ; don't obey parts and startpoints

    (maphash 
	#'(lambda (x y) (funcall f (pair x y) x))
	content
    )
)

#|
(defiterationmethod (map-hash :iterate-key-method) (f)

    (maphash
	#'(lambda (x y) (funcall f x y x))
	content
    )
)
|#

