#|
   module      : set-hash
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 

	ADT set implemented as a hash table

   history     : 90.06.26 Version 1
   errors      :
   bugs        : iteration method does not obey all iteration variants
|#

(defflavor set-hash (content) (set))

; -------------------- constructors (page 57)
(definitmethod (set-hash) (&key init-value) 
    (setq content (make-hash-table))
    (dolist (e init-value)
	[:move-in e]
    )
)

(defmethod (set-hash :clear) () (clrhash content) self)

(defmethod (set-hash :move) (from-pos &key to-pos from) 
    [self :move-in [from :move-out from-pos]]
)

(defmethod (set-hash :move-in) (e  &key to-pos) (setf (gethash e content) t) e)

(defmethod (set-hash :move-out) (pos) 
    (setq pos (if (listp pos) (cadr pos) pos))
    (remhash pos content) 
    pos)

(defmethod (set-hash :index-of) (i) i)
(defmethod (set-hash :get-elem) (e) (if (listp e) (cadr e) e))

; -------------------- Iterators (page 80)
(defmethod (set-hash :iterate-method) (f)
    (maphash #'(lambda (x y) (funcall f x x)) content))
