#|
   module      : lazy-lis
   owner       : Peter Sanders
   author      : Peter Sanders
   explanation : 

	concrete implementation of lis by lazy lists

   history     : 2/13/91 created
   errors      :
   bugs        : parts not implemented
		 normalize-remove/insert-pos
		 should be functions instead of methods !!
		 iterate method does not obey all variants
|#

(mod-require "data/lis/v2/lis")

(defflavor lazy-list (content) (lis))

; --------------------

; kernel method ???
(definitmethod (lis-list) (&key init-value) (setf content nil)
    [:insert-s-exp init-value]
)
; kernel method ???
(defmethod (lis-list :insert-s-exp) (e) 
    (setf content (append content (copy-list e))))
(defmethod (lis-list :move-in-s-exp) (e) (setf content (append content e)))
; kernel method ???
(defmethod (lis-list :clear) () (setf content nil) self)

; kernel method ???
(defmethod (lis-list :slice) (sub) 
    (setf content (car sub))
    (setf (car (cdr sub)) nil)
)
;???? geht  so nicht
(defmethod (lis-list :move) 
    (from-pos &key (to-pos nil to-pos-p) (from nil from-p) &aux help)

    (unless to-pos-p (setf to-pos '(dont-care)))
    (unless from-p (setf from self))
    (setf help [from :move-out from-pos])
    [:move-in help :to-pos to-pos]
)

; kernel method ???
(defmethod (lis-list :move-out) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos
	 (prog1 (cadr pos) (setf (cdr pos) (cddr pos)))
	 (prog1 (car content) (setf content (cdr content)))
    )
)

; kernel method ???
(defmethod (lis-list :move-in) (e &key (to-pos nil to-pos-p))
    (unless to-pos-p (setf to-pos 'dont-care))
    (setq to-pos [:normalize-insert-pos to-pos])
    (if to-pos
	 (setf (cdr to-pos) (cons e (cdr to-pos)))
	 (setf content (cons e content))
    )
)

; -------------------- Operations based on indices (page 73)

;??????????????????????????????????????????????
(defmethod (lis-list :normalize-remove-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (first     nil)
	    (last      [:previous [:last]])
	)
	(case (car pos)
	    (at-index [:previous (cadr pos)])
	    (after    pos)
	    (first     nil)
	    (last      [:previous [:last]])
	)
    )
)

;???????????????????????????????????????????????
(defmethod (lis-list :normalize-insert-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (dont-care nil)
	    (first     nil)
	    (last      [:last])
	)
	(case (car pos)
	    (before [:previous (cadr pos)])
	    (after    pos)
	    (dont-care nil)
	    (first     nil)
	    (last      [:last])
	)
    )
)

;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
(defmethod (lis-list :index-of) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos (cdr pos) content))

; kernel method
;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
(defmethod (lis-list :set-elem) (i e) (setf (car i) e))
(defmethod (lis-list :get-elem) (i) (car i))
(defmethod (lis-list :transfer-elem-out) (i) (car i))
; kernel method ?????
(defmethod (lis-list :transfer-elem-in) (i e) (setf (car i) e))

; -------------------- Iterators (page 80)

;??????????!!!!!!!!!!!
(defiterationmethod (lis-list :iterate-non-recursive-method) 
    (f)
    ; obey iteration-order move part from to 

     (if (eql iteration-order 'from-last)
	(progn
	   (if part
	       (progn
		   (setq from (caddr part))
		   (setq to (cadr part))
	       )
	       (when from (setq from (cdr from)))
	   )
	   (labels
	     ((rec (index)
		  (unless (eql index from) (rec (cdr index)))
		  (funcall (car index) (index))
	      )
	     )
	      (if to (rec to) (rec content))
	   )
	)
	(if move
	   (let (( ff content))
	       (when ff
		   (let ((n (cdr ff)))
		       (loop
			   (funcall f (car ff) ff)
			   (unless n (return))
			   (setq ff n)
			   (setq n (cdr ff))
		       )
		   )
	       )
	   )
	   (mapl #'(lambda (l) (funcall f (car l) l)) content)
	)
     )
)

; -------------------- Parts (page 130)

(compile-flavor-methods lis-list)




