 #|
   module      : flavor
   owner       : Franz Weber
   author      : Franz Weber
   explanation : flavor conventions

	1. short terms for send

	   a) [object :message params ...]

	      means

	      (send object :message params ...)

           b) [:message params ...]

	      means

	      [self :message params ...]
	
	2. special method declarations

	   a) (virtualmethod <method description> <params>)

	      declares a method which has to be implemented by
	      descendant classes. This enables 'abstract classes'.

	   b) (hidemethod <method description>)

	      hides a method for a class (to forbid certain inherited
	      messages

           c) (definitmethod (class [<method combinator]) <params> <forms>)
	      
	      allows the definition of a method, which is called
	      at the creation of an object. Note, that only keyword
	      parameters are allowed, which where declared as
	      'init-keywords' in the flavor definition.

	3. flavor macros

	   (defflavormacro :name objectname params forms)

	   declares a macro, which may be called as

	   [object :name ...] or
	   [:name           ]

	   whereby the lexical variable objectname is bound to the object
	   implied by the semantic of []

   history     :
  
	90.06.12
	90.07.12 short form for self introduced
		 [:f ...] means [self :f ... ]
  
   bugs        :
	
	short terms don't work in debug mode
   errors      :
|# 

(export 
    '(definitmethod 
      virtualmethod 
      hidemethod 
      *flavor-macros* 
      defflavormacro
      undefflavormacro
     )
)

; -------------------- special method declarations
(defmacro definitmethod (message lambda-list &rest forms)
    `(defmethod 
        ,(append message (list :init))
	(args)
	(destructuring-bind
	    ,lambda-list
	    args
	    ,@forms
	)
    )
)

(defmacro virtualmethod (&rest dummy) nil)

(defmacro hidemethod  (message)
    `(defmethod ,(list (first message) :before (car (last message))) (&rest args)
        (error "method may not be implemented by descendant class")
    )
)

(defvar *flavor-macros* nil)

; -------------------- short terms for send
(set-macro-character
    #\[
    #'(lambda (s c &aux l m)
	  (setq l nil)
	  (loop
	      (setq m (read s nil (values) t))
	      (when (eql m #\]) (return))
	      (setf l (append l (list m)))
	  )
	  (setq l (if (keywordp (car l)) (cons 'self l) l))
	  (if (member (cadr l) *flavor-macros*)
	      (cons (cadr l) (cons (car l) (cddr l)))
	      (cons 'send l)
	  )
      )
      nil
)

(set-macro-character
    #\]
    #'(lambda (s c &aux l)
	  #\]
      )
      nil
)

; -------------------- flavor macros

(defmacro defflavormacro (name objectname params &rest forms)

    (setq *flavor-macros* (cons name *flavor-macros*))
    `(defmacro ,name ,(cons objectname params) ,@forms)
)

(defmacro undefflavormacro (name) 
    (setq *flavor-macros* (delete name *flavor-macros*))
    nil
)
