#|
   module      : e-unification
   author      : Olaf Mueller
   explanation : e-unification method for formulas
                 with general K-Imitation and searchtree pruning for C and AC 
                 
   history     : 29.3.93 create
   owner       : Olaf Mueller
                 5.4.93 all examples work
    
   errors      :
  
|# 

; Modules to be imported
; ADT's
(mod-require "data/lis/v2/lis-list")
(mod-require "data/lis/v2/lis-list-readonly")
(mod-require "data/lis/v2/lis-lazy")
(mod-require "data/queue/v2/queue-list")
(mod-require "data/glo/v2/simple-aggregate-empty")

; logic modules
(mod-require "logic/v2/substitution/substitution-hash")
(mod-require "logic/v2/expression")

; Unification method for expression
(defmethod (expression-node :E-unify) 
  (e2 environment &key (substitutions (make-instance 'substitution-hash)) 
	               (patience 5000)
                  &allow-other-keys)
    (setq *uni-pic-count* 0)
    (setq *Neu-count* 0)     ; neue AC-Variablen
    (setq *NV-count* 0)      ; neue Lifting-Variablen
    (setq *count-I* 0)
    (setq *count-P* 0)
    (setq *count-KI* 0)
    (setq *count-S* 0)
    (setq *count-SS* 0)
    (setq *count-triv-S* 0)
    (setq *count-ho-S* 0)
    (setq *count-deadends* 0)
    (setq *count-nodes* 0)
    (make-instance 'lis-lazy :init-value 
		 (uni-unify (uni-init-searchtree self e2 environment substitutions patience))
    )
)

(defmethod (expression-node :time-unify) (e2 env)
    (time  (progn (setq res [ :E-unify e2 env]) nil))
   (format t "~%~%--------------------------")
   (format t "~% Auswertung mit KI (immer)~%")
   (format t "--------------------------~%~%")
   (format t "Anzahl Lsg.:          ~A~%" [res :count])
   (format t "Anzahl toter Aeste:   ~A~%" *count-deadends*)
   (format t "Knoten:          ~A~%" *count-nodes*)
   (format t "Imitationen:     ~A~%" *count-I*)
   (format t "K-Imitationen:   ~A~%" *count-KI*)
   (format t "Projektionen:    ~A~%" *count-P*)
   (format t "Vereinfachungen: ~A~%" *count-S*)
   (format t "Triv. Vereinfachungen (Gr. 2): ~A~%In Prozent aller Vereinfachungen: ~A~%"
       *count-triv-S* 
       (if (> *count-S* 0) (float (/ *count-triv-S* *count-S*)) nil))
   (format t "Durchschnittliche Vereinf. Groesse: ~A~%" 
       (if (> *count-S* 0) (float (/ *count-SS* *count-S*)) nil))
   (format t "ho-Vereinfachungen: ~A~%" 
       (if (> *count-S* 0) (float (/ *count-ho-S* *count-S*)) nil))
)


;(format t "~%Anzahl der durchlaufenen Knoten: ~A~%" count-nodes)
; Function traversing the searchtree. Output is a lisp-lazy-list of preunifiers
(defun uni-unify (tree)
  (do (node dis sub all-sub (pat [tree :patience] (- pat 1)))
      ([tree :is-empty] nil); While there are nodes not yet visited
      (if (= pat 0)(throw 'uni-patience nil))
      (setq node      [tree :pick-node]           ; Pick one of these nodes
	    dis       [node :disagreements]       ; And look at its
	    sub       [node :new-sub]             ; components
            all-subs  [node :all-subs])
      (setq *count-nodes* (1+ *count-nodes*))
      [all-subs :insert-aggregate sub] ; Substitution application that has been postponed
      ; Normalize all disagreemnent pairs
      ;(format t "dis: ~A~%" dis)
      [dis :iterate (x) [[[x :apply-sub sub] :beta-reduce] :toplevel-eta-expand]]
      ;[[node :copy] :print-node]
      (let* ((p [dis :get-rigid-rigid])
            (q-idx [dis :get-flexible-rigid-idx])
            (q (if q-idx [dis :get-elem q-idx] nil))
            (new-diss (make-instance 'lis-list))
            (subs (make-instance 'lis-list))
           )
        (if (or p q)
          (if p
              ; ---------- Vereinfachung -----------------------
              (progn
                 (setq new-diss [dis :simplify [tree :env]])
                 (if new-diss
                    (progn
                      (setq *count-S* (+ [new-diss :count] *count-S*))
                      [tree :insert-new-nodes node new-diss subs]
                    )
                    (setq *count-deadends* (1+ *count-deadends*))
                 )
              )
              (progn
                 ; ----------- Projektion -------------------------
                 (setq new-diss (make-instance 'lis-list))
                 (setq subs [q :project [tree :env]])
                 (setq *count-P* (+ [subs :count] *count-P*))
                 [tree :insert-new-nodes node new-diss subs]
                 ; ----------- Imitation --------------------------
                 (if (typep [q :right |:head |:ref] 'c-declaration)
                    (progn
                        [dis :move-out `(at-index ,q-idx)]
                        (setq *count-KI* (1+ *count-KI*))
                        [tree :K-imitate-for-C node q]
                    )
                    (if (typep [q :right |:head |:ref] 'ac-declaration)
                        (progn
                            [dis :move-out `(at-index ,q-idx)]
                            (setq *count-KI* (+ 4 *count-KI*))
                            [tree :K-imitate-for-AC node q]
                        )
                        (progn
                            [dis :move-out `(at-index ,q-idx)]
                            (setq subs [q :imitate [tree :env]])
                            (if (not [subs :is-empty])
                             (progn
                               (setq *count-I* (1+ *count-I*))
                               [tree :K-imitate node q subs]
                              )
                            )
                        )
                    )
                 )
              )   ; *progn*
          )
          (return (lazy-cons                                ; Create a lisp-lazy-list 
		     (let ((us (make-instance 'uni-sub)))   ; whose car is the first solution
		       [all-subs :iterate ((key info))      ; consisting of the accumulated subs
			    [us :move-in-key key info]]
		       [us :set-flex-flex dis]              ; and the remaining flex-flex pairs.
		       us)
		     (uni-unify tree) ; the cdr ist the promise to compute all solutions
                   )
          )
       ) ; *if*
      )  ;*let*    
   )     ;*do*
)

;--------------------------------------------------------------------------------
;         Substitutions with preunifier-components
;--------------------------------------------------------------------------------

;flex-flex is a disagreement-set containing only flexible-flexible-pairs
(defflavor uni-sub (flex-flex) (substitution-hash)
  (:gettable-instance-variables flex-flex)
  (:settable-instance-variables flex-flex)
  (:initable-instance-variables flex-flex))
  
;--------------------------------------------------------------------------------
;                   searchtree
;--------------------------------------------------------------------------------

; This ADT hides the representation and search strategy of Huet's unification
; search tree
; Additionally it hides the environment and patience

(defflavor searchtree (env patience) (queue-list) ; breadth first search strategy
  (:gettable-instance-variables env patience)
  (:settable-instance-variables env patience)
  (:initable-instance-variables env patience))

; Build an initial searchtree from two expressions, an environment and an initial substitution.
(defun uni-init-searchtree (e1 e2 en substitution patience)
  (let* ((left  e1)  
	 (right e2)
	 (pair  (make-instance 'disagreement-pair :left  left 
			                          :right right))
	 (set   (make-instance 'disagreement-set :init-value `(,pair)))
	 (mtsub (make-instance 'substitution-hash))
	 (node  (make-instance 'searchtree-node :disagreements set
		                                :new-sub       substitution
		                                :all-subs      mtsub
                                                :depth         0))
	 (tree  (make-instance 'searchtree :env en :patience patience)))
    [tree :push node]
    tree))

 
; Give the next searchtree-node to be visited
(defmethod (searchtree :pick-node)() [:dequeue])


; Build new searchtreenodes into the tree after both match and simplify
(defmethod (searchtree :insert-new-nodes)(node new-dissets subs)
  (if [new-dissets :is-empty]
    ; case match
    [subs :iterate (x) 
        (let ((new-node [node :copy]))
	  [new-node :set-new-sub x]
          [new-node :set-depth (1+ [node :depth])]
	  [:push new-node]
        )
    ]
    ; case simplify
    [new-dissets :iterate (x)
        (let ((new-node [node :copy]))
          [new-node :set-new-sub (make-instance 'substitution-hash)]
          [new-node :set-disagreements x]
          [new-node :set-depth (1+ [node :depth])]
          [:push new-node]
        )
    ]
  ) 
)

(defmethod (searchtree :K-imitate) (node q subs)
  (let* ((new-node [node :copy])
         (dis [new-node :disagreements])
         (sub [subs :move-out 'first])
         (pair [q :apply-sub sub | :beta-reduce])
         (l [pair :left])
         (r [pair :right])
        )
     (multiple-iterate ((a [l :arguments])   
	                (b [r :arguments]))
	      [dis :move-in (make-instance 'disagreement-pair 
                          :left  [(assemble-exp-binder [l :binder] a) :toplevel-eta-expand] 
			  :right [(assemble-exp-binder [r :binder] b) :toplevel-eta-expand])
              ]
     )
     ;(format t "dis: ~A~%" dis)
     [new-node :set-new-sub sub]
     [new-node :set-depth (1+ [node :depth])]
     [:push new-node]
  )
)

(defmethod (searchtree :K-imitate-for-C) (node q)
  (let* ((new-node [node :copy])
         (dis [new-node :disagreements])
         (sub-in-list [q :imitate [:env]])
         (sub [sub-in-list :move-out 'first])
         (pair [q :apply-sub sub | :beta-reduce])
         (l [pair :left])
         (r [pair :right])
        )
     (multiple-iterate ((a [l :arguments])   
	                (b [r :arguments]))
	      [dis :move-in (make-instance 'disagreement-pair 
                          :left  [(assemble-exp-binder [l :binder] a) :toplevel-eta-expand] 
			  :right [(assemble-exp-binder [r :binder] b) :toplevel-eta-expand])
              ]
     )
     ;(format t "dis: ~A~%" dis)
     [new-node :set-new-sub sub]
     [new-node :set-depth (1+ [node :depth])]
     [:push new-node]
  )
)

(defmethod (searchtree :K-imitate-for-AC) (node q)
  (let* ((sub-in-list [q :imitate [:env]])
         (sub [sub-in-list :move-out 'first])
         (new-node1 [node :copy])
         (new-node2 [node :copy])
         (new-node3 [node :copy])
         (new-node4 [node :copy])
         (dis1 [new-node1 :disagreements])
         (dis2 [new-node2 :disagreements])
         (dis3 [new-node3 :disagreements])
         (dis4 [new-node4 :disagreements])
         (pair [q :apply-sub sub | :beta-reduce])
         (l [pair :left])
         (r [pair :right])
         (arg1 [l :arguments])
         (arg2 [r :arguments])
         (b1   [l :binder])
         (b2   [r :binder])
         (x1 [arg1 :get-by-pos 1])
         (x2 [arg1 :get-by-pos 2])
         (y1 [arg2 :get-by-pos 1])
         (y2 [arg2 :get-by-pos 2])
         (ac-decl [q :left |:head |:ref])
         (var-type [ac-decl :type |:range |:domain])
         (NV1 [env :sub2-variable var-type "Neu"]) 
         (NV2 [env :sub2-variable var-type "Neu"])
         (NV3 [env :sub2-variable var-type "Neu"])
         (NV4 [env :sub2-variable var-type "Neu"])
        )
     (multiple-iterate ((a arg1)   
	                (b arg2))
	      [dis1 :move-in (make-instance 'disagreement-pair 
                          :left  [(assemble-exp-binder [l :binder] a) :toplevel-eta-expand] 
			  :right [(assemble-exp-binder [r :binder] b) :toplevel-eta-expand])
              ]
     )  
     [dis2 :move-in (make-instance 'disagreement-pair
               :left  (assemble-exp-binder b1 x1)
               :right (assemble-exp-binder b2 [ac-decl :appl-build NV2 y2]))]
     [dis2 :move-in (make-instance 'disagreement-pair
               :left (assemble-exp-binder b1 [ac-decl :appl-build x2 NV2])
               :right (assemble-exp-binder b2 y1))]
     [dis3 :move-in (make-instance 'disagreement-pair
               :left (assemble-exp-binder b1 x2)
               :right  (assemble-exp-binder b2 [ac-decl :appl-build NV4 y2]))]
     [dis3 :move-in (make-instance 'disagreement-pair
               :left  (assemble-exp-binder b1 [ac-decl :appl-build NV4 x1])
               :right  (assemble-exp-binder b2 y1))]
     [dis4 :move-in (make-instance 'disagreement-pair
               :left   (assemble-exp-binder b1 x1)
               :right  (assemble-exp-binder b2 [ac-decl :appl-build NV2 NV1]))]
     [dis4 :move-in (make-instance 'disagreement-pair
               :left   (assemble-exp-binder b1 x2)
               :right  (assemble-exp-binder b2 [ac-decl :appl-build NV4 NV3]))]
     [dis4 :move-in (make-instance 'disagreement-pair
               :left   (assemble-exp-binder b1 [ac-decl :appl-build NV4 NV2])
               :right  (assemble-exp-binder b2 y1))]
     [dis4 :move-in (make-instance 'disagreement-pair
               :left   (assemble-exp-binder b1 [ac-decl :appl-build NV3 NV1])
               :right  (assemble-exp-binder b2 y2))]
     [new-node1 :set-new-sub sub]
     [new-node2 :set-new-sub sub]
     [new-node3 :set-new-sub sub]
     [new-node4 :set-new-sub sub]
     [new-node1 :set-depth (1+ [node :depth])]
     [new-node2 :set-depth (1+ [node :depth])]
     [new-node3 :set-depth (1+ [node :depth])]
     [new-node4 :set-depth (1+ [node :depth])]
     [:push new-node1]
     [:push new-node2]
     [:push new-node3]
     [:push new-node4]
  )
)

  
;--------------------------------------------------------------------------------
;              searchtree-node
;--------------------------------------------------------------------------------

; simple structure  storing a disagreement-set, 
; a substitution to be applied to the disagreements
; and a substitution caching the substitutions found so far

(defflavor searchtree-node (disagreements new-sub all-subs depth)(simple-aggregate-empty simple-aggregate)
  :settable-instance-variables
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (searchtree-node :print-node) ()
   (format t "~%~%Knoten: ~A~%" *count-nodes*)
   [disagreements :iterate (pair)
       [pair :left |:unparse]
       (format t " =?= ")
       [pair :right |:unparse]
       (format t" , ")
   ]
   (format t "Tiefe: ~A" depth)
   (print-real-sub new-sub)
)

; auxiliary function for generic implementation of :print-self
(defmethod (searchtree-node :attribute-lis)()
  (make-instance 'lis-list :init-value `(,disagreements ,new-sub ,all-subs)))

; A :copy of a searchtree-node is really only a copy of new-sub and all-subs
(defmethod (searchtree-node :copy)()
  (make-instance 'searchtree-node :disagreements [disagreements :copy]
		                  :new-sub       [new-sub :copy]
				  :all-subs      [all-subs :copy]
                                  :depth         depth))

;--------------------------------------------------------------------------------
;                    disagreement-pairs
;--------------------------------------------------------------------------------

; Pair of uni-exp's that has to be unified
(defflavor disagreement-pair (left right) (simple-aggregate-empty simple-aggregate)
  (:gettable-instance-variables left right)  
  (:settable-instance-variables left right)
  (:initable-instance-variables left right))

; auxiliary function for generic implementation of :print-self
(defmethod (disagreement-pair :attribute-lis)()
  (make-instance 'lis-list :init-value `(,left ,right)))

; extend some expression-methods to both components
(defmethod (disagreement-pair :apply-sub) (sub)
  (setq left  [left  :apply-sub sub])
  (setq right [right :apply-sub sub])
  self)

(defmethod (disagreement-pair :copy)()
  (make-instance 'disagreement-pair :left  left
		                    :right right))

(defmethod (disagreement-pair :beta-reduce) ()
  (setq left  [left  :beta-reduce])
  (setq right [right :beta-reduce])
  self)

(defmethod (disagreement-pair :toplevel-eta-expand) ()
  (setq left  [left  :toplevel-eta-expand])
  (setq right [right :toplevel-eta-expand])
  self)


; A component of a disagreement-pair is flexible iff it's head is a free variable
; else it is called rigid.
; Therefore there are four types of disagreement-pairs:
; rigid-rigid
; (rigid-flexible)
; flexible-rigid
; flexible-flexible
; If :get-type finds a rigid-flexible pair it normalizes it to
; a flexible-rigid pair and returns flexible-rigid
; both expression has to be beta-normalized
(defmethod (disagreement-pair :get-type)()
  (if [left :is-rigid]
      (if [right :is-rigid]
          :rigid-rigid
	  (progn (psetq left  right
			right left)
		 :flexible-rigid))
      (if [right :is-rigid]
	  :flexible-rigid
	  :flexible-flexible)))

; project returns of a set of substitutions determined by projection
; The env argument specifies an environment where
; new variables are made public
(defmethod (disagreement-pair :project) (env)
  (let* ((subs (make-instance 'set-lis)) ; Substitution to be computed. 
	 (e1 [:left])                    ; bind some parts of the pair
	 (e2 [:right])
	 (a1 [e1 :arguments])
	 (a2 [e2 :arguments])
	 (h1 [e1 :head])
	 (h2 [e2 :head])
	 (beta [[e1 :expression-type] :output-type]) ; Result type of e1 and e2
	 ; some easy parts of the output substitutions
         ; binder for all produced substitutions
	 (binder  [a1 :map :iteration-order 'from-last (x) (uni-new-abstraction x)])
         ; remember: binders are stored in reverse order!
         (rev-binder [binder :reduce (x) 
                             (p (make-instance 'lis-list-readonly))
                             [p :cons x]])
         ; Arguments for new Variables
	 (vars    [binder :map :iteration-order 'from-last (x) (make-instance 'reference :ref x)])
         ; binder is inverse, not so arguments.
        )
       ;----------project--------------------------------------------------
    ; inspect all arguments of e1 for fitting type
    (multiple-iterate ((e1i a1)
                       (z rev-binder))
      (if [beta :equal [[e1i :expression-type] :output-type]]
	  (let ((itypes [[e1i :expression-type] :argument-types]))
	    [subs :move-in 
	          (make-instance 'substitution-hash :init-value `(( 
		     ,[h1 :ref]
		     ,(assemble-exp
		        binder
			(make-instance 'reference :ref z)
  		        [itypes :map (x) 
		          (assemble-exp-head 
                             [env :new-variable 
                                  (assemble-type 
                                     x  
                                     [h1 :expression-type |:argument-types]
                                  )
                                  "I"
                             ]
			     vars
                          )
                        ]
                      )
                   )))
            ]
          )
       )
    )
    subs)) 


; imitate returns all substitutions found by imitation
; The env argument specifies an environment where
; new variables are made public
(defmethod (disagreement-pair :imitate) (env)
  (let* ((subs (make-instance 'set-lis)) ; Substitution to be computed. 
	 (e1 [:left])                    ; bind some parts of the pair
	 (e2 [:right])
	 (a1 [e1 :arguments])
	 (a2 [e2 :arguments])
	 (h1 [e1 :head])
	 (h2 [e2 :head])
	 (beta [[e1 :expression-type] :output-type]) ; Result type of e1 and e2
	 ; some easy parts of the output substitutions
         ; binder for all produced substitutions
	 (binder  [a1 :map :iteration-order 'from-last (x) (uni-new-abstraction x)])
         ; remember: binders are stored in reverse order!
         (rev-binder [binder :reduce (x) 
                             (p (make-instance 'lis-list-readonly))
                             [p :cons x]])
         ; Arguments for new Variables
	 (vars    [binder :map :iteration-order 'from-last (x) (make-instance 'reference :ref x)])
         ; binder is inverse, not so arguments.
         (decl2 [h2 :ref])
         (head-list (make-instance 'lis-list))
        )
    ;----------imitate--------------------------------------------------
     [[decl2 :imitation-decls env] :iterate (decl-i)
            [head-list :move-in (make-instance 'reference :ref decl-i)]
     ]
     [head-list :iterate (hi)
       (if (not [[e2 :binder] :find [hi :ref]]) ; is this check necessary for all hi??
	  [subs :move-in 
	        (make-instance 'substitution-hash :init-value `((
	            ,[h1 :ref]
		    ,(assemble-exp
		       binder
		       hi
		       [a2 :map (x) 
		           (assemble-exp-head 
                              [env :new-variable 
                                   (assemble-type 
                                       [x :expression-type] 
                                       [h1 :expression-type |:argument-types]
                                   )
	                           "G"
                              ]
			      vars
                            )
                       ]
                     )
                  )))
           ]
       )
     ]
     subs
  )
) 

;------------------------------------------------------------------------------
;              disagreement-sets
;------------------------------------------------------------------------------

; set of disagreement pairs
; the following line is efficient but wrong
(defflavor disagreement-set () (lis-list))
;(defflavor disagreement-set () (set-list))

; copy a disagreement set keeping the pairs untouched
(defmethod (disagreement-set :copy)()
  (let ((c (make-instance 'disagreement-set)))
    [:iterate (x) [c :move-in [x :copy]]]
    c))
 

; Apply Substitution sub to each element of self
(defmethod (disagreement-set :apply-sub)(sub)
  [:iterate (el) 
      [el :apply-sub sub] 
      ; Note: apply-sub for pairs is modifying
  ]
)

; Remove one pair of the specified type
; Heuristics for choosing pairs intelligently are hidden here
(defmethod (disagreement-set :pick-rigid-rigid)()
  (let ((x [:find-property (el) (eq [el :get-type] :rigid-rigid)]))
    (if x [:move-out `(at-index ,x)] nil))) 
(defmethod (disagreement-set :get-rigid-rigid)()
  (let ((x [:find-property (el) (eq [el :get-type] :rigid-rigid)]))
    (if x [:get-elem x] nil))) 
(defmethod (disagreement-set :pick-flexible-rigid)()
  (let ((x [:find-property (el) (eq [el :get-type] :flexible-rigid)]))
    (if x [:move-out `(at-index ,x)] nil)))
; same as pick but don't remove the pair
(defmethod (disagreement-set :get-flexible-rigid)()
  (let ((x [:find-property (el) (eq [el :get-type] :flexible-rigid)]))
    (if x [:get-elem x] nil)))
(defmethod (disagreement-set :pick-flexible-flexible)()
  (let ((x [:find-property (el) (eq [el :get-type] :flexible-flexible)]))
    (if x [:move-out `(at-index ,x)] nil)))

(defmethod (disagreement-set :get-flexible-rigid-idx)()
  (let ((x [:find-property (el) (eq [el :get-type] :flexible-rigid)]))
    x))

; --------------------------------------------------------------------
;      Theory-dependent table of suitable declarations for imitation
; --------------------------------------------------------------------

; returns all declarations that are compatible to self in a list
; in order to use references to them as possible heads in the imitation rule.
; In normal declarations only self is able to produce a suitable head for
; imitation. For all root-preserving operators (C, AC, ...) thats the same.
; Therefore in this cases this method need not to be overwritten. 
(defmethod (declaration :imitation-decls ) (env)
  (let  ((result-list (make-instance 'lis-list))
        )
    [result-list :move-in self]
    result-list
  )
)
    


; --------------------------------------------------------------------
;                      Output - Utilities
; --------------------------------------------------------------------

(defun print-sub (s)

    [s :iterate ((var exp))
        (format t "~a  ->  " [var :picture])
        [exp :beta-eta-reduce |:unparse]
        (format t "~%")
    ]
    (format t "- flex~%")
    [s :flex-flex |:iterate (pair)
        [pair :left  |:unparse]
        (format t "  ->  ")
        [pair :right |:beta-eta-reduce |:unparse]
        (format t "~%")
 
    ]
)


(defun print-real-sub (s)
    (if (not [s :is-empty]) (format t "~%"))
    [s :iterate ((var exp))
        (format t "~a  ->  " [var :picture])
        [exp :beta-eta-reduce |:unparse]
        (format t " , ")
    ]
)

(defun print-subs (l)
    [l :iterate (x)
        (format t "----~%")
        (print-sub x)

    ]
)

(defun print-expand (s sub)

    [s :iterate ((var exp))
        (format t "~a  ->  " [var :picture])
        [exp :apply-sub sub |:beta-eta-reduce |:unparse]
        (format t "~%")
    ]
    (format t "- flex~%")
    [s :flex-flex |:iterate (pair)
        [pair :left  |:unparse]
        (format t "  ->  ")
        [pair :right |:apply-sub sub |:beta-eta-reduce |:unparse]
        (format t "~%")
 
    ]
)

(defun print-expand-s (l)
    [l :iterate (x)
        (format t "----~%")
        (print-expand x x)

    ]
)


; ---------------------------------------------------------------------------------------
;                                      Methode Simplify
; ---------------------------------------------------------------------------------------

(defmethod (disagreement-set :print-simp) ()
 (let ((mfus  [*PMAS* :reduce (pair) (h 0)
                (+ h [pair :right |:count-leafs] 
                     [pair :left |:count-leafs])
              ])
       (set  [ :reduce (pair) (h 0)
               (+ h [pair :right |:count-leafs] 
                    [pair :left |:count-leafs])
             ])
      )
  ;(format t "~%Vereinfachung:~%")
  ;(format t "Anzahl der mfu's: ~A~%" [*PMAS* :count])
  ;(format t "Einzelne Groessen: ")
  ; [*PMAS* :iterate (pair)
  ;   (format t "~A, " [pair :right |:count-leafs])
 ;  ]
 ; (format t "~%Anzahl der uebergebenen Paare: ~A~%" [:count])
 ; (format t "Groesse dieses Sets: ~A" set)
  (if (< set 3)  (setq *count-triv-S* (1+ *count-triv-S*)))
  (setq *count-SS* (+ set *count-SS*))
  (if (> mfus set) (setq *count-ho-S* (1+ *count-ho-S*)))
 )
)


(virtualmethod (expression-node :count-leafs) () )

(defmethod (application :count-leafs) () 
   (+ [functor :count-leafs] [argument :count-leafs])
)

(defmethod (abstraction :count-leafs) () 
   (1+ [expression-node :count-leafs])
)

(defmethod (reference :count-leafs) () 1)


(defmethod (disagreement-set :simplify) (env)
(block simplify
  (let ((S1 self)
        (S2 (make-instance 'disagreement-set))    ; dis-set fuer erste Stufe
        (S-print (make-instance 'disagreement-set))    ; S2-copy fuer Ausgabe
        (sigma-list (make-instance 'lis-list))    ; Liste von Unifikatoren erster Stufe
        (res (make-instance 'lis-list))           ; Liste der neuen dis-sets
        (H-env (make-instance 'environment-lis))
        (l-env (make-instance 'environment-lis))
       )
     (setq *exp-identity* (make-instance 'key-stack-list))
     (setq *pmas* (make-instance 'lis-list))
     (setq S2 [S1 :abstract H-env]) 
     ;(format t "S2: ~A~%" S2)
     (setq *H-contexts* (make-instance 'key-stack-list))
     (setq *lambda-list* (make-instance 'lambda-list))
     (setq *converter* (make-instance 'converter))
     [S2 :translate l-env]
     (setq S-print [S2 :copy])
     ;(format t "S2: ~A~%" S2)
     (setq sigma-list [S2 :poormans-AC (make-instance 'substitution-hash) l-env])
     ;(format t "sigs: ~A~%" sigma-list)
     (setq *organizer* (make-instance 'new-variable-organizer))  ; @@or sigma-dependent?
     (if (not sigma-list) (return-from simplify nil))  ; no unifiers at all
     [S-print :print-simp]
     ;(format t "sigs-before: ~A~%" sigma-list)
     (if (> [sigma-list :count] 0) [sigma-list :eliminate-redundant-subs env])
     ;(format t "sigs-after: ~A~%" sigma-list)
     [sigma-list :iterate (sigma)
        (let ((S-hilf (make-instance 'disagreement-set))
             )
          [*pmas* :iterate (pair)
            (let* ((H         [pair :left])
                   (H-decl    [H :ref])
                   (binder    [*H-contexts* :get H-decl :test 'eql])
                   (abs-list  (make-instance 'lis-list))
                   (Y         (make-instance 'algebraic-reference 
                                  :ref [*converter* :convert-variable H-decl l-env]))
                   (Y-sub     [Y :apply-sub sigma])
                   (Y-sub-re  (make-instance 'expression-node))
                   (erg-term  (make-instance 'expression-node))
                  )
              [binder :iterate (x) 
                (let ((abs (make-instance 'abstraction 
                                 :type [x :type]
                                 :picture [x :picture]
                                 :is-constant t))
                     )
                  [*exp-identity* :push-key x abs]
                  [abs-list :move-in abs :to-pos 'last]
                )
              ]
              (setq Y-sub-re  [Y-sub :retranslate binder env])
              ;(format t "Y-sub-re: ~A~%" Y-sub-re)
              (setq erg-term 
                 [abs-list :reduce :iteration-order 'from-last (x) (p Y-sub-re)
                        [x :set-expression-node p]
                 ]
              )     
              [abs-list :iterate (x) [*exp-identity* :pop]]  
              [S-hilf :move-in (make-instance 'disagreement-pair
                                     :left  erg-term
                                     :right [pair :right])
              ]
            )  
          ] 
          [S1 :iterate (pair)              ; alle flex-flex Paare von S1 hinzunehmen
               [S-hilf :move-in pair]
          ]   
          [res :move-in S-hilf]
        )
     ]
    ; (format t "~%Zeit:   ")
     res   
  )
)
)


; ---------------------------------------------------------------------------------------
;                                      Methode Abstract
; ---------------------------------------------------------------------------------------

(defmethod (disagreement-set :abstract) (H-env)
  (let ((S1 self)
        (S2 (make-instance 'disagreement-set)) 
       )     
    (setq *Hi-count* 0)    ; Zaehler der Hi innerhalb von ganz Simplify  
    [S1 :iterate (pair idx)
       (if (eq [pair :get-type] :rigid-rigid)
              ; (eq [pair :get-type] :flexible-rigid)
          ; )    
         (progn
           [S1 :move-out `(at-index ,idx)]
           [S2 :move-in [pair :replace-alien-subterms H-env]]
         )
       )
     ]
     S2
  )
)

; --------------------------------------------------------------------------
;                          Methode replace-alien-subterms
; --------------------------------------------------------------------------
; method is only allowed for beta-reduced terms

(defmethod (disagreement-pair :replace-alien-subterms) (env)
  (setq left  [left  :replace-alien-subterms (make-instance 'lis-list) env])
  (setq right [right :replace-alien-subterms (make-instance 'lis-list) env])
  self
)

(defmethod (reference :eta-expand) ()
   [ :toplevel-eta-expand]
)

(defmethod (application :eta-expand) ()
   (make-instance 'application
         :functor [functor :toplevel-eta-expand]
         :argument [argument :toplevel-eta-expand]
   )
)

(defmethod (abstraction :eta-expand) ()
   (let ((res (make-instance (type-of self) 
                        :picture picture
                        :type type))
       )
     [*exp-identity* :push-key self res]
     [res :set-expression-node [ :expression-node | :eta-expand]]
     [*exp-identity* :pop]
     res
  )
)

(virtualmethod (expression-node :replace-alien-subterms) (abs-list env))

(defmethod (application :replace-alien-subterms) (abs-list env)
  (let* ((fun    functor)
        )
     (if (not [head :ref |:is-constant])  ; only beta-reduced terms allowed!
       (let*  ((type-list [abs-list :map (x) [x :type]])
               (type [type-list :reduce (x) (p [ :expression-type])
                          (make-instance 'function-type
                               :domain x
                               :range p)])
               (H [env :sub-variable type "H"])
              )      
            [*pmas* :move-in (make-instance 'disagreement-pair 
                         :left  H 
                         :right (assemble-exp-binder abs-list self)) :to-pos 'last
            ]  ; therefore correct order.
            [abs-list :reduce :iteration-order 'from-last (x) (p H) 
                   (make-instance 'application
                           :functor p
                           :argument (make-instance 'reference 
                                            :ref [*exp-identity* :get x :test 'eql]))
            ]
         )
         (make-instance (type-of self)                       
                        :functor  [fun       :replace-alien-subterms abs-list env]
                        :argument [argument  :replace-alien-subterms abs-list env])
     )
  )
)

(defmethod (abstraction :replace-alien-subterms) (abs-list env)
  (let ((res (make-instance (type-of self) 
                        :picture picture
                        :type type))
       )
     [abs-list :move-in self]              ; reverse order: directly fit for binder
     [*exp-identity* :push-key self res]
     [res :set-expression-node [ :expression-node | :replace-alien-subterms abs-list env]]
     [*exp-identity* :pop]
     [abs-list :move-out 'first]
     res
  )
)

(defmethod (reference :replace-alien-subterms) (abs-list env)
  (if (not [ :ref | :is-constant])
    (let* ((type-list [abs-list :map (x) [x :type]])
           (type [type-list :reduce (x) (p [ :expression-type])
                          (make-instance 'function-type
                               :domain x
                               :range p)])
           (H [env :sub-variable type "H"])
          )
       [*pmas* :move-in (make-instance 'disagreement-pair 
             :left  H 
             :right (assemble-exp-binder abs-list self)) :to-pos 'last
       ]  ; therefore correct order.
       [abs-list :reduce :iteration-order 'from-last (x) (p H) 
              (make-instance 'application
                      :functor p
                      :argument (make-instance 'reference 
                                       :ref [*exp-identity* :get x :test 'eql]))
       ]
    )
    (let ((bound [*exp-identity* :get [ :ref] :test 'eql]))
       (if bound
          (make-instance 'reference :ref bound)
          [ :copy]
       )
    )
  )
)

; --------------------------------------------------------------------------------
;                         Hilfsprozeduren zur Abstraktion
; --------------------------------------------------------------------------------

; create a declaration with a unique picture beginning with base,
; insert it into the environment and return a reference to it
(defmethod (environment :sub-variable) (type-node base)
  (let ((dec (make-instance 'declaration              ;note the similarity to :declare 
			    :type type-node
			    :picture (new-pic base)
			    :is-constant nil)))
    [:move-in dec]
    (make-instance 'reference :ref dec)))

(setq *Hi-count* 0) 
(defun new-pic (base)
  (setq *Hi-count* (1+ *Hi-count*))
  (format nil "~A~D" base *Hi-count*))


;------------------------------------------------------------------------------
;              Datentyp lambda-list
; -----------------------------------------------------------------------------

(defflavor const-declaration (var) (declaration)
                (:settable-instance-variables var) )

(defflavor lambda-declaration (cons next) (declaration) 
                (:settable-instance-variables cons next) )

(defflavor lambda-list (list top depth) ()
        (:gettable-instance-variables list top depth)
        (:settable-instance-variables list top depth)
        (:initable-instance-variables list top depth) )

(defmethod (lambda-list :new-lambda) (Te S env)
  (incf depth)
  ;(format t "S: ~A~%" S)
  (let* ((T-new [Te :translate])
         (TtoS (make-instance 'function-type 
                     :domain Te
                     :range  S))
         (type (make-instance 'function-type 
                     :domain [S :translate]
                     :range  [TtoS :translate]))
        )
  (if (equal top [list :last])              ; if at the end of list
    (let ((lambda (make-instance 'lambda-declaration           
			    :type type
			    :picture (format nil "~A~D" "lambdax" depth)
			    :is-constant t))
          (varx (make-instance 'declaration           
			    :type Te
			    :picture (format nil "~A~D" "x" depth)
			    :is-constant nil))
          (cons   [env :introduce-constant T-new depth])
         )                   ; create new lambda-declaration
       [cons :set-var varx]
       [lambda :set-cons cons]
       [lambda :set-next nil]
       [ :set-top [list :move-in lambda :to-pos 'last]] ; move it in list
       lambda                               ; and return it
    )
    (let ((lambda nil)                       ; if not at the end of list
         )
       (if top                               ; increase top
          [ :set-top [list :next top]]
          [ :set-top [list :first]]
       )
       (setq lambda [list :get-elem top])    ; get (next) lambda-declaration
       (do () 
           ((or (not [lambda :next])        ; if type disagree + next lambda exists
                (equal [lambda :type] type)))
         (setq lambda [lambda :next])        ; get next lambda
       )
       (if [[lambda :type] :compare type]    ; if type correct. only compare, not equal!!
         lambda                              ; return lambda
                                             ; else no next lambda exist
                                             ; therefore create new lambda
         (let ((next-lambda  (make-instance 'lambda-declaration           
			    :type type
			    :picture (format nil "~A~D" "lambdax" depth)
			    :is-constant t))
               (varx (make-instance 'declaration           
			    :type Te
			    :picture (format nil "~A~D" "x" depth)
			    :is-constant nil))
               (next-cons   [env :introduce-constant T-new depth])
              )
           [next-cons :set-var varx]
           [next-lambda :set-cons next-cons]    
           [lambda :set-next next-lambda]     ; kitt it to the old
           next-lambda                        ; and return it
         )
       )   
    )   ; *let*
  )     ; *if*
  )
)

(defmethod (lambda-list :dispose-lambda) ()
  (decf depth)
  [ :set-top [list :previous top]]
)

(setq *lambda-list* (make-instance 'lambda-list))

(definitmethod (lambda-list) ()
   (setq list (make-instance 'lis-list))
   (setq top nil)
   (setq depth 0)
)

(defmethod (environment :introduce-constant) (type counter)
  (let ((dec (make-instance 'const-declaration           
			    :type type
			    :picture (format nil "~A~D" "consx" counter)
			    :is-constant t)))
    [:move-in dec]  
    dec
  )
)

;------------------------------------------------------------------------------
;              Datentyp converter
; -----------------------------------------------------------------------------

(defflavor converter (to-first to-higher) ()
        (:gettable-instance-variables to-first to-higher)
        (:settable-instance-variables to-first to-higher)
        (:initable-instance-variables to-first to-higher) )

(setq *converter* (make-instance 'converter))

(definitmethod (converter) ()
   (setq to-first  (make-instance 'key-stack-list))
   (setq to-higher (make-instance 'key-stack-list))
)
; here to-higher should be initialized with all constants in the theory, 
; they could be generated by the first-order-precess!

(defmethod (converter :convert-constant) (c l-env)
  (let ((c_exists [to-first :get c :test 'eql]))
    (if c_exists
       c_exists
       (let* ((arg-types [c :type | :argument-types])
              (in  [arg-types :map (type) [type :translate]])
              (out [c :type | :output-type])
              (type (assemble-type out in))
              (new_c  (make-instance (type-of c) 
                           :type type
                           :picture [c :picture]
                           :is-constant t))
             )
         [l-env :move-in new_c]
         [to-first  :push-key c     new_c]
         [to-higher :push-key new_c c    ]
         new_c
       )
    )
  )
)

(defmethod (converter :reconvert-constant) (c)
  [to-higher :get c :test 'eql]
)

(defmethod (converter :convert-variable) (H l-env)
  (let ((Y_exists [to-first :get H :test 'eql]))
    (if Y_exists
       Y_exists
       (let* ((out [H :type | :output-type])
              (Y (make-instance 'declaration 
                      :type out   ; exactly here the argument-types get lost!
                      :picture (format nil "~A~D" 
                                 "Y" (string-left-trim "H" [H :picture]))
                      :is-constant nil))
             )
         [l-env :move-in Y]
         [to-first :push-key H Y]
         Y
       )
    )
  )
)

(defmethod (basic-type :translate) ()
  self 
)

(defmethod (function-type :translate) ()
  (make-instance 'basic-type 
       :picture (format nil 
          "~A~D~A~D~A" "(" [domain :as-string] "->" [range :as-string] ")" )) 
)

(defmethod (function-type :as-string) ()
  (format nil "~A~D~A~D~A" "(" [domain :as-string] "->" [range :as-string] ")" )
)

(defmethod (basic-type :as-string) ()
  picture
)

; ---------------------------------------------------------------------------------------
;                                      Methode Translate
; ---------------------------------------------------------------------------------------

(defflavor algebraic-expression-node () (expression-node) )

(defflavor algebraic-application () (application) )

(defflavor algebraic-reference () (reference) )

(defmethod (disagreement-set :translate) (l-env)
  [ :iterate (pair)
      [pair :translate l-env]
  ]
)

(defmethod (disagreement-pair :translate) (l-env)
  (setq left  [left  :translate (make-instance 'lis-list) l-env])
  (setq right [right :translate (make-instance 'lis-list) l-env])
  self
)

(virtualmethod (expression-node :translate) (x-list l-env))

(defmethod (application :translate) (x-list l-env)
  (if [head :ref | :is-constant]
    [arguments :reduce (s) (h [head :translate x-list l-env])
        (make-instance 'algebraic-application                        
             :functor  h
             :argument [s :translate x-list l-env])
    ]
    [head :translate x-list l-env]  ; self is a simple variable subterm; here
                                    ; the types of the arguments get lost!
  )
)

(defmethod (abstraction :translate) (x-list l-env)
  (let* ((node-type [expression-node :expression-type])
         (l-decl    [*lambda-list* :new-lambda type node-type l-env])
         (fun       (make-instance 'algebraic-reference :ref l-decl))
         res
        )
     [x-list :move-in [l-decl :cons | :var] :to-pos 'last]
     ; (format t "x-list: ~A~%" x-list)
     [*exp-identity* :push-key self l-decl]
     (setq res (make-instance 'algebraic-application
                    :functor  fun
                    :argument [expression-node :translate x-list l-env]))
     [*exp-identity* :pop]
     [x-list :move-out 'last]
     [*lambda-list* :dispose-lambda]
     res
  )
)

(defmethod (reference :translate) (x-list l-env)
  (if [ref :is-constant]
    (let ((lambda-i [*exp-identity* :get [ :ref] :test 'eql]))
      (if lambda-i
        (make-instance 'algebraic-reference :ref [lambda-i :cons])
        (make-instance 'algebraic-reference 
                   :ref [*converter* :convert-constant ref l-env])
      )
    )
    (progn 
      [*H-contexts* :push-key ref [x-list :map (x) x]]
      (make-instance 'algebraic-reference 
                   :ref [*converter* :convert-variable ref l-env])
    )
  )
)


; --------------------------------------------------------------------------------
;                                Methode Retranslate
; --------------------------------------------------------------------------------

(virtualmethod (algebraic-expression-node :retranslate) (var-list env) )

(defmethod (algebraic-application :retranslate) (var-list env)
  (if (and (typep functor 'reference)
           (typep [functor :ref] 'lambda-declaration)
      )
    (let* ((var [functor :ref | :cons | :var])
           (res (make-instance 'abstraction 
                      :picture [var :picture]
                      :type [var :type]))          
         )
       [var-list :move-in var :to-pos 'last] 
       [*exp-identity* :push-key var res]
       [res :set-expression-node [argument :retranslate var-list env]]
       [*exp-identity* :pop]
       [var-list :move-out 'last]
       res
    )
    [arguments :reduce (s) (h [head :retranslate var-list env])
        (make-instance 'application                        
             :functor  h
             :argument [s :retranslate var-list env])
    ]
  )
)

(defmethod (algebraic-reference :retranslate) (var-list env)
  (if [ref :is-constant]
    (if (typep [ :ref] 'const-declaration)
                             ; the first includes two cases: ci of initial var-list
                             ; or of the created part of var-list!
      (make-instance 'reference  
              :ref [*exp-identity* :get [ :ref | :var] :test 'eql])
                             ; these are all constants without the ci.
      (make-instance 'reference 
              :ref [*converter* :reconvert-constant ref])  ; other constants
    )
    [ :lift-variable var-list env]   ; all variables
  )
)

(defmethod (algebraic-reference :lift-variable) (var-list env)
  (let* ((new-decl [*organizer* :new-variable ref var-list env]) 
         (NV (make-instance 'reference :ref new-decl))
       )
     [var-list :reduce (x) (p NV) 
           (make-instance 'application
                   :functor p
                   :argument (make-instance 'reference 
                                :ref [*exp-identity* :get x :test 'eql]))
     ]
  )
)

;------------------------------------------------------------------------------
;              Datentyp New-variable-organizer
; -----------------------------------------------------------------------------

(defflavor new-variable-organizer (old-new) ()
        (:gettable-instance-variables old-new)
        (:settable-instance-variables old-new)
        (:initable-instance-variables old-new) )

(setq *organizer* (make-instance 'new-variable-organizer))

(definitmethod (new-variable-organizer) ()
   (setq old-new (make-instance 'key-stack-list))
)

(defmethod (new-variable-organizer :new-variable) (old-variable var-list env)
 (let* ((pair (make-instance 'disagreement-pair
               :left old-variable
               :right var-list))
        (exist [ :give pair])
       )
    (if exist
       [old-new :get exist :test 'eql]  ; a new decl already exists: return it
       (let* ((type-list [var-list :map (x) [x :type]])
              (type (assemble-type [old-variable :type] type-list))
              (decl (make-instance 'declaration
                      :type type
                      :picture (new-pic-NV "NV")
                      :is-constant nil)) 
             )
          [env :move-in decl]
          [old-new :push-key pair decl]
          decl
       )
     )
  )
)

(defmethod (new-variable-organizer :give) (pair)
  ; Sei pair = (x,<y1, .., yn>)
  (let ((left  [pair :left])
        (right [pair :right])
        (pair-eq nil)
        (existing-pair (make-instance 'disagreement-pair))
       )
     [old-new :iterate ((key info)) 
       ; Sei key = (x',<y'1, .., y'k>)
       (if (equal [key :left] left)    ; x = x'
         (if (equal [key :right |:count] [right :count])  ; n = k
           (let ((list-eq t))
             (multiple-iterate ((dec1 [key :right])
                                (dec2 right))
               (if (not (equal dec1 dec2))
                 (setq list-eq nil)
               )
             )
             (if list-eq    ; y1 = y'1, ... , yn = y'k
               (progn
                 (setq pair-eq t)
                 (setq existing-pair key)
               )
             )
           )    ; let
         )      ; n = k
       )        ; x = x'
     ]
     (if pair-eq
        existing-pair
        nil
     )
  )
)

(setq *NV-count* 0) 
(defun new-pic-NV (base)
  (setq *NV-count* (1+ *NV-count*))
  (format nil "~A~D" base *NV-count*))


; --------------------------------------------------------------------------------
;                          Methode poormans-AC
; --------------------------------------------------------------------------------

; Gibt Liste mit Unifikatoren erster Stufe zurueck, die self AC-gerecht unifizieren.
; env ist das l-env, indem alle first-order Konstanten udn Variablen enthalten sind,
; bis auf die lambda-i!

(defmethod (disagreement-set :poormans-AC) (sub env)
  (if [ :is-empty]
    (make-instance 'lis-list :init-value `(,sub))  ; ready 
    (let* ((pair [ :move-out 'last])
           (left [pair :left])
           (right [pair :right])
          )
      (if (eq [pair :get-type] :rigid-rigid)
         (if (equal [left :head |:ref] [right :head |:ref])  ; case [f(si), f(ti)]
            (if (typep [left :head |:ref] 'c-declaration)
              [:c-variante left right sub env]
              (if (typep [left :head |:ref] 'ac-declaration)
                [:ac-variante [left :head |:ref] left right sub env]
                (progn
                  (multiple-iterate ((a [left :arguments])
                                     (b [right :arguments]))
                    [:move-in (make-instance 'disagreement-pair :left a :right b)] 
                  )
                  [ :poormans-AC sub env]
                )
              )
            )
            nil                                        ; case [f(si), g(ti)]
         )
         (if (eq [pair :get-type] :flexible-rigid)     ; case [x,f(si)] or [f(si),x]
         (progn             
            (if (not [[pair :right] :occur-check [pair :left]])
              (let* ((left  [pair :left])
                     (right [pair :right])
                     (new-sub  (make-instance 'substitution-hash 
                                :init-value `(( ,[left :ref] ,right))))
                   )
                [ :apply-sub new-sub] 
                [sub :move-in-key [left :ref] right] 
                ;[sub :expand]
                [ :poormans-AC sub env]
              )
              nil
            ))
            (if (equal [left :ref] [right :ref])       ; case [x,x]  
               [ :poormans-AC sub env]
               (let ((new-sub  (make-instance 'substitution-hash    ; case [x,y]
                                  :init-value `(( ,[left :ref] ,right))))
                    )
                  [ :apply-sub new-sub] 
                  [sub :move-in-key [left :ref] right] 
                  ;[sub :expand]
                  [ :poormans-AC sub env]
               )
            )
          )
       )   ; if
    )      ; let
  )        ; if
)

(defmethod (disagreement-set :c-variante) (left right sub env)
   (let ((help-dis [self :copy]))
      (multiple-iterate ((a [left :arguments])
                         (b [right :arguments]))
        [:move-in (make-instance 'disagreement-pair :left a :right b)] 
      )
      (multiple-iterate ((a [left :arguments])
                         (b [[right :arguments] 
                   :map :iteration-order 'from-last (x) x]))
        [help-dis :move-in (make-instance 'disagreement-pair :left a :right b)] 
      )
      (let ((first [ :poormans-AC [sub :copy] env])
            (sec   [help-dis :poormans-AC [sub :copy] env]))
        (if first
           (if sec
              (progn
                [sec :iterate (x) [first :move-in x]]
                first
              )
              first
           )
           (if sec
              sec
              nil
           )
         ) 
      )
    )
)

(defmethod (disagreement-set :ac-variante) (ac-decl left right sub env)
 (let* ((res-list (make-instance 'lis-list))
        (solution1 [self :copy])
        (solution2 [self :copy])
        (solution3 [self :copy])
        (solution4 [self :copy])
        (solution5 [self :copy])
        (solution6 [self :copy])
        (solution7 [self :copy])
        (arg1 [left :arguments])
        (arg2 [right :arguments])
        (x1 [arg1 :get-by-pos 1])
        (x2 [arg1 :get-by-pos 2])
        (y1 [arg2 :get-by-pos 1])
        (y2 [arg2 :get-by-pos 2])
        (var-type [ac-decl :type |:range |:domain])
        (NV1 [env :sub2-variable var-type "Neu"]) 
        (NV2 [env :sub2-variable var-type "Neu"])
        (NV3 [env :sub2-variable var-type "Neu"])
        (NV4 [env :sub2-variable var-type "Neu"])
       )
         (multiple-iterate ((a arg1)
                            (b arg2))
           [solution1 :move-in (make-instance 'disagreement-pair 
                               :left a 
                               :right b)] 
         )          
         [solution2 :move-in (make-instance 'disagreement-pair
                                :left  x1
                                :right y2)]
         [solution2 :move-in (make-instance 'disagreement-pair
                                :left  x2
                                :right y1)]
         [solution3 :move-in (make-instance 'disagreement-pair
                                :left  x1
                                :right [ac-decl :appl-build y1 NV1])]
         [solution3 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build x2 NV1] 
                                :right y2)]
         [solution4 :move-in (make-instance 'disagreement-pair
                                :left  x1
                                :right [ac-decl :appl-build NV2 y2])]
         [solution4 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build x2 NV2]
                                :right y1)]
         [solution5 :move-in (make-instance 'disagreement-pair
                                :left  x2
                                :right [ac-decl :appl-build y1 NV3])]
         [solution5 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build NV3 x1]
                                :right y2)]
         [solution6 :move-in (make-instance 'disagreement-pair
                                :left  x2
                                :right [ac-decl :appl-build NV4 y2])]
         [solution6 :move-in (make-instance 'disagreement-pair
                                 :left [ac-decl :appl-build NV4 x1]
                                 :right y1)]
         [solution7 :move-in (make-instance 'disagreement-pair
                                :left  x1
                                :right [ac-decl :appl-build NV2 NV1])]
         [solution7 :move-in (make-instance 'disagreement-pair
                                :left  x2
                                :right [ac-decl :appl-build NV4 NV3])]
         [solution7 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build NV4 NV2]
                                :right y1)]
         [solution7 :move-in (make-instance 'disagreement-pair
                                :left  [ac-decl :appl-build NV3 NV1]
                                :right y2)]
         [res-list :move-in [solution1 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution2 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution3 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution4 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution5 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution6 :poormans-AC [sub :copy] env]]
         [res-list :move-in [solution7 :poormans-AC [sub :copy] env]]
         (let ((sol nil))
           (block suche-eine-lsg
           [res-list :iterate (sub idx) 
              (if sub 
                  (progn
                     (setq sol sub)
                     [res-list :move-out `(at-index ,idx)]
                     (return-from suche-eine-lsg nil)
                  ))
           ]
           )
           [res-list :iterate (sub) 
               (if sub
                   [sub :iterate (x) [sol :move-in x]]
               )
           ]
           sol
         )
  )
)

; @@@@@@@@@Sollte man mit sub-variable zusammentun!!
(defmethod (environment :sub2-variable) (type-node base)
  (let ((dec (make-instance 'declaration              ;note the similarity to :declare 
			    :type type-node
			    :picture (neu-pic base)
			    :is-constant nil)))
    [:move-in dec]
    (make-instance 'algebraic-reference :ref dec)))

(setq *Neu-count* 0) 
(defun neu-pic (base)
  (setq *Neu-count* (1+ *Neu-count*))
  (format nil "~A~D" base *Neu-count*))


(defflavor c-declaration () (declaration) )

(defflavor ac-declaration () (declaration) )

(defmethod (environment-lis :new-declare) (decl)
  [:move-in decl]
)

(defmethod (ac-declaration :appl-build) (x y)
  (make-instance 'algebraic-application 
          :functor (make-instance 'algebraic-application 
                             :functor (make-instance 'algebraic-reference :ref self)
                             :argument x)
          :argument y
  )
)

; Beachte: Solange :replace-alien-subterms jedes Vorkommen einer Variable 
; (oder eines Terms) durch eine andere Variable ersetzt, braucht man den
; occur-check nicht ...
(virtualmethod (expression-node :occur-check) (x))

(defmethod (application :occur-check) (x)
  (let ((occurs nil))
    [arguments :iterate (si)
      (if [si :occur-check x] 
        (setq occurs t)
      )
    ]
    occurs
  )
)

(defmethod (reference :occur-check) (x)
  (equal ref [x :ref])
)

; ---------------------------------------------------------------------------------------
;                                      Methode less-equal
; ---------------------------------------------------------------------------------------

(defmethod (substitution-hash :less-equal) (rho env)
  (let ((sigma [:expand self]) 
        (rho   [rho :expand rho])
        (dis-set (make-instance ' disagreement-set))
        (sub (make-instance 'substitution-hash))
       )
     [sigma :iterate ((var exp))
        (let ((varx (make-instance 'reference :ref var)))
           [dis-set :move-in (make-instance 'disagreement-pair 
                 :left exp
                 :right [varx :apply-sub rho | :free-vars-to-new-cons sub])
           ]
        )
     ]
     ;(format t "dis-set: ~A" dis-set)
     (if [dis-set :poormans-AC (make-instance 'substitution-hash) env] t nil)
  )
)

(virtualmethod (expression-node :collect-free-variables) (var-list) )

(defmethod (application :collect-free-variables) (var-list)
  (let ((list1 [functor  :collect-free-variables [var-list :copy]])
        (list2 [argument :collect-free-variables [var-list :copy]]) 
       )
     [list1 :iterate (var)
        (if (not [list2 :find var])
            [list2 :move-in var]
        ) 
     ]
     list2
  )
)

(defmethod (abstraction :collect-free-variables) (var-list)
  [expression-node :collect-free-variables [var-list :copy]]
)

(defmethod (reference :collect-free-variables) (var-list)
  (if (not [ :ref | :is-constant]) [var-list :move-in [self :ref]])
  var-list
)

(defmethod (expression-node :free-vars-to-new-cons) (sub) 
  (let ((frees [ :apply-sub sub |:collect-free-variables (make-instance 'lis-list)])
       )
     [frees :iterate (var) 
        (let ((cons (make-instance 'declaration 
                           :type [var :type]
                           :picture (format nil "~A~D" "cons-" [var :picture])
                           :is-constant t))
             )
           [sub :move-in-key var (make-instance 'reference :ref cons)]
        )
     ]
     [ :apply-sub sub]
  )
)

(defmethod (lis-list :eliminate-redundant-subs) (env)
    [ :iterate (sub1 idx1)
       [ :iterate (sub2 idx2)
           (if (AND (not (eq idx1 idx2))
                    [sub2 :less-equal sub1 env])
              [ :move-out `(at-index ,idx1)]
           )
       ]
       ; (format t "Rest-subs: ~%")
       ; (print-subsx self)  Ausgabe zu Testzwecken
    ]
    self
)


(defmethod (substitution-hash :expand) (sub)
   (let ((res (make-instance 'substitution-hash))
         )
      [ :iterate ((var exp))
         [res :move-in-key var [exp :apply-sub sub]]
      ]
      res
   )
)

; ---------------------------------------------------------------------------------------
;                                 (h-o) - Unifikation fuer Mengen
; ---------------------------------------------------------------------------------------

; Unification method for disagreement-sets
(defmethod (disagreement-set :set-unify) 
     (environment &key (substitutions (make-instance 'substitution-hash)) 
	               (patience 500)
                  &allow-other-keys)
  (setq *count-nodes* 0)
  (setq *uni-pic-count* 0)
  (let* ((mtsub (make-instance 'substitution-hash))
	 (node  (make-instance 'searchtree-node :disagreements self
		                                :new-sub       substitutions
		                                :all-subs      mtsub
                                                :depth         0))
	 (tree  (make-instance 'searchtree :env environment :patience patience))
        )
    [tree :push node]
    (make-instance 'lis-lazy :init-value (uni-unify tree))
  )
)
; ---------------------------------------------------------------------------------------
;                                      Methode h-o-less-equal
; ---------------------------------------------------------------------------------------

(defmethod (substitution-hash :h-o-less-equal) (rho env)
  (let ((sigma self)  ; expand??@@@
        (dis-set (make-instance ' disagreement-set))
        (sub (make-instance 'substitution-hash))
       )
     [sigma :iterate ((var exp))
        (let ((varx (make-instance 'reference :ref var)))
           [dis-set :move-in (make-instance 'disagreement-pair 
                 :left exp
                 :right [varx :apply-sub rho | :free-vars-to-new-cons sub])
           ]
        )
     ]
     (if [[dis-set :set-unify env] :get-by-pos '1] t nil)
  )
)

(defmethod (uni-sub :expand) (sub)
   (let ((res (make-instance 'uni-sub))
         )
      [ :iterate ((var exp))
         [res :move-in-key var [exp :apply-sub sub]]
      ]
      [res :set-flex-flex flex-flex]
      res
   )
)

(defmethod (uni-sub :project-help-variables) (exp1 exp2)
 (let* ((res (make-instance 'uni-sub))
       (exp (make-instance 'application 
                   :functor exp1
                   :argument exp2))
       (var-list [exp :collect-free-variables (make-instance 'lis-list)])
      )
   [ :iterate ((var exp))
      (if [var-list :find var]
          [res :move-in-key var exp]
      )
   ]
   [res :set-flex-flex flex-flex]
   res
 )
)


(defmethod (lis-list :eliminate-redundant-ho-subs) (env)
    [ :iterate (sub1 idx1)
       [ :iterate (sub2 idx2)
           (if (AND (not (eq idx1 idx2))
                    [sub2 :h-o-less-equal sub1 env])
              [ :move-out `(at-index ,idx1)]
           )
       ]
       ; (format t "Rest-subs: ~%")
       ; (print-subsx self)  Ausgabe zu Testzwecken
    ]
    self
)


; Eliminiert das Ergebnis von :E-unify um abhaengige ho-subs, indem vorher alle
; Hilfsvariablen wegprojeziert wurden. Geht aber nur bei richtigen Substitutionen,
; meist nciht, wenn flex-flex Paare da sind.

(defmethod (lis-lazy :optimize) (env exp1 exp2)
   (let ((list (make-instance 'lis-list))
        )
     [ :iterate (uni-sub) 
        [list :move-in [uni-sub :expand uni-sub |:project-help-variables exp1 exp2]]
     ]
     [list :eliminate-redundant-ho-subs env]
     list
   )
)

