
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;         W I N D O W   T O O L   K I T             ;;;
;;;    (Chapter 12 of SUN COMMON LISP USER'S GUIDE)   ;;;
;;;                                                   ;;;
;;;      EXTENSIONs by P. Weber, FZI, Karlsruhe       ;;;
;;;                                                   ;;;
;;;           part for DRAWing functions              ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(in-package 'windows)
(export '(; Drawing functions
          drawf ; draws a series of one-argument functions simultaneously
          drawf-2 ; draws a two-argument function in a 3-dimensional view
          rotate  ; for animation of drawf-2
          ; the parameter which influence the display
          *drawf-width* *drawf-height* *drawf-offset*
          *alpha*))


(proclaim '(special *drawf-bitmap* *drawf-viewport*))
  ; bitmap and viewport used by drawf and drawf-2
(defparameter *drawf-width* 520)  ; defines the width and height 
(defparameter *drawf-height* 520) ; of the region used for drawf
(defparameter *drawf-offset* 20)  ; the window for drawf is as large
                          ; as offset added to the region at each side.

(defparameter *alpha* 45) ; this parameter defines the angle between x-axis
       ; and y-axis for the 3-dimensional view used by drawf-2


(defun DRAWF (fcts low_x high_x &optional (steps 100) )
"draws simultaneously the functions in the list FCTS or the single
 function FCTS with x-axis from LOW_X to HIGH_X by interpolating
 the function-values at steps points"
   (initialize-windows)
   (setq *drawf-bitmap* (make-bitmap :width 530 :height 530))
   ;; the functions are drawn in rectangle from (10,10) to (510,510)
   (setq *drawf-viewport* (make-viewport :bitmap *drawf-bitmap*)) 
 (let ((step (/ (- high_x low_x) steps))
       (high_y most-negative-fixnum)
       (low_y most-positive-fixnum) 
       x_pos y_pos
       (x_diff (- high_x low_x)) y_diff
       old_pos (cur_pos (make-position 10 0)) )
     (unless (listp fcts)(setq fcts (list fcts)))
     ;; compute high and low bound for y
     (mapc #' (lambda (fct)
        (do* ((x low_x (+ x step))
              (y (funcall fct x) (funcall fct x)))
             ((> x high_x) (setq y_diff (- high_y low_y)))
             (when (> y high_y)(setq high_y y))
             (when (< y low_y)(setq low_y y))
            ))
        fcts)
      ;; draw axis 
      (let (x_axis-left_x  x_axis-right_x  x_axis-y
            y_axis-bot_y  y_axis-top_y  y_axis-x)
           (setq x_axis-y
              (cond ((>= low_y 0) 510)
                    ((<= high_y 0) 10)
                    (t (round (+ 10 (* (/ high_y y_diff ) 500)))) ))
           (setq x_axis-left_x
               (if (<= low_x 0) 0 20))
           (setq x_axis-right_x
               (if (>= 0 high_x) 520 500))
           (setq y_axis-x
              (cond ((>= low_x 0) 10)
                    ((<= high_x 0) 510)
                    (t (round (+ 10 (* (/ (- low_x) x_diff ) 500)))) ))
           (setq y_axis-bot_y
               (if (<= low_y 0) 520 500))
           (setq y_axis-top_y
               (if (>= 0 high_y) 0 20))
           ; draw x-axis
           (draw-line *drawf-viewport*
                 (make-position x_axis-left_x x_axis-y)
                 (make-position x_axis-right_x x_axis-y))
           (stringblt *drawf-bitmap*
                 (make-position x_axis-left_x (+ x_axis-y 10))
                 *default-font*
                 (format nil "~D" low_x))
           (stringblt *drawf-bitmap*
                 (make-position x_axis-right_x (+ x_axis-y 10))
                 *default-font*
                 (format nil "~D" high_x))
           ; draw y-axis
           (draw-line *drawf-viewport*
                 (make-position y_axis-x y_axis-top_y)
                 (make-position y_axis-x y_axis-bot_y))
           (stringblt *drawf-bitmap*
                 (make-position (- y_axis-x 10) y_axis-bot_y)
                 *default-font*
                 (format nil "~D" low_y))
           (stringblt *drawf-bitmap*
                 (make-position (- y_axis-x 10) y_axis-top_y)
                 *default-font*
                 (format nil "~D" high_y))
         )      
      ;; draw functions
     (mapc #'(lambda (fct)
         ; set first point
         (setq y_pos 
              (round (+ 10 (* (/ (- high_y (funcall fct low_x)) y_diff )
                                 500))))
         (setq cur_pos (make-position  10 y_pos))
         (do* ((x (+ low_x step)(+ x step))
               (y (funcall fct x) (funcall fct x)))
              ((> x high_x))
              (setq old_pos cur_pos)
              (setq x_pos (round (+ 10 (* (/ (- x low_x) x_diff ) 500))))
              (setq y_pos (round (+ 10 (* (/ (- high_y y) y_diff ) 500))))
              (setq cur_pos (make-position  x_pos y_pos))
              (draw-line *drawf-viewport* old_pos cur_pos)
           ))
         fcts)
   ))
   
        
(defun DRAWF-2 (fct low_x high_x 
                &optional (low_y nil) (high_y nil) ; are set eq bounds for x
                          (steps 15) (axes nil))
 "draws the 2-argument function FCT in a 3-dimensional way with x-axis from 
  LOW_X to HIGH_X and y-axis from Y_LOW to Y_HIGH by interpolating
  the function-values at steps points for each axis"
   ; set defaults for y-bounds
  (unless low_y  (setq low_y  low_x ))
  (unless high_y (setq high_y high_x))
   ; create viewport for display
  (initialize-windows)
  (setq *drawf-bitmap*
        (make-bitmap :width  (+ *drawf-width*  (* 2 *drawf-offset*))
                     :height (+ *drawf-height* (* 2 *drawf-offset*))) )
  (setq *drawf-viewport* (make-viewport :bitmap *drawf-bitmap*)) 
 (let* ((x-diff (- high_x low_x))
         (step-x (/ x-diff (1- steps)))
        (y-diff (- high_y low_y))
         (step-y (/ y-diff (1- steps)))
        (high_z most-negative-fixnum)
         (low_z most-positive-fixnum)
         z-diff z-array
        (alpha-rad (/ (* *alpha* pi) 180))
        (y-factor (* (sin alpha-rad)(cos alpha-rad)))
        (x-factor (* (cos alpha-rad)(cos alpha-rad))) 
        )
    ;; compute high and low bound for z
    ;; and store function values into z-array
    (setq z-array (make-array (list steps steps)))
    (let ((x low_x))
      (dotimes (i steps)
        (let ((y low_y))
          (dotimes (j steps)
            (let ((z (funcall fct x y)))
              (setf (aref z-array i j) z)
              (when (> z high_z)(setq high_z z))
              (when (< z low_z)(setq low_z z)) )
            (incf y step-y)))
        (incf x step-x) ))
    (setq z-diff (- high_z low_z))
    (when (zerop z-diff)(setq z-diff 1)) ; diff must not be 0!
 ; define function to transform (x,y,z) coordinates into positions
 (flet ((make-point (x y z)
          (let  ((rel-x (/ (- x low_x) x-diff)) ; relativate coordinate
                 (rel-y (/ (- y low_y) y-diff)) ; into unity cube
                 (rel-z (/ (- z low_z) z-diff))
                 x-pos y-pos)
            ; project y-coord 3-dim
            (setq x-pos (+ rel-x (* x-factor rel-y)))
            (setq y-pos (+ rel-z (* y-factor rel-y)))
            ; expand unity cube onto full window
           (make-position 
                (round (+ *drawf-offset* 
                          (/ (* x-pos *drawf-width*) (1+ x-factor))))
                (round (- (+ *drawf-offset* *drawf-height*)
                          (/ (* y-pos *drawf-height*) (1+ y-factor)))))
         )))     
    ;; draw 3-dim axes
  (when axes
    (let ((origin ( make-point low_x low_y low_z))
          (right  ( make-point high_x low_y low_z))
          (top    ( make-point low_x low_y high_z))
          (back   ( make-point low_x high_y low_z)))
     ; main axes
      (draw-line *drawf-viewport* ; x-axis
                 origin right :width 3)
      (draw-line *drawf-viewport* ; y-axis
                 origin back :width 2)
      (draw-line *drawf-viewport* ; z-axis
                 origin top :width 3)
     ; make skala
      (let ((o-x (position-x origin))
            (o-y (position-y origin))
            (r-x (position-x right))
            (t-y (position-y top))
            (b-x (position-x back))
            (b-y (position-y back)) )
      (stringblt *drawf-bitmap* (make-position o-x (+ o-y 10))
                 *default-font* (format nil "~D" low_x))
      (stringblt *drawf-bitmap* (make-position r-x (+ o-y 10))
                 *default-font* (format nil "~D" high_x))
      (stringblt *drawf-bitmap* (make-position (+ o-x 10)(- o-y 10))
                 *default-font* (format nil "~D" low_y))
      (stringblt *drawf-bitmap* (make-position (- b-x 10)(+ b-y 10))
                 *default-font* (format nil "~D" high_y))
      (stringblt *drawf-bitmap* (make-position (- o-x 10) o-y)
                 *default-font* (format nil "~D" low_z))
      (stringblt *drawf-bitmap* (make-position (- o-x 10) t-y)
                 *default-font* (format nil "~D" high_z))
       )
     ; hidden axes
      (let ((back_top ( make-point low_x high_y high_z))
            (right_top ( make-point high_x low_y high_z))
            (right_back ( make-point high_x high_y low_z))
            (right_back_top ( make-point high_x high_y high_z))
            )
      (draw-line *drawf-viewport* top back_top)
      (draw-line *drawf-viewport* top right_top)
      (draw-line *drawf-viewport* right right_top)
      (draw-line *drawf-viewport* right right_back)
      (draw-line *drawf-viewport* right_back back)
      (draw-line *drawf-viewport* back back_top)
      (draw-line *drawf-viewport* back_top right_back_top)
      (draw-line *drawf-viewport* right_back_top right_back)
      (draw-line *drawf-viewport* right_back_top right_top)
    ) ))

  ;; draw function y-layers and transform z-values into positions in z-array
    (let ((y low_y))
      (dotimes (j steps)
        (let ((x low_x) old_pos cur_pos)
          (dotimes (i steps)
             (setq old_pos cur_pos)
             (setq cur_pos (make-point x y (aref z-array i j)))
             (when old_pos  
               (draw-line *drawf-viewport* old_pos cur_pos))
             (setf (aref z-array i j) cur_pos)            
             (incf x step-x) ))
        (incf y step-y) ))
  ;; draw function x-layers
     (dotimes (i steps)
        (let (old_pos cur_pos)
          (dotimes (j steps)
             (setq old_pos cur_pos)
             (setq cur_pos (aref z-array i j))
             (when old_pos  
               (draw-line *drawf-viewport* old_pos cur_pos))
      ) )  )
 ))) ; defun, let*, flet

 
(defvar *rotate-delay* 0.2) ; pause between two bitmaps  

(defun ROTATE (drawf-2-call 
                  &optional (steps 20) (min_a 0) (max_a 90) )
 "evaluates drawf2-call steps times with different angles and
  rotates the generated bitmaps circulary"
  (clear-root)
  (setf (viewport-children (root-viewport)) nil)
  (let ((old-alpha *alpha*)
        (old-delay *rotate-delay*)
        (speed-menu
           (make-menu '(("+" 0.8)("-" 1.2)("stop" 0)) 
              :x (+ *drawf-width* (* 2 *drawf-offset*) 5)
              :y (+ *drawf-height* (* 2 *drawf-offset*) 5)
              :action #'(lambda (c)
                          (setq *rotate-delay* 
                                (* *rotate-delay* c)))
              :title "speed" ))
       )   
   (unwind-protect ; *alpha* and *rotate-delay* must be reset
     (progn
       (when (> min_a max_a) ; exchange boundaries
             (setq min_a max_a 
                   max_a min_a))
       ; evaluate drawf-2-call with different angles
       (let ((step (/ (- max_a min_a) steps)))
          (do ((a min_a (+ a step)))
              ((> a max_a))
             (setq *alpha* a)
             (eval drawf-2-call)
        ) )     
       ; rotate all generated bitmaps in infinite loop
       (display-menu speed-menu)
       (let (views)  
        (setq views (cdr (viewport-children (root-viewport)))) ;
                ; all children but the menu ! 
        (setq views (append (reverse views) (cdr views))) ; and going up
        (setf (cdr (last views)) views) ; make circular 
        (loop
          (when (<= *rotate-delay* 0) (return))  
          (sleep *rotate-delay*) 
          (expose-viewport (car views))
          (setq views (cdr views)) )
        (setq *alpha* old-alpha)
        (setq *rotate-delay* old-delay)
        (clear-menu speed-menu)    
       ))
     (progn
      (setq *alpha* old-alpha)
      (setq *rotate-delay* old-delay)
      (clear-menu speed-menu)) )    
 ))    
  
