
(defvar *counted-fcts* nil)   

(defun counter-name (f)
  "generates a symbol used as counter for calls to f"
    (read-from-string (format nil "c-~A" f))
 )
   
(defun fcount (&optional (flist nil)  &key (reset t) (new nil))
  "when flist=nil: the counts since the last reset to fcount are presented.
   Else every function in  flist is added to the counted functions starting
   with count=0. When the :new argument is non-nil the counted-fcts are
   first cleared. Flist could also be a single symbol.
   The counts of all fcts are set to 0, unless the :reset argument is nil."
 
   ; print first counters if required  
 (unless flist ; print counters if flist is nil
   (dolist (f *counted-fcts*)
      (format t "~%~10D -- ~A" (eval (counter-name f)) f)))
   ; then update *counted-fcts* -  could have changed due to (untrace)
   ; and reset counters if reset is non-nil
 (setq *counted-fcts* 
    (if new nil (intersection *counted-fcts* (trace))))
 (when reset
   (dolist (cf *counted-fcts*) 
        (set (counter-name cf) 0)))
  ; add the new fcts in flist for counting
 (when flist
    (when (not (listp flist))
        (setq flist (list flist)))       
    (let ((counters (mapcar #'counter-name flist)))
      (dolist (c counters) (set c 0))
      (mapc #'(lambda (f c)
                (eval `(trace (,f :exitcond nil :entry ((incf ,c))))))
            flist counters))
    (setq *counted-fcts* (nunion *counted-fcts* flist))
 ))
          

(defun silent-trace (&optional (wanted t))
 "directs the trace output in a sink if wanted is non-nil"
   (setq *trace-output*
     (if wanted
      (make-broadcast-stream)
      (make-synonym-stream *standard-output*)))
)  

    
   
