;;;;-----------------------------------------------------------;;;;   
;;;;      Counting GC-Time for efficiency -tests               ;;;;
;;;;-----------------------------------------------------------;;;;

;; Usage: The function to be measured must (periodically) call TIMING-GC 
;;        to prevent the system from invoking gc. It is impossible to 
;;        mesure the time for a gc called from system. 
;;        Then use TTIME like TIME to get timing information including
;;        gc-times caused by TIMING-GC. 

(defvar *gc-run-time*  0)
(defvar *gc-real-time* 0)
(defun timing-gc ()  
   (let ((real-time (get-internal-real-time))
         (run-time  (get-internal-run-time ))
         (silence system::*gc-silence*)      )
        (setf system::*gc-silence* t)
        (multiple-value-prog1
           (system:gc)
           (setf system::*gc-silence* silence)
           (incf *gc-run-time*  (- (get-internal-run-time)  run-time ))
           (incf *gc-real-time* (- (get-internal-real-time) real-time)) )))
(defmacro ttime (eval-form)
    `(let ((gc-real-time *gc-real-time*)
           (gc-run-time  *gc-run-time* ) )
          (time ,eval-form)
          (format *trace-output*
                  "~%GC: [elapsed real time = ~D ] [total cpu time = ~D ]"  
                  (- *gc-real-time* gc-real-time)
                  (- *gc-run-time* gc-run-time))  )) 




