;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;              LIB-MODULE  LOAD-FONTS                       ;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (provide 'load-fonts) at the end of file

(require-lib 'window-extension)

(in-package 'windows )

(export '(
           require-font
           select-font 
           invert-font
           show-all-fonts
         ))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *font-homedir* (concatenate 'string user::*mod-homedir* "/fonts/"))

(defun require-font (&rest fonts)
 "Loads the fonts if not already loaded. Returns the last font.
  Font-names are either given by symbols or strings." 
  (let (res)
    (dolist (f fonts)
       (setq res
         (or (find-font (string-downcase (format nil "~A" f)))
             (find-font (format nil "~A" f))
             (find-font (format nil "~:@(~A~)" f)) ; uppercase
             (let ((fontfile
                     (pathname (format nil "~A~A.font" *font-homedir* f)))
                   )
                (when (probe-file fontfile)
                    (load-font fontfile)  )
              )
        ) )
     )
    res
  ) 
) 
                  
(defun select-font ()
 "makes a menu to select from all fonts in *font-homedir*
  the selected font becomes the *default-font*"
  (let ((font-files (directory *font-homedir*))
        font-choices font-menu)
   (setq font-choices 
     (mapcar #'(lambda (f) (read-from-string (pathname-name f)))
             font-files))
    ; menu shows the font-names and returns the font-filenames
   (setq font-menu (make-pop-up-menu font-choices))   
   (let ((menu-value (pop-up-menu-choose font-menu)) 
         )
     (when menu-value
        (setq *default-font* (require-font menu-value))
      )     
     *default-font*
   )
 ))
 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun invert-font (font-name)
 "defines for an existing font <f> 
     the inverted font 'inverted-<f> and returns it."

  (let* ((font (require-font font-name))
         (code-limit (font-code-limit font))
         (fixed-width (font-fixed-width font))
         inv-font inv-bitmap
       )
   ; initialize the inverted font
   (setq inv-font
    (make-font 
      (lisp-extension:format-symbol "inverted-~A" (font-name font))
      :code-limit  code-limit
      :fixed-width fixed-width
      :height      (font-height font)
      :baseline    (font-baseline font)
    ))
   ;; now define the character images
     ; first invert the original bitmap
   (setq inv-bitmap (copy-bitmap (font-bitmap font)))
   (bitblt  inv-bitmap 0 0 
            inv-bitmap 0 0 
            (bitmap-width inv-bitmap)
            (bitmap-height inv-bitmap)
            boole-c1)
   (dotimes (i code-limit)
     (font-set-char inv-font
        (code-char i) (* i fixed-width)
        nil inv-bitmap (make-position (* i fixed-width) 0)
      )
    )
   inv-font
  )
)
 
(defun show-all-fonts ()
 "prints all available fonts in a window"
  (let ((fonts (mapcar #'(lambda (f) (require-font (pathname-name f)))
                     (directory *font-homedir*)) )
        (h 20)
        w)
         
     (dolist (f fonts)
       (incf h (+ (font-height f) 20))
      )
     (setq w (make-window :x 10 :y 10 :width 1000 :height h
               :viewport-height (min (- (window-height (root-viewport)) 40)
                                       h)
               :viewport-width (- (window-width (root-viewport)) 40)
               :scroll t))
     (setq h 10)
     (dolist (f fonts)
       (stringblt w
                (make-position 10 (incf h (font-height f)))
                f
                (font-name f) )
       (incf h 20)
      )
   )
)    
       

(provide 'load-fonts)
