;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; MEASURING efficiency of functions
; measuring is turned on and off like tracing
; WARNING: don't apply measure to recursive functions or to functions
;    whose side-effects are not invariant when repeatedly called!

(defvar *measured-functions* nil)

(defun measure (fct &key (repeat 1) (stream t) msg )
 ; msg is an additional statement that is executed after mesuring
 ; Sample (... :msg '(format t " for ~A and ~A" 
 ;                   (first lisp-extension::args)(second lisp-extension::args)) )

 (if (null fct)
    *measured-functions*
   (let (orig-def)
     ; save the original function definiton if not already measured (->saved)
    (unless (member fct *measured-functions*) 
          (setf (get fct 'saved-fct) (symbol-function fct)))
      ; in either case original fdef. now in proplist
    (setq orig-def (get fct 'saved-fct))
     ; define new function with measuring around the original definition
    (eval
    `(defun ,fct (&rest args)
       (let (time)
        (msecs (dotimes (i ,repeat)
                 (apply ,orig-def args))
               time)
        (format ,stream "~%~D* ~A: ~D msecs" ,repeat (quote ,fct) time)
        ,msg)
        (apply ,orig-def args)
      ))
    (compile fct)
    (pushnew fct *measured-functions*))
))

(defun unmeasure (&rest fcts)
  ; restores the original definitions for the functions fcts
  ; if no argument is supplied all measured functions are unmeasured
   (unless fcts (setq fcts *measured-functions*))
   (dolist (fct fcts)
      (if (member fct *measured-functions*)
         (setf (symbol-function fct) (get fct 'saved-fct))
         (format t "~%Warning: ~A is not measured" fct)) )
   (setq *measured-functions*
      (nset-difference *measured-functions* fcts))
 ) 
  

