;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;								;;;;
;;;; 		M O D U L E    U P G R A D E 2 - 1		;;;;
;;;;								;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;           Some functions of CL 2.1. not in V. 2.0.	        ;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;           	AUTHOR:  Paul Weber, FZI, Karlsruhe	        ;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'upgrade2-1)


(setq *load-binary-pathname-types* '("lbin"))

(in-package 'upgrade2-1)

(export
   '(shell ashell cd pwd ls
  ))

  
(defun SHELL (string)                                               
"String is interpreted as csh-command.
 WAITS for completion of the command"
  ; creates a process for each call and exits it after execution
  (check-type string string)
  (with-open-file (csh-input "/usr/tmp/lsh.out" :direction :output)
     (format csh-input "cd ~A~%~A~%~A"
        (namestring *default-pathname-defaults*) ; cd to current lisp dir
        string                                   ; execute the command
        "exit"					 ; and exit
      )
    )        
  (system:run-unix-program "csh" :input "/usr/tmp/lsh.out")
)
  
  
(defun ASHELL (string)    ;; asynchron version of Shell
"String is interpreted as csh-command.
 ASYNCHRON, does not wait for completion."
; creates only one process for cshell when first called. 
 (declare (special *cshell*))
  (check-type string string)
  (when (not (and (boundp '*cshell*)
                  (streamp *cshell*)))
    (setq *cshell*  ; can't wait with input from stream
       (system:run-unix-program "csh" :input :stream :wait nil)))
  (format *cshell* "~A~%" string)
  (values)
  )

    
(defun CD (&optional (path nil))
   ;; set *default-pathname-defaults*
    (setq *default-pathname-defaults* 
          (pathname (lisp-extension::full-pathname path)))
 )
  
(defun PWD ()
  (namestring *default-pathname-defaults*)
 )
  
(defun LS ()
  (shell "ls")
 )



;;;; import UPGRADE into package USER

(in-package 'USER)
(use-package 'UPGRADE2-1)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;  			Adapt LOAD-FOREIGN-FILES   		 ;;;;

;; In CL 2.0 the first argument to Load-foreign-files must be a list
;; of full(!) pathnames whereas in 2.1 it could be also a single 
;; string and the pathnames could be relative to the current directory
;;
;; Because load-foreign-files never returns if a file does not exist
;; file-existance is checked before a call to load-foreign-files

;;; DOES NOT WORK because load-foreign-files seems to be recursive!

#|
;;; first save the 2.0 definition
(setf (symbol-function '*load-foreign-files-V2.0*)
      (symbol-function 'load-foreign-files))

;; now redefine load-foreign-files
(defun load-foreign-files (files &optional (libs nil lsup))
   (unless (listp files) 
      (setq files (list files)))
   (setq files
     (mapcan #'(lambda (f)
               (let ((p (probe-file f))) ; returns the full pathname or nil
                  (if p 
                     (list (namestring p))
                     (format t "No foreign file ~A" f)
              ))  )
            files ))
  (when (car files) ; there exists a file to load
   (format t  "current directory is : ~A " *default-pathname-defaults*)
   (format t "Loading foreign files : ~{~%~10T~S~}" files)
   (unwind-protect
   (if lsup ; take care of defaults for optional parameter
            ; lib of original def.
     (*load-foreign-files-V2.0* files libs)
     (*load-foreign-files-V2.0* files)
   ))
 )
)

|#
  
