#| 
   module      : mod
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 

      module management

      *mod-home-path* contains module directories

      module string or symbol
      file   string

      (defun mod-set (module))
	  set current module

      (defun mod-get ())
	  get current module

      (defun mod-sub-modules (&optional (module *mod-current-module*) )
          get list of sub-modules contained in module 

      (defun mod-recompile (&optional (module *mod-current-module*)
			    &rest  r &key &allow-other-keys)
	   compile all submodules of module
	   pass keyword parameters to function 'compile-file'

      (defun mod-compile (&optional (module *mod-current-module*)
			  &rest  r &key &allow-other-keys)
	   compile changed submodules of module
	   pass keyword parameters to function 'compile-file'

      (defun mod-reoptimize (&optional (module *mod-current-module*)
			   &rest  r &key &allow-other-keys)
	   compile changed submodules of module
	   pass keyword parameters to function 'compile-file'
	   optimization version

      (defun mod-optimize (&optional (module *mod-current-module*)
			   &rest  r &key &allow-other-keys)
	   compile changed submodules of module
	   pass keyword parameters to function 'compile-file'
	   optimization version

      (defun mod-load (&optional (module *mod-current-module*)
		       &rest r &key &allow-other-keys)
	  load module, pass keyword parameters to function 'load'

      (defmacro mod-require (&optional (module *mod-current-module*))
	  if module isn't available, load and starup module

      (defun mod-probe-file (file 
			     &optional (module *mod-current-module*)
			     &rest r &key &allow-other-keys)
	  get path of file 'file' of module 'module'
	  pass keyword parameters to probe-file

      (defun mod-ed (file 
		     &optional (module *mod-current-module*)
		     &rest r &key &allow-other-keys)
          start editor *mod-edit* with file 'file'
	  pass keyword parameters to *mod-edit*

      (defun mod-textedit (file &optional (module *mod-current-module*))
	  start textedit with file 'file'

      (defun mod-vi (file
		     &optional (module *mod-current-module*))
          start vi with file 'file'

      (defun mod-view (file
		       &optional (module *mod-current-module*))
          start vi with file 'file'

   history     : 90.07.04 emacs support added (FW) (Version 6)
   errors      :
   bugs        :
|#
  
(export
    '(user::mod-vi user::mod-set user::mod-load user::mod-require 
      user::mod-require-kern 
      user::mod-compile
      user::*mod-praefix*
      )
    'user
)

(defmacro mod-shell (string)

   `(multiple-value-bind (in-out err return-code proc-nr)
	 (,(if (search " 2.0." (lisp-implementation-version))
	      'run-unix-program 
	      'run-program
	   )
	      "csh"
	      :arguments 
		  (list
		      "-c"
		      ,string
		  )
	  )
	return-code
    )
)

(defmacro mod-quick-shell (string)

   `(multiple-value-bind (in-out err return-code proc-nr)
	 (,(if (search " 2.0." (lisp-implementation-version))
	      'run-unix-program 
	      'run-program
	   )
	      "csh"
	      :arguments 
		  (list
		      "-fc"
		      ,string
		  )
	  )
	return-code
    )
)

(defun mod-is-directory (name)

    (equal (mod-quick-shell (concatenate 'string "test -d '" name "'")) 0)
)

(defvar *mod-module-file-name* "MODULES")

(defvar *mod-current-module* "current module not specified")

(defun mod-set (module)

    (setq *mod-current-module* 
	(mod-string-of module)
    )
)

(defun mod-get () *mod-current-module*)

(defun mod-rel-full-module-name (&optional (module *mod-current-module*) 
			    &key home-dir (praefix "") (with-versions nil))

    (let* ((m (mod-string-of module)   
	   res
	   )
	   (n (if (equal home-dir "")
		 m
		 (concatenate 
		  'string
		   
		  home-dir
		  "/"
		  m
	        )
	     )
	 ))

	(cond
	    ((let ((pos (search ".lisp" n)))
		 (when (and pos (equal (length n) (+ pos 5)))
		     (let ((res (probe-file n)))
			 (when res (namestring res))
		     )
		 )
	     )
	    )
	    ((setq res (or (probe-file (concatenate 'string n ".lisp"))
	                   (probe-file (concatenate 'string n ".sbin"))
	                   (probe-file (concatenate 'string n ".lbin"))
	     ))

	     (subseq (namestring res) 0 (search "." (namestring res) :from-end t))

	    )
	    ((and 
		(not (or with-versions 
			 (probe-file (concatenate 'string n "/startup.lisp"))))
		(probe-file (concatenate 'string n "/v")))

	     (mod-rel-full-module-name (concatenate 'string m "/v") 
		 :home-dir home-dir
		 :praefix (concatenate 'string praefix "/v")
		 :with-versions with-versions
	     )
	    )
	    ((and (setq res (probe-file n))
		  (mod-is-directory n)
	     )

	     (namestring res)
	    )
	    (t nil)
	)
    )
)

(defun mod-rel-sub-modules (&optional (module *mod-current-module*) 
			    &key home-dir (praefix "") (with-versions
			    nil))

    (let* ((m (mod-string-of module)   
	   )
	   (n (mod-rel-full-module-name module :home-dir home-dir
	       :with-versions with-versions)
	 ))

	(cond
	    ; single file module
            ((let ((pos (search ".lisp" n)))
		 (when (and pos (equal (length n) (+ pos 5)))
		     (list n (list ""))
		 )
	     )
	    )
	    ((and n
	     (or (probe-file (concatenate 'string n ".lisp"))
	         (probe-file (concatenate 'string n ".sbin"))
	         (probe-file (concatenate 'string n ".lbin"))
	     )
	     )

	     (list n (list ""))
	    )
	    (n

	     (if (probe-file
		     (concatenate 'string n "/" *mod-module-file-name*))
		 (with-open-file 
		     (f (concatenate 'string n "/" *mod-module-file-name*))
		     (let ((item (read-line f nil nil)) erg)
			 (loop
			     (unless item (return))
			     (setq erg
				 (nconc erg
				     (list (concatenate 'string "/" item))))
			     (setq item (read-line f nil nil))
			 )
			 (list n erg)
		     )
		 )
		 (let ((erg (directory (concatenate 'string n "/"))))
		     (list
			 n
			 (delete-duplicates
			     (mapcan
				 #'(lambda (x)
				       (let ((y (concatenate 'string
						  n "/" 
						  (file-namestring x)
						))
					    )
				       (if (mod-is-directory y)
					  (list (file-namestring x))
					  (let ((n (mod-get-kern x)))
					      (when n (list n))
					  )
				       )
				       )
				   )
				 erg
			     )
			     :test 'equal
			 )
		     )
		 )
	     )
	    )
	    (t nil)
	)
    )
)

(defun mod-string-of (name)

    (if (symbolp name) 
	(symbol-name name)
	name
    )
)

(defun mod-sub-modules (&optional (module *mod-current-module*) &key
(with-versions nil))

    (let ((mn (mod-string-of module)))

	(cond
	    ; absolute name
	    ((search "/" mn :start2 0 :end2 1)
		(mod-rel-sub-modules module :home-dir ""
		    :with-versions with-versions)
	    )
	    ; relative name
	    (t
		(let ((res
			  (or
			  (some
			      #'(lambda (x)
				    (mod-rel-sub-modules
					module
					:home-dir x
					:with-versions with-versions
				    )
				)
				*mod-home-path*
			  )
			  (some
			      #'(lambda (x)
				    (mod-rel-sub-modules
					module
					:home-dir 
					    (concatenate 'string x "/" *mod-current-module*)
					:with-versions with-versions
				    )
				)
				*mod-home-path*
			  )
			  )
		     ))

		    (if res
			res
			(error "Module ~A does not exist~%" mn)
		    )
		)
	    )
	)
    )
)

(defun mod-full-module-name (&optional (module *mod-current-module*)
			     &key (with-versions nil))

    (let ((mn (mod-string-of module)))

	(cond
	    ; absolute name
	    ((search "/" mn :start2 0 :end2 1)
		(mod-rel-full-module-name module :home-dir ""
		    :with-versions with-versions)
	    )
	    ; relative name
	    (t
		(let ((res
			  (some
			      #'(lambda (x)
				    (mod-rel-full-module-name
					module
					:home-dir x
					:with-versions with-versions
				    )
				)
			      *mod-home-path*
			  )
		     ))

		    (if res
			res
			(error "Module ~A does not exist~%" mn)
		    )
		)
	    )
	)
    )
)

(defun mod-get-kern (p)

    (let* ((fn (file-namestring p))
	   (l  (length fn))
	   (in (when (>= l 5)
		   (or (search ".lisp" fn :from-end t :start2 (- l 5))
		       (search ".sbin" fn :from-end t :start2 (- l 5))
		       (search ".lbin" fn :from-end t :start2 (- l 5))
		   )
	       )
	   )
	  )

	(when in
	    (concatenate 'string "/" (subseq fn 0 in))
	)
    )
)


(proclaim '(special *mod-praefix*))

(defvar *mod-praefix* "")

(defun mod-recompile (&optional (module *mod-current-module*)
		     &rest  r &key &allow-other-keys)

    (apply 'mod-comp module t r)
)

(defun mod-compile (&optional (module *mod-current-module*)
		    &rest  r &key &allow-other-keys)

    (apply 'mod-comp module nil r)
)

(defun mod-reoptimize (&rest r)

     (unwind-protect
	 (progn
	     (proclaim '(optimize (compilation-speed 0) (safety 1) (speed 3)))
	     (apply 'mod-recompile r)
	 )
	 (proclaim '(optimize (compilation-speed 3) (safety 3) (speed 2)))
     )
)

(defun mod-optimize (&rest r)

     (unwind-protect
	 (progn
	     (proclaim '(optimize (compilation-speed 0) (safety 1) (speed 3)))
	     (apply 'mod-compile r)
	 )
	 (proclaim '(optimize (compilation-speed 3) (safety 3) (speed 2)))
     )
)

(defun mod-comp    (module force
		    &rest  r &key &allow-other-keys)

    (let* ((mn (mod-string-of module))
	   (sms (mod-sub-modules mn))
	   (*mod-praefix* (car sms))
	   (names (cadr sms))
          )

	(dolist (m names)
	    (let ((fn (concatenate 'string  *mod-praefix* m)))

		(when
		    (or
			force
			(not
			    (file-write-date
				(concatenate 'string
				    fn "." (car *load-binary-pathname-types*)
				)
			    )
			)
			(>
			   (file-write-date
			       (concatenate 'string
				   fn ".lisp"
			       )
			   )
			   (file-write-date
			       (concatenate 'string
				   fn "." (car *load-binary-pathname-types*)
			       )
			   )
			)
                    )

		    (apply 'compile-file fn r)
		)
	    )
	)
    )
)

(defvar *mod-current-loads* nil)
(proclaim '(special *mod-current-loads*))
(defun mod-load (&optional (module *mod-current-module*)
		 &rest r &key &allow-other-keys)

    (let* ((mn (mod-string-of module))
	   (sms (mod-sub-modules mn))
	   (*mod-praefix* (car sms))
	   (names (cadr sms))
          )

	(dolist (m names)
	    (unless (equal m "/startup")
		(apply
		    `load 
		    (concatenate 'string  *mod-praefix* m) 
		    r
		)
		(provide (concatenate 'string  *mod-praefix* m))
	    )
	)
      (provide mn)
      (provide *mod-praefix*)
      )
)

(defmacro mod-require (&optional (module nil module-p))
    `(eval-when (compile load eval)
	 ,(if module-p
	      `(mod-require-kern ,module)
	      `(mod-require-kern)
	  )
     )
)

(defun mod-require-kern (&optional (module *mod-current-module*))

    (let ((mn (mod-string-of module)))
	(unless (member mn *modules* :test 'equal)
	    (let* ((sms (mod-sub-modules mn))
		   (*mod-praefix* (car sms))
		   (names (cadr sms))
		  )
	(unless (or (member *mod-praefix* *modules* :test 'equal)
	            (member *mod-praefix* *mod-current-loads* :test 'equal))
           (let ((*mod-current-loads* (cons *mod-praefix* *mod-current-loads*)))
		(if (member "/startup" names :test 'equal)
		    (load (concatenate 'string *mod-praefix* "/startup"))
		    (dolist (m names)
			(load (concatenate 'string *mod-praefix* m))
		        (provide (concatenate 'string  *mod-praefix* m))
		    )
		)
	    )
	    (provide mn)
	    (provide *mod-praefix*)
	    )
	)
	)
    )
)

(defun mod-probe-file (file 
		       &optional (module *mod-current-module*)
		       &key (with-versions nil))

    (if file
       (let* ( (fn (mod-string-of file))
	      (*mod-praefix* (mod-full-module-name module :with-versions
	      with-versions))
	     )

	   (probe-file (concatenate 'string  *mod-praefix* "/" fn))
       )
       (let* ( (*mod-praefix* (mod-full-module-name module :with-versions
	      with-versions))
	     )

	   (probe-file (concatenate 'string  *mod-praefix* ))
       )
    )
)

(defun mod-ed-file (file module &key (with-versions nil))

    (namestring 
	(or
	    (when file (mod-probe-file 
		(concatenate 'string file ".lisp")
		module
		:with-versions with-versions
	    ))
	    (unless file (mod-probe-file 
		nil
		(concatenate 'string (mod-string-of module) ".lisp")
		:with-versions with-versions
	    ))
	    (let ((m (mod-probe-file file module :with-versions with-versions)))
		(when (and m (not (mod-is-directory (namestring m)))) m)
	    )
	)
    )
)

(defvar *mod-editor* 'mod-textedit)

(defun mod-ed (&optional file (module *mod-current-module*)
	       &rest r &key &allow-other-keys)

    (apply
	*mod-editor*
	file
	module
	r
    )
)

(defun mod-textedit (&optional file (module *mod-current-module*))

    (mod-shell (concatenate 'string 
	       "textedit "
	       "-width 80 -height 35 "
	       (mod-ed-file file module)
	       " &"
	   )
    )
)

#|
(defun mod-textedit (file
		     &optional (module *mod-current-module*))

    (let* ((nam (mod-ed-file file module))
	   (dir (directory-namestring nam))
	   (fil (file-namestring nam))
	  )

	(mod-shell (format nil
		  "(cd ~A ; ~
		   textedit ~@[~A~] ~
		      -width 80 -height 35 ~
		      ~@[-icon_image ~A~]   &"
		   dir
		   nam
		     ; if there is a icon-file *cases-homedir*/cases-ed.icon
		     ; this icon is used for textedit.
		)    
	)
    )
))
|#

(defun mod-vi (&optional file
	       (module *mod-current-module*))

    (let* ((nam (mod-ed-file file module))
	   (dir (directory-namestring nam))
	   (fil (file-namestring nam))
	  )

	(mod-shell
	    (concatenate 'string
		"( cd "
		dir
		"; vi "
		fil
		" )"
	    )
	)
    )
)

(defun mod-view ( &optional file (module *mod-current-module*))

    (mod-shell
	(concatenate 'string
	    "view "
	    (mod-ed-file file module)
	)
    )
)

(defun mod-emacs (&optional file (module *mod-current-module*))

    (ed (mod-ed-file file module))
)
