#| 
   module      : try
   owner       : Franz Weber
   author      : Franz Weber
   explanation : graphical lisp library browser
   history     : 
   errors      :
   bugs        :
|#

(mod-require "gra/v3")

(defun mod-try-filter (name)
    (dolist (postfix '(".sbin" ".s3bin" ".lbin" ".l3bin" "%" "~"))
	(when
	    (let ((found (search postfix name)))
	       (and found
		  (equal
		      (+ found (length postfix))
		      (length name)
		  )
	       )
	    )
	    (return-from mod-try-filter nil)
	)
    )
    t
)

(defmacro mod-try-adjust (list line pathname)
   `(let ((name (enough-namestring 
		    (subseq ,line 0 (search ":" ,line)) ,pathname)))
       (when (mod-try-filter name)
	  (setf ,list (cons name ,list))
       )
    )
)

(defun mod-try-find-modules (pathname &aux texts directories)

   (let ((dd (directory (concatenate 'string pathname "/")))
	 (pn (probe-file (concatenate 'string pathname "/")))
	)
   (when dd
   (with-open-stream 
       (s (run-program
	      "/bin/file"
	      :arguments
	      (mapcar
		 #'(lambda (x)
			(namestring x)
	            )
		 dd
	      )
	      :output :stream
	      :wait nil
	  ))

       (loop
	   (let ((line (read-line s nil)))
	       (unless line (return))
	       (cond
		   ((or (search " directory" line)
			(search (concatenate 'string (string #\tab) "directory") line)
		    )
		      (mod-try-adjust directories line pn)
		   )
		   ((search " text" line)
		      (mod-try-adjust texts line pn)
		   )
		   ((or (search " empty" line)
			(search 
			    (concatenate 'string (string #\tab) "empty") line)
		    )
		      (mod-try-adjust texts line pn)
		   )
	       )
	   )
       )
   )
   )
   )
   (values texts directories)
)

(defflavor browser (pn module) (item-field)
    (:settable-instance-variables module)
    (:gettable-instance-variables module)
    (:init-keywords :module)
)

(defmethod (browser :open) (pathname)
    (with-fast-drawing-environment
    (setf pn pathname)
    [:set-title pn]
    (multiple-value-bind (texts directories) (mod-try-find-modules pn)
    [:transfer-s-exp
    (append
	(list
	   (make-instance 'menu-item :selectable t :picture "quit/reload"
	       :procedure
	       #'(lambda (father item event)
		     (case event
		         (:mouse-left-down  [father :set-visibility nil]
		                            [the-menues :remove-elem father
					    :test 'eql])
		         (:mouse-right-down [father :reload])
		     )
		 )
	   )
	)
       (sort
       (append
	   (mapcar 
	       #'(lambda (e) (make-instance 'text-module :pathname pn :module e))
		 texts
	   )
	   (mapcar
	       #'(lambda (e) (make-instance 'directory-module :pathname pn :module e))
	       directories
	   )
	)
	#'(lambda (x y)
	      (string< [x :picture] [y :picture])
	  )
	)
     )
   ]
   )
   )
)

(defmethod (browser :reload) ()
    (with-mouse-methods-preempted nil [:open pn])
)

(defmethod (browser :expose) ()
    (unless [the-menues :is-elem self :test 'eql]
	  [:set-visibility t]
	  [the-menues :move-in self :to-pos 'last]
    )
)

(defmethod (browser :selected) (event)
    [module :selected event]
)

(defmethod (browser :touched) (event)
    ;[module :touched event]
)

; -------------------- text modules
(defflavor text-module 
    (pn spn editor) (menu-item comparison copies) 
    (:init-keywords :module :pathname))

(definitmethod (text-module) (&key pathname module) 
    (setf editor nil)
    (setf pn (concatenate 'string pathname "/" module))
    (setf spn
       (let ((i (search ".lisp" pn :from-end t)))
	   (if i
	     (subseq  pn 0 i)
	     pn
	   )
       )
    )
    (setf picture module)
    (setf procedure 
	 #'(lambda (father item event)
	       (case event
		  (:mouse-left-down
		      [item :inspect]
		  )   
		  (:mouse-middle-down
		      [item :load]
		  )   
		  (:mouse-right-down
		      [item :menu-choice]
		  )   
	       )  
	   )   
    )
)

(defmethod (text-module :compare) (right) 
    (flavor-compare picture [right :picture])
)

(defmethod (text-module :inspect) ()
    (unless editor
	 (setq editor (make-instance 'aneditor))
	 [editor :open pn ]
    )
    [editor :expose]
)

(defmethod (text-module :load) () 
    (mod-load spn :if-source-newer :load-source)
)

(defmethod (text-module :menu-choice) () 
    (make-menu :title "Options" :x (mouse-x) :y (mouse-y) :visibility t
	:init-value `(
	("require" ,#'(lambda (father item event) [father :close] (mod-require spn)) ) 
	("load" ,#'(lambda (father item event) [father :close] (mod-load spn
	:if-source-newer :load-source)))
	("load-source" ,#'(lambda (father item event) [father :close] (load pn
	)) ) 
	("compile" ,#'(lambda (father item event) [father :close] (mod-compile spn)) ) 
	("recompile" ,#'(lambda (father item event) [father :close] (mod-recompile spn)) ) 
	("edit" ,#'(lambda (father item event) [father :close] [:inspect] ) )
	("remove" ,#'(lambda (fath item event)
			 (when (equal (car (gra-input
			     (format nil "do you want to remove ~a ?" pn)
			      :x (mouse-x) :y (mouse-y) :edit-length 0
			      :exits '("yes" "no")
			       )) "yes")
			 (shell (format nil "rm ~a" pn))
			 [father :remove-elem self :test 'eql]
			 )
			 [fath :close]
		     )
	)
	("print" ,#'(lambda (father item event) 
			(shell (format nil "enscript ~a" pn))
			[father :close]
		    )
	)
	("quit" ,#'(lambda (father item event) [father :close]))
	)
    )
)

; -------------------- editor
(defflavor aneditor (pn) ())
(defmethod (aneditor :open) (pathname) (setf pn pathname))

(defmethod (aneditor :expose) () 
    (run-program "/bin/shelltool" 
	:arguments 
	`("-font" "/fzi/prost/fweber/icon/bigfont" 
	  "-label" ,(file-namestring pn)
	  "-icon_label" ,(file-namestring pn)
	  "vi" ,pn) 
	:wait nil)
)


; -------------------- directory modules
(defflavor directory-module 
    (editor pn) (menu-item comparison copies) 
    (:settable-instance-variables editor)
    (:init-keywords :module :pathname)
)

(definitmethod (directory-module) (&key pathname module) 
    (setf editor nil)
    (setf pn (concatenate 'string pathname "/" module))
    (setf picture module)
    (setf procedure 
	 #'(lambda (father item event)
	       (case event
		  (:mouse-left-down
		      [item :inspect]
		  )   
		  (:mouse-middle-down
		      [item :load]
		  )   
		  (:mouse-right-down
		      [item :menu-choice]
		  )   
	       )  
	   )   
    )
)

(defmethod (directory-module :compare) (right)
    (flavor-compare picture [right :picture])
)

(defmethod (directory-module :inspect) ()
    (if editor
	 (let ((i [the-menues :find editor :test 'eql]))
	     (when i [the-menues :move `(at-index ,i) :to-pos 'last])
	 )
	(progn
	   (setq editor 
	       (make-instance 'browser :visibility nil :module self))
	   [editor :open pn ]
	 )
    )
    [editor :expose]
)

(defmethod (directory-module :load) ()

    (mod-load pn :if-source-newer :load-source)
)

(defmethod (directory-module :menu-choice) () 
    (make-menu :title "Options" :x (mouse-x) :y (mouse-y) :visibility t
	:init-value `(
	("require" ,#'(lambda (father item event) [father :close] (mod-require pn)) ) 
	("load" ,#'(lambda (father item event) [father :close] (mod-load pn
	:if-source-newer :load-source)) ) 
	("compile" ,#'(lambda (father item event) [father :close] (mod-compile pn)) ) 
	("recompile" ,#'(lambda (father item event) [father :close] (mod-recompile pn)) ) 
	("edit" ,#'(lambda (father item event) [father :close] [:inspect] ) )
	("remove" ,#'(lambda (fath item event)
			 (when (equal (car (gra-input
			     (format nil "do you want to remove ~a ?" pn)
			      :x (mouse-x) :y (mouse-y) :edit-length 0
			      :exits '("yes" "no")
			       )) "yes")
			 (shell (format nil "rm -r ~a" pn))
			 [father :remove-elem self :test 'eql]
			 )
			 [fath :close]
		     )
	)
	("insert-file" ,#'(lambda (fath item event)
	     (let ((name (format nil "touch ~a/~a" pn 
			      (cdr (gra-input "Enter file name" 
				       :x (mouse-x) :y (mouse-y))))))
		(shell (format nil "touch ~a" name))
		(when editor [editor :reload])
		[fath :close]
	    )
	    )
	)
	("insert-directory" ,#'(lambda (fath item event)
			 (shell (format nil "mkdir ~a/~a" pn
			     (cdr (gra-input "Enter directory name" 
				      :x (mouse-x) :y (mouse-y)))))
		         (when editor [editor :reload])
			 [fath :close]
		     )
	)
	("shelltool" ,#'(lambda (fath item event)
	    (let ((pwd (pwd)))
	       (cd pn)
	       (run-program "/bin/shelltool" 
		   :arguments `("-font" "/fzi/prost/fweber/icon/bigfont" 
				"-label" ,pn
				"-icon_label" ,pn
			       ) 
		   :wait nil)
	       (cd pwd)
	    )
	    [fath :close]
	    )
	)
	("cmdtool" ,#'(lambda (fath item event)
	    (let ((pwd (pwd)))
	       (cd pn)
	       (run-program "/bin/cmdtool" 
		   :arguments `("-font" "/fzi/prost/fweber/icon/bigfont"
				"-label" ,pn
				"-icon_label" ,pn
			       ) 
		   :wait nil)
	       (cd pwd)
	    )
	    [fath :close]
	    )
	)
	("quit" ,#'(lambda (father item event) [father :close]))
	)
    )
)


(defun mod-try-start ()

;(when (boundp 'try-root) [try-root :close])
(setq try-module (make-instance 'directory-module 
		      :pathname "/fzi/prost/lisp/lib" :module "/"))
(setq try-root (make-instance 'browser :visibility nil :module
try-module))

[try-module :set-editor try-root]

[try-root :open "/fzi/prost/lisp/lib"]

#|
(when (boundp 'the-menues)
    (iterate
	(e the-menues)
	[e :close]
    )
)
|#
(setq the-menues (make-instance 'h-list :x 0 :y 0))

(setq ttt
    (make-instance 'v-list :x 0 :y 0 :init-value (list try-root the-menues)))

[try-root :set-visibility t]

)
(compile-flavor-methods text-module directory-module browser aneditor)
