#| 
   module      : tra
   owner       : Franz Weber
   author      : Franz Weber
   explanation : Tracing Support
   history     :

	 05.06.89

   errors      :
|#

(require 'otr "/fzi/prost/lisp/lib/earley/v2/otr")

(defvar *tra-specials* nil)

; (unless (search " 2.0." (lisp-implementation-version))

(defun tra-place (f s)

    ;`(getf (symbol-plist (quote ,f)) (quote ,s))
    (string-symbol (cat
	"*tra-var-" 
	(symbol-name f) 
	"-" 
	(symbol-name s)
	"*"
      ))
)

(defun tra-symbols-of (module)

    (let ((s (cat (symbol-name module) "-"))
	  (h nil))

	(do-symbols (v 'user)
	     (when (equal (search s (symbol-name v)) 0)
		 (setq h (cons v h))
	     )
	 )
	h
    )
)

(defmacro tra-do-symbols (head body)

    (let ((args (cadr head))
	  (var  (car head))
	 )
	`(dolist (arg ,args)
	     (if (listp arg)
		(dolist (,var arg) (eval ,body))
		(dolist (,var (tra-symbols-of arg)) (eval ,body))
	     )
	 )
    )
)

(defmacro tra-reset-measure (&rest fs)

    `(progn
	 (tra-reset-count ,@fs)
	 (tra-reset-time ,@fs)
     )
)

(defmacro tra-unmeasure (&rest fs)

    (tra-do-symbols (f fs)
	`(progn
	     (remove-advice (quote ,f) 'tra-measure)
	     (remprop (quote ,f) 'tra-measure)
	     t
	 )
    )
)

(defmacro tra-measure (&rest fs)

    (tra-do-symbols (f fs)
	`(when (fboundp (quote ,f))
	    (tra-reset-measure (,f))
	    (proclaim '(bignum ,(tra-place f 'tra-time)))
	    (proclaim '(fixnum ,(tra-place f 'tra-count)))
	    (let ((f1 (compile nil 
		       #'(lambda ()
			 ;(declare (optimize (safety 0)))
			 (incf (the fixnum ,(tra-place f 'tra-count)))
			 (the bignum
			   (- (the bignum ,(tra-place f 'tra-time))
			       (the bignum (get-internal-run-time)))
			 ))))
		  (f2 (compile nil
		       #'(lambda (start-time)
			 ;(declare (optimize (safety 0)))
			 (declare (bignum start-time))
			 (setf ,(tra-place f 'tra-time)
			       (the bignum 
				   (+  (the bignum (get-internal-run-time))
					start-time))
		         )(values))))
		 )
		(defadvice
		    (,f tra-measure) (&rest args)
		    (let ((start-time (funcall f1)))
			(prog1
			    (apply-advice-continue args)
			    (funcall f2 start-time)
			)
		    )
		)
	    )
	)
    )
)

(defmacro tra-get-measure (&rest fs)

    (tra-do-symbols (f fs)
	`(when (and (boundp (quote ,(tra-place f 'tra-count)))
		    (< 0 ,(tra-place f 'tra-count))
	       )
	    (format t "~A~20T : ~10@A (times) ~12,3F (secs)~%" (quote ,f)
		,(tra-place f 'tra-count)
		(float (/ ,(tra-place f 'tra-time)
		    internal-time-units-per-second))
	    )
	 )
    )
)

(defmacro tra-reset-count (&rest fs)

    (tra-do-symbols (f fs)
	  `(setf ,(tra-place f 'tra-count) 0)
      )
  )

  (defmacro tra-count (&rest fs)

      (tra-do-symbols (f fs)
	  `(when (fboundp (quote ,f))
	      (tra-reset-count (,f))
	      (let ((f1 (compile nil
			 #'(lambda ()  (incf (the fixnum ,(tra-place f 'tra-count)))))))
		  (defadvice
		      (,f tra-count) (&rest args)
		      (funcall f1)
		      (apply-advice-continue args)
		  )
	      )
	  )
      )
  )

  (defmacro tra-uncount (&rest fs)

      (tra-do-symbols (f fs)
	  `(progn
	       (remove-advice (quote ,f) 'tra-count)
	       (remprop (quote ,f) 'tra-count)
	       t
	   )
      )
  )

  (defmacro tra-get-count (&rest fs)

      (tra-do-symbols (f fs)
	  `(when (and (boundp (quote ,(tra-place f 'tra-count)))
		      (< 0 ,(tra-place f 'tra-count))
		 )
	      (format t "~A~20T : ~10@A (times)~%" (quote ,f)
		  ,(tra-place f 'tra-count)
	      )
	   )
      )
  )

  (defmacro tra-reset-time (&rest fs)

      (tra-do-symbols (f fs)
	  `(setf ,(tra-place f 'tra-time) 0)
    )
)

(defmacro tra-time (&rest fs)

    (tra-do-symbols (f fs)
	`(when (fboundp (quote ,f))
	    (tra-reset-time (,f))
	    (proclaim '(bignum ,(tra-place f 'tra-time)))
	    (let ((f1 (compile nil 
		       #'(lambda ()
			 (- ,(tra-place f 'tra-time) (get-internal-run-time))
			 )))
		  (f2 (compile nil
		       #'(lambda (start-time)
			 (declare (bignum start-time))
			 (setf ,(tra-place f 'tra-time)
			       (+  (get-internal-run-time) start-time)
		         ))))
		 )
		(defadvice
		    (,f tra-time) (&rest args)
		    (let ((start-time (funcall f1)))
			(prog1
			    (apply-advice-continue args)
			    (funcall f2 start-time)
			)
		    )
		)
	    )
	)
    )
)

(defmacro tra-untime (&rest fs)

    (tra-do-symbols (f fs)
	`(progn
	     (remove-advice (quote ,f) 'tra-time)
	     (remprop (quote ,f) 'tra-time)
	     t
	 )
    )
)

(defmacro tra-get-time (&rest fs)

    (tra-do-symbols (f fs)
	`(when (and (boundp (quote ,(tra-place f 'tra-time)))
	            (< 0 ,(tra-place f 'tra-time))
	       )
	     (format t
		 "~A~20T : ~12,3F (secs)~%" (quote ,f)
		 (float (/ ,(tra-place f 'tra-time)
		    internal-time-units-per-second))
	     )
	 )
    )
)

; ) ; unless (search " 2.0." (lisp-implementation-version))

(defmacro tra (&rest l)

    (cons
	'trace
	(mapcar
	    #'(lambda (x)
	          `(,x :entry ((tra-entry)) :exit ((tra-exit)))
	      )
	    l
 	)
    )
)

(defmacro tra-special (&rest v)

    `(setq *tra-specials* (append (quote ,v) *tra-specials*))
)

(defmacro un-tra-special () (setq *tra-specials* nil))	

(defmacro un-tra (&rest l)

    (cons `untrace l)
)
  
(defun tra-entry ()

    (format t "~%")
    (mapc
	#'(lambda (x)
	      (tra-print x)
	  )
	*trace-arglist*
    )
    (mapc
	#'(lambda (x)
	      (print x)
	      (tra-print (eval x))
	  )
	*tra-specials*
    )
    (values)
)

(defun tra-exit ()

    (format t "~%")
    (mapc
	#'(lambda (x)
	      (tra-print x)
	  )
	*trace-values*
    )
    (values)
)

(defun tra-print (x)

    (cond
	((op-tree-p x)
	    (tra-print-op-tree x)
	)
	((hash-table-p x)
	    (tra-print-hash-table x)
	)
	((listp x)
	    (format t "(")
	    (tra-print-list x)
	    (format t ")~%")
	)
	(t (print x))
    )
)

(defun tra-print-list (x)

    (when x
	(tra-print (car x))
	(if (listp (cdr x))
	    (tra-print-list (cdr x))
	    (progn
		(format t " . ")
		(tra-print (cdr x))
	    )
	)
    )
)

(defun tra-print-op-tree (x)

    (cond
	((otr-is-a *gra-nonterminal* (otr-tag (otr-tag x)))
	    (ppr-format x)
	    (format t "~%")
	)
	(t (otr-format-tree x))
    )
)

(defun tra-print-hash-table (x)

    (format t "[")
    (maphash
	#'(lambda (x y) (tra-print x) (tra-print y))
	x
    )
    (format t "]")
)

(provide `tra)
